{*******************************************************}
{                                                       }
{         Delphi VCL Extensions (RX)                    }
{                                                       }
{         Copyright (c) 2010 Jaro Benes                 }
{                                                       }
{         Main library's language file integrator       }
{                                                       }
{*******************************************************}
unit RxResConst;
{$WARNINGS OFF}
{$HINTS OFF}
interface
{$I RX.INC}

{$IFDEF UNICODE}
  {$DEFINE _LNG_ALL_}
{$ELSE}
  {$IFDEF RX_D9}
    {$DEFINE _LNG_ALL_}
  {$ENDIF}
{$ENDIF}

{$IFDEF _LNG_ONE_}
  {$UNDEF _LNG_ALL_}
{$ENDIF}

//------------------------------------------------------------------------------
// NOTE: for compilers before Delphi 2005 cannot be use utf8 files
// English is used by default because file is non-UTF8 file (pure ASCII)
// For different language by Englist, convert own UTF8 file to locale ASCII
// and write name of directive to section for older Delphi version
//------------------------------------------------------------------------------

{$IFDEF _LNG_ALL_}
  {$DEFINE RXLANG_Bul}
  {$DEFINE RXLANG_Chi}
  {$DEFINE RXLANG_Cht}
  {$DEFINE RXLANG_Cze}
  {$DEFINE RXLANG_Dan}
  {$DEFINE RXLANG_Dut}
  {$DEFINE RXLANG_Eng}
  {$DEFINE RXLANG_Fra}
  {$DEFINE RXLANG_Gal}
  {$DEFINE RXLANG_Ger}
  {$DEFINE RXLANG_Gre}
  {$DEFINE RXLANG_Hun}
  {$DEFINE RXLANG_Ita}
  {$DEFINE RXLANG_Kor}
  {$DEFINE RXLANG_Mex}
  {$DEFINE RXLANG_Pol}
  {$DEFINE RXLANG_Rom}
  {$DEFINE RXLANG_Rus}
  {$DEFINE RXLANG_Slo}
  {$DEFINE RXLANG_Spa}
  {$DEFINE RXLANG_Srb}
  {$DEFINE RXLANG_Svk}
  {$DEFINE RXLANG_Swe}
  {$DEFINE RXLANG_Tur}
  {$DEFINE RXLANG_Ukr}
{$ELSE}
//==============================================================================
// ASCII file for Delphi 7, 6, 5, and older...
//==============================================================================
  {$DEFINE RXLANG_Eng} //default language for not UTF8 files
{$ENDIF}

type
  TRxLang = (rlNone, rlBul, rlChi, rlCht, rlCze, rlDan, rlDut, rlEng, rlFra, rlGal,
    rlGer, rlGre, rlHun, rlIta, rlKor, rlMex, rlPol, rlPor, rlRom, rlRus, rlSlo,
    rlSpa, rlSrb, rlSvk, rlSwe, rlTur, rlUkr);

//==============================================================================
// Global variable for set up Language for all messages of Rx Library
// only well translated (basic names origin from Englist).
//==============================================================================
var
  RxDefaultLang: TRxLang = {$IFDEF UNICODE}rlEng{$ELSE}rlCze{$ENDIF};

function RxLoadStr(IDMsg: Integer): string;

//==============================================================================
// Constants for simple evaluate symbolic name to returned message string
//==============================================================================
const
  { RXCConst }
  SBrowse = 1;
  SDefaultFilter = 2;
  SDateDlgTitle = 3;
  SNextYear = 4;
  SNextMonth = 5;
  SPrevYear = 6;
  SPrevMonth = 7;
  SNotImplemented = 8;
  SFileNotExec = 9;
  SLoadLibError = 10;
  SDetails = 11;
  SDateOutOfRange = 12;
  SDateOutOfMin = 13;
  SDateOutOfMax = 14;
  SDateMinLimit = 15;
  SDateMaxLimit = 16;
  { RXDConst }
  SLocalDatabase = 17;
  SRetryLogin = 18;
  SExprNotBoolean = 19;
  SExprBadNullTest = 20;
  SExprBadField = 21;
  SCaptureFilter = 22;
  SNotCaptureFilter = 23;
  SInactiveData = 24;
  SBrowseData = 25;
  SEditData = 26;
  SInsertData = 27;
  SSetKeyData = 28;
  SCalcFieldsData = 29;
  SRegistration = 30;
  SAppTitleLabel = 31;
  SHintLabel = 32;
  SUserNameLabel = 33;
  SPasswordLabel = 34;
  SInvalidUserName = 35;
  SChangePassword = 36;
  SOldPasswordLabel = 37;
  SNewPasswordLabel = 38;
  SConfirmPasswordLabel = 39;
  SPasswordChanged = 40;
  SPasswordNotChanged = 41;
  SPasswordsMismatch = 42;
  SDBExceptCaption = 43;
  SBDEErrorLabel = 44;
  SServerErrorLabel = 45;
  SErrorMsgLabel = 46;
  SNextButton = 47;
  SPrevButton = 48;
  SExprIncorrect = 49;
  SExprTermination = 50;
  SExprNameError = 51;
  SExprStringError = 52;
  SExprInvalidChar = 53;
  SExprNoRParen = 54;
  SExprExpected = 55;
  SExprBadCompare = 56;
  SConfirmSave = 57;
  SDatabaseName = 58;
  SUnlockCaption = 59;
  SUnlockHint = 60;
  SDeleteMultipleRecords = 61;
  { RXGConst }
  SGIFImage = 62;
  SChangeGIFSize = 63;
  SNoGIFData = 64;
  SUnrecognizedGIFExt = 65;
  SWrongGIFColors = 66;
  SBadGIFCodeSize = 67;
  SGIFDecodeError = 68;
  SGIFEncodeError = 69;
  SGIFVersion = 70;
  { RXTConst }
  SDualListSrcCaption = 71;
  SDualListDestCaption = 72;
  SClipbrdUnknown = 73;
  SClipbrdEmpty = 74;
  SCustomizeSpeedbar = 75;
  SAvailButtons = 76;
  SSpeedbarCategories = 77;
  SSpeedbarEditHint = 78;
  SParseSyntaxError = 79;
  SParseNotCramp = 80;
  SParseDivideByZero = 81;
  SParseSqrError = 82;
  SParseLogError = 83;
  SParseInvalidFloatOperation = 84;
  { RXLConst }
  srRXControls = 85;
  srRXDBAware = 86;
  srRXTools = 87;

  srStorageDesigner = 88;

  srProxyEditor = 89;
  srPageProxies = 90;
  srProxyName = 91;
  srPageName = 92;

  srSBItemNotCreate = 93;
  srConfirmSBDelete = 94;
  srSpeedbarDesigner = 95;
  srNewSectionName = 96;

  srEventNotCreate = 97;
  srTimerDesigner = 98;
  srTimerEvents = 99;

  srAniCurFilter = 100;
  srEditPicture = 101;
  srLoadAniCursor = 102;

  srLoadIcon = 103;
  srSaveImageList = 104;

  srCaptionDesigner = 105;
  srGradientCaptions = 106;
  srBorrowStructure = 107;

  ccTimeEditHours = 108;
  ccTimeEditMins = 109;
  ccTimeEditSecs = 110;

  { additional color constant names }
  SColorBlack = 111;
  SColorMaroon = 112;
  SColorGreen = 113;
  SColorOlive = 114;
  SColorNavy = 115;
  SColorPurple = 116;
  SColorTeal = 117;
  SColorGray = 118;
  SColorSilver = 119;
  SColorRed = 120;
  SColorLime = 121;
  SColorYellow = 122;
  SColorBlue = 123;
  SColorFuchsia = 124;
  SColorAqua = 125;
  SColorWhite = 126;
  SColorScrollBar = 127;
  SColorBackground = 128;
  SColorActiveCaption = 129;
  SColorInactiveCaption = 130;
  SColorMenu = 131;
  SColorWindow = 132;
  SColorWindowFrame = 133;
  SColorMenuText = 134;
  SColorWindowText = 135;
  SColorCaptionText = 136;
  SColorActiveBorder = 137;
  SColorInactiveBorder = 138;
  SColorAppWorkSpace = 139;
  SColorHighlight = 140;
  SColorHighlightText = 141;
  SColorBtnFace = 142;
  SColorBtnShadow = 173;
  SColorGrayText = 144;
  SColorBtnText = 145;
  SColorInactiveCaptionText = 146;
  SColorBtnHighlight = 147;
  SColor3DDkShadow = 148;
  SColor3DLight = 149;
  SColorInfoText = 150;
  SColorInfoBk = 151;
  SColorCream = 152;
  SColorMoneyGreen = 153;
  SColorSkyBlue = 154;
  SColorNone = 155;
  SColorDefault = 156;
  SColorCustom = 157;
  SMemNoRecords = 158;
  SInvalidFields = 159;
  SAutoSpeedbarMode = 160;

implementation

uses
  {$IFDEF RXLANG_Bul}RxResConst_bul,{$ENDIF}
  {$IFDEF RXLANG_Chi}RxResConst_chi,{$ENDIF}
  {$IFDEF RXLANG_Cht}RxResConst_cht,{$ENDIF}
  {$IFDEF RXLANG_Cze}RxResConst_cze,{$ENDIF}
  {$IFDEF RXLANG_Dan}RxResConst_dan,{$ENDIF}
  {$IFDEF RXLANG_Dut}RxResConst_dut,{$ENDIF}
  {$IFDEF RXLANG_Eng}RxResConst_eng,{$ENDIF}
  {$IFDEF RXLANG_Fra}RxResConst_fra,{$ENDIF}
  {$IFDEF RXLANG_Gal}RxResConst_gal,{$ENDIF}
  {$IFDEF RXLANG_Ger}RxResConst_ger,{$ENDIF}
  {$IFDEF RXLANG_Gre}RxResConst_gre,{$ENDIF}
  {$IFDEF RXLANG_Hun}RxResConst_hun,{$ENDIF}
  {$IFDEF RXLANG_Ita}RxResConst_ita,{$ENDIF}
  {$IFDEF RXLANG_Kor}RxResConst_kor,{$ENDIF}
  {$IFDEF RXLANG_Mex}RxResConst_mex,{$ENDIF}
  {$IFDEF RXLANG_Pol}RxResConst_pol,{$ENDIF}
  {$IFDEF RXLANG_Por}RxResConst_por,{$ENDIF}
  {$IFDEF RXLANG_Rom}RxResConst_rom,{$ENDIF}
  {$IFDEF RXLANG_Rus}RxResConst_rus,{$ENDIF}
  {$IFDEF RXLANG_Slo}RxResConst_slo,{$ENDIF}
  {$IFDEF RXLANG_Spa}RxResConst_spa,{$ENDIF}
  {$IFDEF RXLANG_Srb}RxResConst_srb,{$ENDIF}
  {$IFDEF RXLANG_Svk}RxResConst_svk,{$ENDIF}
  {$IFDEF RXLANG_Swe}RxResConst_swe,{$ENDIF}
  {$IFDEF RXLANG_Tur}RxResConst_tur,{$ENDIF}
  {$IFDEF RXLANG_Ukr}RxResConst_ukr,{$ENDIF}
  Sysutils, Windows;

{-------------------------------------------------------------------------------
  Procedure: RxLoadStr
  Author:    jaro.benes
  DateTime:  2011.03.02
  Arguments: IDMsg: Integer
  Result:    string
  Purpose:   Function brings message string by number identifier
-------------------------------------------------------------------------------}
function RxLoadStr(IDMsg: Integer): string;
  {$IFDEF RXLANG_Bul}
  function RxResConst_Bul_RxLoadStr(IDMsg: Integer): string;
  begin
    case IDMsg of
     SBrowse : Result := RxResConst_Bul.SBrowse;
     SDefaultFilter : Result := RxResConst_Bul.SDefaultFilter;
     SDateDlgTitle : Result := RxResConst_Bul.SDateDlgTitle;
     SNextYear : Result := RxResConst_Bul.SNextYear;
     SNextMonth : Result := RxResConst_Bul.SNextMonth;
     SPrevYear : Result := RxResConst_Bul.SPrevYear;
     SPrevMonth : Result := RxResConst_Bul.SPrevMonth;
     SNotImplemented : Result := RxResConst_Bul.SNotImplemented;
     SFileNotExec : Result := RxResConst_Bul.SFileNotExec;
     SLoadLibError : Result := RxResConst_Bul.SLoadLibError;
     SDetails : Result := RxResConst_Bul.SDetails;
     SDateOutOfRange : Result := RxResConst_Bul.SDateOutOfRange;
     SDateOutOfMin : Result := RxResConst_Bul.SDateOutOfMin;
     SDateOutOfMax : Result := RxResConst_Bul.SDateOutOfMax;
     SDateMinLimit : Result := RxResConst_Bul.SDateMinLimit;
     SDateMaxLimit : Result := RxResConst_Bul.SDateMaxLimit;
     SLocalDatabase : Result := RxResConst_Bul.SLocalDatabase;
     SRetryLogin : Result := RxResConst_Bul.SRetryLogin;
     SExprNotBoolean : Result := RxResConst_Bul.SExprNotBoolean;
     SExprBadNullTest : Result := RxResConst_Bul.SExprBadNullTest;
     SExprBadField : Result := RxResConst_Bul.SExprBadField;
     SCaptureFilter : Result := RxResConst_Bul.SCaptureFilter;
     SNotCaptureFilter : Result := RxResConst_Bul.SNotCaptureFilter;
     SInactiveData : Result := RxResConst_Bul.SInactiveData;
     SBrowseData : Result := RxResConst_Bul.SBrowseData;
     SEditData : Result := RxResConst_Bul.SEditData;
     SInsertData : Result := RxResConst_Bul.SInsertData;
     SSetKeyData : Result := RxResConst_Bul.SSetKeyData;
     SCalcFieldsData : Result := RxResConst_Bul.SCalcFieldsData;
     SRegistration : Result := RxResConst_Bul.SRegistration;
     SAppTitleLabel : Result := RxResConst_Bul.SAppTitleLabel;
     SHintLabel : Result := RxResConst_Bul.SHintLabel;
     SUserNameLabel : Result := RxResConst_Bul.SUserNameLabel;
     SPasswordLabel : Result := RxResConst_Bul.SPasswordLabel;
     SInvalidUserName : Result := RxResConst_Bul.SInvalidUserName;
     SChangePassword : Result := RxResConst_Bul.SChangePassword;
     SOldPasswordLabel : Result := RxResConst_Bul.SOldPasswordLabel;
     SNewPasswordLabel : Result := RxResConst_Bul.SNewPasswordLabel;
     SConfirmPasswordLabel : Result := RxResConst_Bul.SConfirmPasswordLabel;
     SPasswordChanged : Result := RxResConst_Bul.SPasswordChanged;
     SPasswordNotChanged : Result := RxResConst_Bul.SPasswordNotChanged;
     SPasswordsMismatch : Result := RxResConst_Bul.SPasswordsMismatch;
     SDBExceptCaption : Result := RxResConst_Bul.SDBExceptCaption;
     SBDEErrorLabel : Result := RxResConst_Bul.SBDEErrorLabel;
     SServerErrorLabel : Result := RxResConst_Bul.SServerErrorLabel;
     SErrorMsgLabel : Result := RxResConst_Bul.SErrorMsgLabel;
     SNextButton : Result := RxResConst_Bul.SNextButton;
     SPrevButton : Result := RxResConst_Bul.SPrevButton;
     SExprIncorrect : Result := RxResConst_Bul.SExprIncorrect;
     SExprTermination : Result := RxResConst_Bul.SExprTermination;
     SExprNameError : Result := RxResConst_Bul.SExprNameError;
     SExprStringError : Result := RxResConst_Bul.SExprStringError;
     SExprInvalidChar : Result := RxResConst_Bul.SExprInvalidChar;
     SExprNoRParen : Result := RxResConst_Bul.SExprNoRParen;
     SExprExpected : Result := RxResConst_Bul.SExprExpected;
     SExprBadCompare : Result := RxResConst_Bul.SExprBadCompare;
     SConfirmSave : Result := RxResConst_Bul.SConfirmSave;
     SDatabaseName : Result := RxResConst_Bul.SDatabaseName;
     SUnlockCaption : Result := RxResConst_Bul.SUnlockCaption;
     SUnlockHint : Result := RxResConst_Bul.SUnlockHint;
     SDeleteMultipleRecords : Result := RxResConst_Bul.SDeleteMultipleRecords;
     SGIFImage : Result := RxResConst_Bul.SGIFImage;
     SChangeGIFSize : Result := RxResConst_Bul.SChangeGIFSize;
     SNoGIFData : Result := RxResConst_Bul.SNoGIFData;
     SUnrecognizedGIFExt : Result := RxResConst_Bul.SUnrecognizedGIFExt;
     SWrongGIFColors : Result := RxResConst_Bul.SWrongGIFColors;
     SBadGIFCodeSize : Result := RxResConst_Bul.SBadGIFCodeSize;
     SGIFDecodeError : Result := RxResConst_Bul.SGIFDecodeError;
     SGIFEncodeError : Result := RxResConst_Bul.SGIFEncodeError;
     SGIFVersion : Result := RxResConst_Bul.SGIFVersion;
     SDualListSrcCaption : Result := RxResConst_Bul.SDualListSrcCaption;
     SDualListDestCaption : Result := RxResConst_Bul.SDualListDestCaption;
     SClipbrdUnknown : Result := RxResConst_Bul.SClipbrdUnknown;
     SClipbrdEmpty : Result := RxResConst_Bul.SClipbrdEmpty;
     SCustomizeSpeedbar : Result := RxResConst_Bul.SCustomizeSpeedbar;
     SAvailButtons : Result := RxResConst_Bul.SAvailButtons;
     SSpeedbarCategories : Result := RxResConst_Bul.SSpeedbarCategories;
     SSpeedbarEditHint : Result := RxResConst_Bul.SSpeedbarEditHint;
     SParseSyntaxError : Result := RxResConst_Bul.SParseSyntaxError;
     SParseNotCramp : Result := RxResConst_Bul.SParseNotCramp;
     SParseDivideByZero : Result := RxResConst_Bul.SParseDivideByZero;
     SParseSqrError : Result := RxResConst_Bul.SParseSqrError;
     SParseLogError : Result := RxResConst_Bul.SParseLogError;
     SParseInvalidFloatOperation : Result := RxResConst_Bul.SParseInvalidFloatOperation;
     srRXControls : Result := RxResConst_Bul.srRXControls;
     srRXDBAware : Result := RxResConst_Bul.srRXDBAware;
     srRXTools : Result := RxResConst_Bul.srRXTools;
     srStorageDesigner : Result := RxResConst_Bul.srStorageDesigner;
     srProxyEditor : Result := RxResConst_Bul.srProxyEditor;
     srPageProxies : Result := RxResConst_Bul.srPageProxies;
     srProxyName : Result := RxResConst_Bul.srProxyName;
     srPageName : Result := RxResConst_Bul.srPageName;
     srSBItemNotCreate : Result := RxResConst_Bul.srSBItemNotCreate;
     srConfirmSBDelete : Result := RxResConst_Bul.srConfirmSBDelete;
     srSpeedbarDesigner : Result := RxResConst_Bul.srSpeedbarDesigner;
     srNewSectionName : Result := RxResConst_Bul.srNewSectionName;
     srEventNotCreate : Result := RxResConst_Bul.srEventNotCreate;
     srTimerDesigner : Result := RxResConst_Bul.srTimerDesigner;
     srTimerEvents : Result := RxResConst_Bul.srTimerEvents;
     srAniCurFilter : Result := RxResConst_Bul.srAniCurFilter;
     srEditPicture : Result := RxResConst_Bul.srEditPicture;
     srLoadAniCursor : Result := RxResConst_Bul.srLoadAniCursor;
     srLoadIcon : Result := RxResConst_Bul.srLoadIcon;
     srSaveImageList : Result := RxResConst_Bul.srSaveImageList;
     srCaptionDesigner : Result := RxResConst_Bul.srCaptionDesigner;
     srGradientCaptions : Result := RxResConst_Bul.srGradientCaptions;
     srBorrowStructure : Result := RxResConst_Bul.srBorrowStructure;
     ccTimeEditHours : Result := RxResConst_Bul.ccTimeEditHours;
     ccTimeEditMins : Result := RxResConst_Bul.ccTimeEditMins;
     ccTimeEditSecs : Result := RxResConst_Bul.ccTimeEditSecs;
     SColorBlack : Result := RxResConst_Bul.SColorBlack;
     SColorMaroon : Result := RxResConst_Bul.SColorMaroon;
     SColorGreen : Result := RxResConst_Bul.SColorGreen;
     SColorOlive : Result := RxResConst_Bul.SColorOlive;
     SColorNavy : Result := RxResConst_Bul.SColorNavy;
     SColorPurple : Result := RxResConst_Bul.SColorPurple;
     SColorTeal : Result := RxResConst_Bul.SColorTeal;
     SColorGray : Result := RxResConst_Bul.SColorGray;
     SColorSilver : Result := RxResConst_Bul.SColorSilver;
     SColorRed : Result := RxResConst_Bul.SColorRed;
     SColorLime : Result := RxResConst_Bul.SColorLime;
     SColorYellow : Result := RxResConst_Bul.SColorYellow;
     SColorBlue : Result := RxResConst_Bul.SColorBlue;
     SColorFuchsia : Result := RxResConst_Bul.SColorFuchsia;
     SColorAqua : Result := RxResConst_Bul.SColorAqua;
     SColorWhite : Result := RxResConst_Bul.SColorWhite;
     SColorScrollBar : Result := RxResConst_Bul.SColorScrollBar;
     SColorBackground : Result := RxResConst_Bul.SColorBackground;
     SColorActiveCaption : Result := RxResConst_Bul.SColorActiveCaption;
     SColorInactiveCaption : Result := RxResConst_Bul.SColorInactiveCaption;
     SColorMenu : Result := RxResConst_Bul.SColorMenu;
     SColorWindow : Result := RxResConst_Bul.SColorWindow;
     SColorWindowFrame : Result := RxResConst_Bul.SColorWindowFrame;
     SColorMenuText : Result := RxResConst_Bul.SColorMenuText;
     SColorWindowText : Result := RxResConst_Bul.SColorWindowText;
     SColorCaptionText : Result := RxResConst_Bul.SColorCaptionText;
     SColorActiveBorder : Result := RxResConst_Bul.SColorActiveBorder;
     SColorInactiveBorder : Result := RxResConst_Bul.SColorInactiveBorder;
     SColorAppWorkSpace : Result := RxResConst_Bul.SColorAppWorkSpace;
     SColorHighlight : Result := RxResConst_Bul.SColorHighlight;
     SColorHighlightText : Result := RxResConst_Bul.SColorHighlightText;
     SColorBtnFace : Result := RxResConst_Bul.SColorBtnFace;
     SColorBtnShadow : Result := RxResConst_Bul.SColorBtnShadow;
     SColorGrayText : Result := RxResConst_Bul.SColorGrayText;
     SColorBtnText : Result := RxResConst_Bul.SColorBtnText;
     SColorInactiveCaptionText : Result := RxResConst_Bul.SColorInactiveCaptionText;
     SColorBtnHighlight : Result := RxResConst_Bul.SColorBtnHighlight;
     SColor3DDkShadow : Result := RxResConst_Bul.SColor3DDkShadow;
     SColor3DLight : Result := RxResConst_Bul.SColor3DLight;
     SColorInfoText : Result := RxResConst_Bul.SColorInfoText;
     SColorInfoBk : Result := RxResConst_Bul.SColorInfoBk;
     SColorCream : Result := RxResConst_Bul.SColorCream;
     SColorMoneyGreen : Result := RxResConst_Bul.SColorMoneyGreen;
     SColorSkyBlue : Result := RxResConst_Bul.SColorSkyBlue;
     SColorNone : Result := RxResConst_Bul.SColorNone;
     SColorDefault : Result := RxResConst_Bul.SColorDefault;
     SColorCustom : Result := RxResConst_Bul.SColorCustom;
     SMemNoRecords : Result := RxResConst_Bul.SMemNoRecords;
     SInvalidFields : Result := RxResConst_Bul.SInvalidFields;
     SAutoSpeedbarMode : Result := RxResConst_Bul.SAutoSpeedbarMode;
    end;
  end;
  {$ENDIF}
  {$IFDEF RXLANG_Chi}
  function RxResConst_Chi_RxLoadStr(IDMsg: Integer): string;
  begin
    case IDMsg of
     SBrowse : Result := RxResConst_Chi.SBrowse;
     SDefaultFilter : Result := RxResConst_Chi.SDefaultFilter;
     SDateDlgTitle : Result := RxResConst_Chi.SDateDlgTitle;
     SNextYear : Result := RxResConst_Chi.SNextYear;
     SNextMonth : Result := RxResConst_Chi.SNextMonth;
     SPrevYear : Result := RxResConst_Chi.SPrevYear;
     SPrevMonth : Result := RxResConst_Chi.SPrevMonth;
     SNotImplemented : Result := RxResConst_Chi.SNotImplemented;
     SFileNotExec : Result := RxResConst_Chi.SFileNotExec;
     SLoadLibError : Result := RxResConst_Chi.SLoadLibError;
     SDetails : Result := RxResConst_Chi.SDetails;
     SDateOutOfRange : Result := RxResConst_Chi.SDateOutOfRange;
     SDateOutOfMin : Result := RxResConst_Chi.SDateOutOfMin;
     SDateOutOfMax : Result := RxResConst_Chi.SDateOutOfMax;
     SDateMinLimit : Result := RxResConst_Chi.SDateMinLimit;
     SDateMaxLimit : Result := RxResConst_Chi.SDateMaxLimit;
     SLocalDatabase : Result := RxResConst_Chi.SLocalDatabase;
     SRetryLogin : Result := RxResConst_Chi.SRetryLogin;
     SExprNotBoolean : Result := RxResConst_Chi.SExprNotBoolean;
     SExprBadNullTest : Result := RxResConst_Chi.SExprBadNullTest;
     SExprBadField : Result := RxResConst_Chi.SExprBadField;
     SCaptureFilter : Result := RxResConst_Chi.SCaptureFilter;
     SNotCaptureFilter : Result := RxResConst_Chi.SNotCaptureFilter;
     SInactiveData : Result := RxResConst_Chi.SInactiveData;
     SBrowseData : Result := RxResConst_Chi.SBrowseData;
     SEditData : Result := RxResConst_Chi.SEditData;
     SInsertData : Result := RxResConst_Chi.SInsertData;
     SSetKeyData : Result := RxResConst_Chi.SSetKeyData;
     SCalcFieldsData : Result := RxResConst_Chi.SCalcFieldsData;
     SRegistration : Result := RxResConst_Chi.SRegistration;
     SAppTitleLabel : Result := RxResConst_Chi.SAppTitleLabel;
     SHintLabel : Result := RxResConst_Chi.SHintLabel;
     SUserNameLabel : Result := RxResConst_Chi.SUserNameLabel;
     SPasswordLabel : Result := RxResConst_Chi.SPasswordLabel;
     SInvalidUserName : Result := RxResConst_Chi.SInvalidUserName;
     SChangePassword : Result := RxResConst_Chi.SChangePassword;
     SOldPasswordLabel : Result := RxResConst_Chi.SOldPasswordLabel;
     SNewPasswordLabel : Result := RxResConst_Chi.SNewPasswordLabel;
     SConfirmPasswordLabel : Result := RxResConst_Chi.SConfirmPasswordLabel;
     SPasswordChanged : Result := RxResConst_Chi.SPasswordChanged;
     SPasswordNotChanged : Result := RxResConst_Chi.SPasswordNotChanged;
     SPasswordsMismatch : Result := RxResConst_Chi.SPasswordsMismatch;
     SDBExceptCaption : Result := RxResConst_Chi.SDBExceptCaption;
     SBDEErrorLabel : Result := RxResConst_Chi.SBDEErrorLabel;
     SServerErrorLabel : Result := RxResConst_Chi.SServerErrorLabel;
     SErrorMsgLabel : Result := RxResConst_Chi.SErrorMsgLabel;
     SNextButton : Result := RxResConst_Chi.SNextButton;
     SPrevButton : Result := RxResConst_Chi.SPrevButton;
     SExprIncorrect : Result := RxResConst_Chi.SExprIncorrect;
     SExprTermination : Result := RxResConst_Chi.SExprTermination;
     SExprNameError : Result := RxResConst_Chi.SExprNameError;
     SExprStringError : Result := RxResConst_Chi.SExprStringError;
     SExprInvalidChar : Result := RxResConst_Chi.SExprInvalidChar;
     SExprNoRParen : Result := RxResConst_Chi.SExprNoRParen;
     SExprExpected : Result := RxResConst_Chi.SExprExpected;
     SExprBadCompare : Result := RxResConst_Chi.SExprBadCompare;
     SConfirmSave : Result := RxResConst_Chi.SConfirmSave;
     SDatabaseName : Result := RxResConst_Chi.SDatabaseName;
     SUnlockCaption : Result := RxResConst_Chi.SUnlockCaption;
     SUnlockHint : Result := RxResConst_Chi.SUnlockHint;
     SDeleteMultipleRecords : Result := RxResConst_Chi.SDeleteMultipleRecords;
     SGIFImage : Result := RxResConst_Chi.SGIFImage;
     SChangeGIFSize : Result := RxResConst_Chi.SChangeGIFSize;
     SNoGIFData : Result := RxResConst_Chi.SNoGIFData;
     SUnrecognizedGIFExt : Result := RxResConst_Chi.SUnrecognizedGIFExt;
     SWrongGIFColors : Result := RxResConst_Chi.SWrongGIFColors;
     SBadGIFCodeSize : Result := RxResConst_Chi.SBadGIFCodeSize;
     SGIFDecodeError : Result := RxResConst_Chi.SGIFDecodeError;
     SGIFEncodeError : Result := RxResConst_Chi.SGIFEncodeError;
     SGIFVersion : Result := RxResConst_Chi.SGIFVersion;
     SDualListSrcCaption : Result := RxResConst_Chi.SDualListSrcCaption;
     SDualListDestCaption : Result := RxResConst_Chi.SDualListDestCaption;
     SClipbrdUnknown : Result := RxResConst_Chi.SClipbrdUnknown;
     SClipbrdEmpty : Result := RxResConst_Chi.SClipbrdEmpty;
     SCustomizeSpeedbar : Result := RxResConst_Chi.SCustomizeSpeedbar;
     SAvailButtons : Result := RxResConst_Chi.SAvailButtons;
     SSpeedbarCategories : Result := RxResConst_Chi.SSpeedbarCategories;
     SSpeedbarEditHint : Result := RxResConst_Chi.SSpeedbarEditHint;
     SParseSyntaxError : Result := RxResConst_Chi.SParseSyntaxError;
     SParseNotCramp : Result := RxResConst_Chi.SParseNotCramp;
     SParseDivideByZero : Result := RxResConst_Chi.SParseDivideByZero;
     SParseSqrError : Result := RxResConst_Chi.SParseSqrError;
     SParseLogError : Result := RxResConst_Chi.SParseLogError;
     SParseInvalidFloatOperation : Result := RxResConst_Chi.SParseInvalidFloatOperation;
     srRXControls : Result := RxResConst_Chi.srRXControls;
     srRXDBAware : Result := RxResConst_Chi.srRXDBAware;
     srRXTools : Result := RxResConst_Chi.srRXTools;
     srStorageDesigner : Result := RxResConst_Chi.srStorageDesigner;
     srProxyEditor : Result := RxResConst_Chi.srProxyEditor;
     srPageProxies : Result := RxResConst_Chi.srPageProxies;
     srProxyName : Result := RxResConst_Chi.srProxyName;
     srPageName : Result := RxResConst_Chi.srPageName;
     srSBItemNotCreate : Result := RxResConst_Chi.srSBItemNotCreate;
     srConfirmSBDelete : Result := RxResConst_Chi.srConfirmSBDelete;
     srSpeedbarDesigner : Result := RxResConst_Chi.srSpeedbarDesigner;
     srNewSectionName : Result := RxResConst_Chi.srNewSectionName;
     srEventNotCreate : Result := RxResConst_Chi.srEventNotCreate;
     srTimerDesigner : Result := RxResConst_Chi.srTimerDesigner;
     srTimerEvents : Result := RxResConst_Chi.srTimerEvents;
     srAniCurFilter : Result := RxResConst_Chi.srAniCurFilter;
     srEditPicture : Result := RxResConst_Chi.srEditPicture;
     srLoadAniCursor : Result := RxResConst_Chi.srLoadAniCursor;
     srLoadIcon : Result := RxResConst_Chi.srLoadIcon;
     srSaveImageList : Result := RxResConst_Chi.srSaveImageList;
     srCaptionDesigner : Result := RxResConst_Chi.srCaptionDesigner;
     srGradientCaptions : Result := RxResConst_Chi.srGradientCaptions;
     srBorrowStructure : Result := RxResConst_Chi.srBorrowStructure;
     ccTimeEditHours : Result := RxResConst_Chi.ccTimeEditHours;
     ccTimeEditMins : Result := RxResConst_Chi.ccTimeEditMins;
     ccTimeEditSecs : Result := RxResConst_Chi.ccTimeEditSecs;
     SColorBlack : Result := RxResConst_Chi.SColorBlack;
     SColorMaroon : Result := RxResConst_Chi.SColorMaroon;
     SColorGreen : Result := RxResConst_Chi.SColorGreen;
     SColorOlive : Result := RxResConst_Chi.SColorOlive;
     SColorNavy : Result := RxResConst_Chi.SColorNavy;
     SColorPurple : Result := RxResConst_Chi.SColorPurple;
     SColorTeal : Result := RxResConst_Chi.SColorTeal;
     SColorGray : Result := RxResConst_Chi.SColorGray;
     SColorSilver : Result := RxResConst_Chi.SColorSilver;
     SColorRed : Result := RxResConst_Chi.SColorRed;
     SColorLime : Result := RxResConst_Chi.SColorLime;
     SColorYellow : Result := RxResConst_Chi.SColorYellow;
     SColorBlue : Result := RxResConst_Chi.SColorBlue;
     SColorFuchsia : Result := RxResConst_Chi.SColorFuchsia;
     SColorAqua : Result := RxResConst_Chi.SColorAqua;
     SColorWhite : Result := RxResConst_Chi.SColorWhite;
     SColorScrollBar : Result := RxResConst_Chi.SColorScrollBar;
     SColorBackground : Result := RxResConst_Chi.SColorBackground;
     SColorActiveCaption : Result := RxResConst_Chi.SColorActiveCaption;
     SColorInactiveCaption : Result := RxResConst_Chi.SColorInactiveCaption;
     SColorMenu : Result := RxResConst_Chi.SColorMenu;
     SColorWindow : Result := RxResConst_Chi.SColorWindow;
     SColorWindowFrame : Result := RxResConst_Chi.SColorWindowFrame;
     SColorMenuText : Result := RxResConst_Chi.SColorMenuText;
     SColorWindowText : Result := RxResConst_Chi.SColorWindowText;
     SColorCaptionText : Result := RxResConst_Chi.SColorCaptionText;
     SColorActiveBorder : Result := RxResConst_Chi.SColorActiveBorder;
     SColorInactiveBorder : Result := RxResConst_Chi.SColorInactiveBorder;
     SColorAppWorkSpace : Result := RxResConst_Chi.SColorAppWorkSpace;
     SColorHighlight : Result := RxResConst_Chi.SColorHighlight;
     SColorHighlightText : Result := RxResConst_Chi.SColorHighlightText;
     SColorBtnFace : Result := RxResConst_Chi.SColorBtnFace;
     SColorBtnShadow : Result := RxResConst_Chi.SColorBtnShadow;
     SColorGrayText : Result := RxResConst_Chi.SColorGrayText;
     SColorBtnText : Result := RxResConst_Chi.SColorBtnText;
     SColorInactiveCaptionText : Result := RxResConst_Chi.SColorInactiveCaptionText;
     SColorBtnHighlight : Result := RxResConst_Chi.SColorBtnHighlight;
     SColor3DDkShadow : Result := RxResConst_Chi.SColor3DDkShadow;
     SColor3DLight : Result := RxResConst_Chi.SColor3DLight;
     SColorInfoText : Result := RxResConst_Chi.SColorInfoText;
     SColorInfoBk : Result := RxResConst_Chi.SColorInfoBk;
     SColorCream : Result := RxResConst_Chi.SColorCream;
     SColorMoneyGreen : Result := RxResConst_Chi.SColorMoneyGreen;
     SColorSkyBlue : Result := RxResConst_Chi.SColorSkyBlue;
     SColorNone : Result := RxResConst_Chi.SColorNone;
     SColorDefault : Result := RxResConst_Chi.SColorDefault;
     SColorCustom : Result := RxResConst_Chi.SColorCustom;
     SMemNoRecords : Result := RxResConst_Chi.SMemNoRecords;
     SInvalidFields : Result := RxResConst_Chi.SInvalidFields;
     SAutoSpeedbarMode : Result := RxResConst_Chi.SAutoSpeedbarMode;
    end;
  end;
  {$ENDIF}
  {$IFDEF RXLANG_Cht}
  function RxResConst_Cht_RxLoadStr(IDMsg: Integer): string;
  begin
    case IDMsg of
     SBrowse : Result := RxResConst_Cht.SBrowse;
     SDefaultFilter : Result := RxResConst_Cht.SDefaultFilter;
     SDateDlgTitle : Result := RxResConst_Cht.SDateDlgTitle;
     SNextYear : Result := RxResConst_Cht.SNextYear;
     SNextMonth : Result := RxResConst_Cht.SNextMonth;
     SPrevYear : Result := RxResConst_Cht.SPrevYear;
     SPrevMonth : Result := RxResConst_Cht.SPrevMonth;
     SNotImplemented : Result := RxResConst_Cht.SNotImplemented;
     SFileNotExec : Result := RxResConst_Cht.SFileNotExec;
     SLoadLibError : Result := RxResConst_Cht.SLoadLibError;
     SDetails : Result := RxResConst_Cht.SDetails;
     SDateOutOfRange : Result := RxResConst_Cht.SDateOutOfRange;
     SDateOutOfMin : Result := RxResConst_Cht.SDateOutOfMin;
     SDateOutOfMax : Result := RxResConst_Cht.SDateOutOfMax;
     SDateMinLimit : Result := RxResConst_Cht.SDateMinLimit;
     SDateMaxLimit : Result := RxResConst_Cht.SDateMaxLimit;
     SLocalDatabase : Result := RxResConst_Cht.SLocalDatabase;
     SRetryLogin : Result := RxResConst_Cht.SRetryLogin;
     SExprNotBoolean : Result := RxResConst_Cht.SExprNotBoolean;
     SExprBadNullTest : Result := RxResConst_Cht.SExprBadNullTest;
     SExprBadField : Result := RxResConst_Cht.SExprBadField;
     SCaptureFilter : Result := RxResConst_Cht.SCaptureFilter;
     SNotCaptureFilter : Result := RxResConst_Cht.SNotCaptureFilter;
     SInactiveData : Result := RxResConst_Cht.SInactiveData;
     SBrowseData : Result := RxResConst_Cht.SBrowseData;
     SEditData : Result := RxResConst_Cht.SEditData;
     SInsertData : Result := RxResConst_Cht.SInsertData;
     SSetKeyData : Result := RxResConst_Cht.SSetKeyData;
     SCalcFieldsData : Result := RxResConst_Cht.SCalcFieldsData;
     SRegistration : Result := RxResConst_Cht.SRegistration;
     SAppTitleLabel : Result := RxResConst_Cht.SAppTitleLabel;
     SHintLabel : Result := RxResConst_Cht.SHintLabel;
     SUserNameLabel : Result := RxResConst_Cht.SUserNameLabel;
     SPasswordLabel : Result := RxResConst_Cht.SPasswordLabel;
     SInvalidUserName : Result := RxResConst_Cht.SInvalidUserName;
     SChangePassword : Result := RxResConst_Cht.SChangePassword;
     SOldPasswordLabel : Result := RxResConst_Cht.SOldPasswordLabel;
     SNewPasswordLabel : Result := RxResConst_Cht.SNewPasswordLabel;
     SConfirmPasswordLabel : Result := RxResConst_Cht.SConfirmPasswordLabel;
     SPasswordChanged : Result := RxResConst_Cht.SPasswordChanged;
     SPasswordNotChanged : Result := RxResConst_Cht.SPasswordNotChanged;
     SPasswordsMismatch : Result := RxResConst_Cht.SPasswordsMismatch;
     SDBExceptCaption : Result := RxResConst_Cht.SDBExceptCaption;
     SBDEErrorLabel : Result := RxResConst_Cht.SBDEErrorLabel;
     SServerErrorLabel : Result := RxResConst_Cht.SServerErrorLabel;
     SErrorMsgLabel : Result := RxResConst_Cht.SErrorMsgLabel;
     SNextButton : Result := RxResConst_Cht.SNextButton;
     SPrevButton : Result := RxResConst_Cht.SPrevButton;
     SExprIncorrect : Result := RxResConst_Cht.SExprIncorrect;
     SExprTermination : Result := RxResConst_Cht.SExprTermination;
     SExprNameError : Result := RxResConst_Cht.SExprNameError;
     SExprStringError : Result := RxResConst_Cht.SExprStringError;
     SExprInvalidChar : Result := RxResConst_Cht.SExprInvalidChar;
     SExprNoRParen : Result := RxResConst_Cht.SExprNoRParen;
     SExprExpected : Result := RxResConst_Cht.SExprExpected;
     SExprBadCompare : Result := RxResConst_Cht.SExprBadCompare;
     SConfirmSave : Result := RxResConst_Cht.SConfirmSave;
     SDatabaseName : Result := RxResConst_Cht.SDatabaseName;
     SUnlockCaption : Result := RxResConst_Cht.SUnlockCaption;
     SUnlockHint : Result := RxResConst_Cht.SUnlockHint;
     SDeleteMultipleRecords : Result := RxResConst_Cht.SDeleteMultipleRecords;
     SGIFImage : Result := RxResConst_Cht.SGIFImage;
     SChangeGIFSize : Result := RxResConst_Cht.SChangeGIFSize;
     SNoGIFData : Result := RxResConst_Cht.SNoGIFData;
     SUnrecognizedGIFExt : Result := RxResConst_Cht.SUnrecognizedGIFExt;
     SWrongGIFColors : Result := RxResConst_Cht.SWrongGIFColors;
     SBadGIFCodeSize : Result := RxResConst_Cht.SBadGIFCodeSize;
     SGIFDecodeError : Result := RxResConst_Cht.SGIFDecodeError;
     SGIFEncodeError : Result := RxResConst_Cht.SGIFEncodeError;
     SGIFVersion : Result := RxResConst_Cht.SGIFVersion;
     SDualListSrcCaption : Result := RxResConst_Cht.SDualListSrcCaption;
     SDualListDestCaption : Result := RxResConst_Cht.SDualListDestCaption;
     SClipbrdUnknown : Result := RxResConst_Cht.SClipbrdUnknown;
     SClipbrdEmpty : Result := RxResConst_Cht.SClipbrdEmpty;
     SCustomizeSpeedbar : Result := RxResConst_Cht.SCustomizeSpeedbar;
     SAvailButtons : Result := RxResConst_Cht.SAvailButtons;
     SSpeedbarCategories : Result := RxResConst_Cht.SSpeedbarCategories;
     SSpeedbarEditHint : Result := RxResConst_Cht.SSpeedbarEditHint;
     SParseSyntaxError : Result := RxResConst_Cht.SParseSyntaxError;
     SParseNotCramp : Result := RxResConst_Cht.SParseNotCramp;
     SParseDivideByZero : Result := RxResConst_Cht.SParseDivideByZero;
     SParseSqrError : Result := RxResConst_Cht.SParseSqrError;
     SParseLogError : Result := RxResConst_Cht.SParseLogError;
     SParseInvalidFloatOperation : Result := RxResConst_Cht.SParseInvalidFloatOperation;
     srRXControls : Result := RxResConst_Cht.srRXControls;
     srRXDBAware : Result := RxResConst_Cht.srRXDBAware;
     srRXTools : Result := RxResConst_Cht.srRXTools;
     srStorageDesigner : Result := RxResConst_Cht.srStorageDesigner;
     srProxyEditor : Result := RxResConst_Cht.srProxyEditor;
     srPageProxies : Result := RxResConst_Cht.srPageProxies;
     srProxyName : Result := RxResConst_Cht.srProxyName;
     srPageName : Result := RxResConst_Cht.srPageName;
     srSBItemNotCreate : Result := RxResConst_Cht.srSBItemNotCreate;
     srConfirmSBDelete : Result := RxResConst_Cht.srConfirmSBDelete;
     srSpeedbarDesigner : Result := RxResConst_Cht.srSpeedbarDesigner;
     srNewSectionName : Result := RxResConst_Cht.srNewSectionName;
     srEventNotCreate : Result := RxResConst_Cht.srEventNotCreate;
     srTimerDesigner : Result := RxResConst_Cht.srTimerDesigner;
     srTimerEvents : Result := RxResConst_Cht.srTimerEvents;
     srAniCurFilter : Result := RxResConst_Cht.srAniCurFilter;
     srEditPicture : Result := RxResConst_Cht.srEditPicture;
     srLoadAniCursor : Result := RxResConst_Cht.srLoadAniCursor;
     srLoadIcon : Result := RxResConst_Cht.srLoadIcon;
     srSaveImageList : Result := RxResConst_Cht.srSaveImageList;
     srCaptionDesigner : Result := RxResConst_Cht.srCaptionDesigner;
     srGradientCaptions : Result := RxResConst_Cht.srGradientCaptions;
     srBorrowStructure : Result := RxResConst_Cht.srBorrowStructure;
     ccTimeEditHours : Result := RxResConst_Cht.ccTimeEditHours;
     ccTimeEditMins : Result := RxResConst_Cht.ccTimeEditMins;
     ccTimeEditSecs : Result := RxResConst_Cht.ccTimeEditSecs;
     SColorBlack : Result := RxResConst_Cht.SColorBlack;
     SColorMaroon : Result := RxResConst_Cht.SColorMaroon;
     SColorGreen : Result := RxResConst_Cht.SColorGreen;
     SColorOlive : Result := RxResConst_Cht.SColorOlive;
     SColorNavy : Result := RxResConst_Cht.SColorNavy;
     SColorPurple : Result := RxResConst_Cht.SColorPurple;
     SColorTeal : Result := RxResConst_Cht.SColorTeal;
     SColorGray : Result := RxResConst_Cht.SColorGray;
     SColorSilver : Result := RxResConst_Cht.SColorSilver;
     SColorRed : Result := RxResConst_Cht.SColorRed;
     SColorLime : Result := RxResConst_Cht.SColorLime;
     SColorYellow : Result := RxResConst_Cht.SColorYellow;
     SColorBlue : Result := RxResConst_Cht.SColorBlue;
     SColorFuchsia : Result := RxResConst_Cht.SColorFuchsia;
     SColorAqua : Result := RxResConst_Cht.SColorAqua;
     SColorWhite : Result := RxResConst_Cht.SColorWhite;
     SColorScrollBar : Result := RxResConst_Cht.SColorScrollBar;
     SColorBackground : Result := RxResConst_Cht.SColorBackground;
     SColorActiveCaption : Result := RxResConst_Cht.SColorActiveCaption;
     SColorInactiveCaption : Result := RxResConst_Cht.SColorInactiveCaption;
     SColorMenu : Result := RxResConst_Cht.SColorMenu;
     SColorWindow : Result := RxResConst_Cht.SColorWindow;
     SColorWindowFrame : Result := RxResConst_Cht.SColorWindowFrame;
     SColorMenuText : Result := RxResConst_Cht.SColorMenuText;
     SColorWindowText : Result := RxResConst_Cht.SColorWindowText;
     SColorCaptionText : Result := RxResConst_Cht.SColorCaptionText;
     SColorActiveBorder : Result := RxResConst_Cht.SColorActiveBorder;
     SColorInactiveBorder : Result := RxResConst_Cht.SColorInactiveBorder;
     SColorAppWorkSpace : Result := RxResConst_Cht.SColorAppWorkSpace;
     SColorHighlight : Result := RxResConst_Cht.SColorHighlight;
     SColorHighlightText : Result := RxResConst_Cht.SColorHighlightText;
     SColorBtnFace : Result := RxResConst_Cht.SColorBtnFace;
     SColorBtnShadow : Result := RxResConst_Cht.SColorBtnShadow;
     SColorGrayText : Result := RxResConst_Cht.SColorGrayText;
     SColorBtnText : Result := RxResConst_Cht.SColorBtnText;
     SColorInactiveCaptionText : Result := RxResConst_Cht.SColorInactiveCaptionText;
     SColorBtnHighlight : Result := RxResConst_Cht.SColorBtnHighlight;
     SColor3DDkShadow : Result := RxResConst_Cht.SColor3DDkShadow;
     SColor3DLight : Result := RxResConst_Cht.SColor3DLight;
     SColorInfoText : Result := RxResConst_Cht.SColorInfoText;
     SColorInfoBk : Result := RxResConst_Cht.SColorInfoBk;
     SColorCream : Result := RxResConst_Cht.SColorCream;
     SColorMoneyGreen : Result := RxResConst_Cht.SColorMoneyGreen;
     SColorSkyBlue : Result := RxResConst_Cht.SColorSkyBlue;
     SColorNone : Result := RxResConst_Cht.SColorNone;
     SColorDefault : Result := RxResConst_Cht.SColorDefault;
     SColorCustom : Result := RxResConst_Cht.SColorCustom;
     SMemNoRecords : Result := RxResConst_Cht.SMemNoRecords;
     SInvalidFields : Result := RxResConst_Cht.SInvalidFields;
     SAutoSpeedbarMode : Result := RxResConst_Cht.SAutoSpeedbarMode;
    end;
  end;
  {$ENDIF}
  {$IFDEF RXLANG_Cze}
  function RxResConst_Cze_RxLoadStr(IDMsg: Integer): string;
  begin
    case IDMsg of
     SBrowse : Result := RxResConst_Cze.SBrowse;
     SDefaultFilter : Result := RxResConst_Cze.SDefaultFilter;
     SDateDlgTitle : Result := RxResConst_Cze.SDateDlgTitle;
     SNextYear : Result := RxResConst_Cze.SNextYear;
     SNextMonth : Result := RxResConst_Cze.SNextMonth;
     SPrevYear : Result := RxResConst_Cze.SPrevYear;
     SPrevMonth : Result := RxResConst_Cze.SPrevMonth;
     SNotImplemented : Result := RxResConst_Cze.SNotImplemented;
     SFileNotExec : Result := RxResConst_Cze.SFileNotExec;
     SLoadLibError : Result := RxResConst_Cze.SLoadLibError;
     SDetails : Result := RxResConst_Cze.SDetails;
     SDateOutOfRange : Result := RxResConst_Cze.SDateOutOfRange;
     SDateOutOfMin : Result := RxResConst_Cze.SDateOutOfMin;
     SDateOutOfMax : Result := RxResConst_Cze.SDateOutOfMax;
     SDateMinLimit : Result := RxResConst_Cze.SDateMinLimit;
     SDateMaxLimit : Result := RxResConst_Cze.SDateMaxLimit;
     SLocalDatabase : Result := RxResConst_Cze.SLocalDatabase;
     SRetryLogin : Result := RxResConst_Cze.SRetryLogin;
     SExprNotBoolean : Result := RxResConst_Cze.SExprNotBoolean;
     SExprBadNullTest : Result := RxResConst_Cze.SExprBadNullTest;
     SExprBadField : Result := RxResConst_Cze.SExprBadField;
     SCaptureFilter : Result := RxResConst_Cze.SCaptureFilter;
     SNotCaptureFilter : Result := RxResConst_Cze.SNotCaptureFilter;
     SInactiveData : Result := RxResConst_Cze.SInactiveData;
     SBrowseData : Result := RxResConst_Cze.SBrowseData;
     SEditData : Result := RxResConst_Cze.SEditData;
     SInsertData : Result := RxResConst_Cze.SInsertData;
     SSetKeyData : Result := RxResConst_Cze.SSetKeyData;
     SCalcFieldsData : Result := RxResConst_Cze.SCalcFieldsData;
     SRegistration : Result := RxResConst_Cze.SRegistration;
     SAppTitleLabel : Result := RxResConst_Cze.SAppTitleLabel;
     SHintLabel : Result := RxResConst_Cze.SHintLabel;
     SUserNameLabel : Result := RxResConst_Cze.SUserNameLabel;
     SPasswordLabel : Result := RxResConst_Cze.SPasswordLabel;
     SInvalidUserName : Result := RxResConst_Cze.SInvalidUserName;
     SChangePassword : Result := RxResConst_Cze.SChangePassword;
     SOldPasswordLabel : Result := RxResConst_Cze.SOldPasswordLabel;
     SNewPasswordLabel : Result := RxResConst_Cze.SNewPasswordLabel;
     SConfirmPasswordLabel : Result := RxResConst_Cze.SConfirmPasswordLabel;
     SPasswordChanged : Result := RxResConst_Cze.SPasswordChanged;
     SPasswordNotChanged : Result := RxResConst_Cze.SPasswordNotChanged;
     SPasswordsMismatch : Result := RxResConst_Cze.SPasswordsMismatch;
     SDBExceptCaption : Result := RxResConst_Cze.SDBExceptCaption;
     SBDEErrorLabel : Result := RxResConst_Cze.SBDEErrorLabel;
     SServerErrorLabel : Result := RxResConst_Cze.SServerErrorLabel;
     SErrorMsgLabel : Result := RxResConst_Cze.SErrorMsgLabel;
     SNextButton : Result := RxResConst_Cze.SNextButton;
     SPrevButton : Result := RxResConst_Cze.SPrevButton;
     SExprIncorrect : Result := RxResConst_Cze.SExprIncorrect;
     SExprTermination : Result := RxResConst_Cze.SExprTermination;
     SExprNameError : Result := RxResConst_Cze.SExprNameError;
     SExprStringError : Result := RxResConst_Cze.SExprStringError;
     SExprInvalidChar : Result := RxResConst_Cze.SExprInvalidChar;
     SExprNoRParen : Result := RxResConst_Cze.SExprNoRParen;
     SExprExpected : Result := RxResConst_Cze.SExprExpected;
     SExprBadCompare : Result := RxResConst_Cze.SExprBadCompare;
     SConfirmSave : Result := RxResConst_Cze.SConfirmSave;
     SDatabaseName : Result := RxResConst_Cze.SDatabaseName;
     SUnlockCaption : Result := RxResConst_Cze.SUnlockCaption;
     SUnlockHint : Result := RxResConst_Cze.SUnlockHint;
     SDeleteMultipleRecords : Result := RxResConst_Cze.SDeleteMultipleRecords;
     SGIFImage : Result := RxResConst_Cze.SGIFImage;
     SChangeGIFSize : Result := RxResConst_Cze.SChangeGIFSize;
     SNoGIFData : Result := RxResConst_Cze.SNoGIFData;
     SUnrecognizedGIFExt : Result := RxResConst_Cze.SUnrecognizedGIFExt;
     SWrongGIFColors : Result := RxResConst_Cze.SWrongGIFColors;
     SBadGIFCodeSize : Result := RxResConst_Cze.SBadGIFCodeSize;
     SGIFDecodeError : Result := RxResConst_Cze.SGIFDecodeError;
     SGIFEncodeError : Result := RxResConst_Cze.SGIFEncodeError;
     SGIFVersion : Result := RxResConst_Cze.SGIFVersion;
     SDualListSrcCaption : Result := RxResConst_Cze.SDualListSrcCaption;
     SDualListDestCaption : Result := RxResConst_Cze.SDualListDestCaption;
     SClipbrdUnknown : Result := RxResConst_Cze.SClipbrdUnknown;
     SClipbrdEmpty : Result := RxResConst_Cze.SClipbrdEmpty;
     SCustomizeSpeedbar : Result := RxResConst_Cze.SCustomizeSpeedbar;
     SAvailButtons : Result := RxResConst_Cze.SAvailButtons;
     SSpeedbarCategories : Result := RxResConst_Cze.SSpeedbarCategories;
     SSpeedbarEditHint : Result := RxResConst_Cze.SSpeedbarEditHint;
     SParseSyntaxError : Result := RxResConst_Cze.SParseSyntaxError;
     SParseNotCramp : Result := RxResConst_Cze.SParseNotCramp;
     SParseDivideByZero : Result := RxResConst_Cze.SParseDivideByZero;
     SParseSqrError : Result := RxResConst_Cze.SParseSqrError;
     SParseLogError : Result := RxResConst_Cze.SParseLogError;
     SParseInvalidFloatOperation : Result := RxResConst_Cze.SParseInvalidFloatOperation;
     srRXControls : Result := RxResConst_Cze.srRXControls;
     srRXDBAware : Result := RxResConst_Cze.srRXDBAware;
     srRXTools : Result := RxResConst_Cze.srRXTools;
     srStorageDesigner : Result := RxResConst_Cze.srStorageDesigner;
     srProxyEditor : Result := RxResConst_Cze.srProxyEditor;
     srPageProxies : Result := RxResConst_Cze.srPageProxies;
     srProxyName : Result := RxResConst_Cze.srProxyName;
     srPageName : Result := RxResConst_Cze.srPageName;
     srSBItemNotCreate : Result := RxResConst_Cze.srSBItemNotCreate;
     srConfirmSBDelete : Result := RxResConst_Cze.srConfirmSBDelete;
     srSpeedbarDesigner : Result := RxResConst_Cze.srSpeedbarDesigner;
     srNewSectionName : Result := RxResConst_Cze.srNewSectionName;
     srEventNotCreate : Result := RxResConst_Cze.srEventNotCreate;
     srTimerDesigner : Result := RxResConst_Cze.srTimerDesigner;
     srTimerEvents : Result := RxResConst_Cze.srTimerEvents;
     srAniCurFilter : Result := RxResConst_Cze.srAniCurFilter;
     srEditPicture : Result := RxResConst_Cze.srEditPicture;
     srLoadAniCursor : Result := RxResConst_Cze.srLoadAniCursor;
     srLoadIcon : Result := RxResConst_Cze.srLoadIcon;
     srSaveImageList : Result := RxResConst_Cze.srSaveImageList;
     srCaptionDesigner : Result := RxResConst_Cze.srCaptionDesigner;
     srGradientCaptions : Result := RxResConst_Cze.srGradientCaptions;
     srBorrowStructure : Result := RxResConst_Cze.srBorrowStructure;
     ccTimeEditHours : Result := RxResConst_Cze.ccTimeEditHours;
     ccTimeEditMins : Result := RxResConst_Cze.ccTimeEditMins;
     ccTimeEditSecs : Result := RxResConst_Cze.ccTimeEditSecs;
     SColorBlack : Result := RxResConst_Cze.SColorBlack;
     SColorMaroon : Result := RxResConst_Cze.SColorMaroon;
     SColorGreen : Result := RxResConst_Cze.SColorGreen;
     SColorOlive : Result := RxResConst_Cze.SColorOlive;
     SColorNavy : Result := RxResConst_Cze.SColorNavy;
     SColorPurple : Result := RxResConst_Cze.SColorPurple;
     SColorTeal : Result := RxResConst_Cze.SColorTeal;
     SColorGray : Result := RxResConst_Cze.SColorGray;
     SColorSilver : Result := RxResConst_Cze.SColorSilver;
     SColorRed : Result := RxResConst_Cze.SColorRed;
     SColorLime : Result := RxResConst_Cze.SColorLime;
     SColorYellow : Result := RxResConst_Cze.SColorYellow;
     SColorBlue : Result := RxResConst_Cze.SColorBlue;
     SColorFuchsia : Result := RxResConst_Cze.SColorFuchsia;
     SColorAqua : Result := RxResConst_Cze.SColorAqua;
     SColorWhite : Result := RxResConst_Cze.SColorWhite;
     SColorScrollBar : Result := RxResConst_Cze.SColorScrollBar;
     SColorBackground : Result := RxResConst_Cze.SColorBackground;
     SColorActiveCaption : Result := RxResConst_Cze.SColorActiveCaption;
     SColorInactiveCaption : Result := RxResConst_Cze.SColorInactiveCaption;
     SColorMenu : Result := RxResConst_Cze.SColorMenu;
     SColorWindow : Result := RxResConst_Cze.SColorWindow;
     SColorWindowFrame : Result := RxResConst_Cze.SColorWindowFrame;
     SColorMenuText : Result := RxResConst_Cze.SColorMenuText;
     SColorWindowText : Result := RxResConst_Cze.SColorWindowText;
     SColorCaptionText : Result := RxResConst_Cze.SColorCaptionText;
     SColorActiveBorder : Result := RxResConst_Cze.SColorActiveBorder;
     SColorInactiveBorder : Result := RxResConst_Cze.SColorInactiveBorder;
     SColorAppWorkSpace : Result := RxResConst_Cze.SColorAppWorkSpace;
     SColorHighlight : Result := RxResConst_Cze.SColorHighlight;
     SColorHighlightText : Result := RxResConst_Cze.SColorHighlightText;
     SColorBtnFace : Result := RxResConst_Cze.SColorBtnFace;
     SColorBtnShadow : Result := RxResConst_Cze.SColorBtnShadow;
     SColorGrayText : Result := RxResConst_Cze.SColorGrayText;
     SColorBtnText : Result := RxResConst_Cze.SColorBtnText;
     SColorInactiveCaptionText : Result := RxResConst_Cze.SColorInactiveCaptionText;
     SColorBtnHighlight : Result := RxResConst_Cze.SColorBtnHighlight;
     SColor3DDkShadow : Result := RxResConst_Cze.SColor3DDkShadow;
     SColor3DLight : Result := RxResConst_Cze.SColor3DLight;
     SColorInfoText : Result := RxResConst_Cze.SColorInfoText;
     SColorInfoBk : Result := RxResConst_Cze.SColorInfoBk;
     SColorCream : Result := RxResConst_Cze.SColorCream;
     SColorMoneyGreen : Result := RxResConst_Cze.SColorMoneyGreen;
     SColorSkyBlue : Result := RxResConst_Cze.SColorSkyBlue;
     SColorNone : Result := RxResConst_Cze.SColorNone;
     SColorDefault : Result := RxResConst_Cze.SColorDefault;
     SColorCustom : Result := RxResConst_Cze.SColorCustom;
     SMemNoRecords : Result := RxResConst_Cze.SMemNoRecords;
     SInvalidFields : Result := RxResConst_Cze.SInvalidFields;
     SAutoSpeedbarMode : Result := RxResConst_Cze.SAutoSpeedbarMode;
    end;
  end;
  {$ENDIF}
  {$IFDEF RXLANG_Dan}
  function RxResConst_Dan_RxLoadStr(IDMsg: Integer): string;
  begin
    case IDMsg of
     SBrowse : Result := RxResConst_Dan.SBrowse;
     SDefaultFilter : Result := RxResConst_Dan.SDefaultFilter;
     SDateDlgTitle : Result := RxResConst_Dan.SDateDlgTitle;
     SNextYear : Result := RxResConst_Dan.SNextYear;
     SNextMonth : Result := RxResConst_Dan.SNextMonth;
     SPrevYear : Result := RxResConst_Dan.SPrevYear;
     SPrevMonth : Result := RxResConst_Dan.SPrevMonth;
     SNotImplemented : Result := RxResConst_Dan.SNotImplemented;
     SFileNotExec : Result := RxResConst_Dan.SFileNotExec;
     SLoadLibError : Result := RxResConst_Dan.SLoadLibError;
     SDetails : Result := RxResConst_Dan.SDetails;
     SDateOutOfRange : Result := RxResConst_Dan.SDateOutOfRange;
     SDateOutOfMin : Result := RxResConst_Dan.SDateOutOfMin;
     SDateOutOfMax : Result := RxResConst_Dan.SDateOutOfMax;
     SDateMinLimit : Result := RxResConst_Dan.SDateMinLimit;
     SDateMaxLimit : Result := RxResConst_Dan.SDateMaxLimit;
     SLocalDatabase : Result := RxResConst_Dan.SLocalDatabase;
     SRetryLogin : Result := RxResConst_Dan.SRetryLogin;
     SExprNotBoolean : Result := RxResConst_Dan.SExprNotBoolean;
     SExprBadNullTest : Result := RxResConst_Dan.SExprBadNullTest;
     SExprBadField : Result := RxResConst_Dan.SExprBadField;
     SCaptureFilter : Result := RxResConst_Dan.SCaptureFilter;
     SNotCaptureFilter : Result := RxResConst_Dan.SNotCaptureFilter;
     SInactiveData : Result := RxResConst_Dan.SInactiveData;
     SBrowseData : Result := RxResConst_Dan.SBrowseData;
     SEditData : Result := RxResConst_Dan.SEditData;
     SInsertData : Result := RxResConst_Dan.SInsertData;
     SSetKeyData : Result := RxResConst_Dan.SSetKeyData;
     SCalcFieldsData : Result := RxResConst_Dan.SCalcFieldsData;
     SRegistration : Result := RxResConst_Dan.SRegistration;
     SAppTitleLabel : Result := RxResConst_Dan.SAppTitleLabel;
     SHintLabel : Result := RxResConst_Dan.SHintLabel;
     SUserNameLabel : Result := RxResConst_Dan.SUserNameLabel;
     SPasswordLabel : Result := RxResConst_Dan.SPasswordLabel;
     SInvalidUserName : Result := RxResConst_Dan.SInvalidUserName;
     SChangePassword : Result := RxResConst_Dan.SChangePassword;
     SOldPasswordLabel : Result := RxResConst_Dan.SOldPasswordLabel;
     SNewPasswordLabel : Result := RxResConst_Dan.SNewPasswordLabel;
     SConfirmPasswordLabel : Result := RxResConst_Dan.SConfirmPasswordLabel;
     SPasswordChanged : Result := RxResConst_Dan.SPasswordChanged;
     SPasswordNotChanged : Result := RxResConst_Dan.SPasswordNotChanged;
     SPasswordsMismatch : Result := RxResConst_Dan.SPasswordsMismatch;
     SDBExceptCaption : Result := RxResConst_Dan.SDBExceptCaption;
     SBDEErrorLabel : Result := RxResConst_Dan.SBDEErrorLabel;
     SServerErrorLabel : Result := RxResConst_Dan.SServerErrorLabel;
     SErrorMsgLabel : Result := RxResConst_Dan.SErrorMsgLabel;
     SNextButton : Result := RxResConst_Dan.SNextButton;
     SPrevButton : Result := RxResConst_Dan.SPrevButton;
     SExprIncorrect : Result := RxResConst_Dan.SExprIncorrect;
     SExprTermination : Result := RxResConst_Dan.SExprTermination;
     SExprNameError : Result := RxResConst_Dan.SExprNameError;
     SExprStringError : Result := RxResConst_Dan.SExprStringError;
     SExprInvalidChar : Result := RxResConst_Dan.SExprInvalidChar;
     SExprNoRParen : Result := RxResConst_Dan.SExprNoRParen;
     SExprExpected : Result := RxResConst_Dan.SExprExpected;
     SExprBadCompare : Result := RxResConst_Dan.SExprBadCompare;
     SConfirmSave : Result := RxResConst_Dan.SConfirmSave;
     SDatabaseName : Result := RxResConst_Dan.SDatabaseName;
     SUnlockCaption : Result := RxResConst_Dan.SUnlockCaption;
     SUnlockHint : Result := RxResConst_Dan.SUnlockHint;
     SDeleteMultipleRecords : Result := RxResConst_Dan.SDeleteMultipleRecords;
     SGIFImage : Result := RxResConst_Dan.SGIFImage;
     SChangeGIFSize : Result := RxResConst_Dan.SChangeGIFSize;
     SNoGIFData : Result := RxResConst_Dan.SNoGIFData;
     SUnrecognizedGIFExt : Result := RxResConst_Dan.SUnrecognizedGIFExt;
     SWrongGIFColors : Result := RxResConst_Dan.SWrongGIFColors;
     SBadGIFCodeSize : Result := RxResConst_Dan.SBadGIFCodeSize;
     SGIFDecodeError : Result := RxResConst_Dan.SGIFDecodeError;
     SGIFEncodeError : Result := RxResConst_Dan.SGIFEncodeError;
     SGIFVersion : Result := RxResConst_Dan.SGIFVersion;
     SDualListSrcCaption : Result := RxResConst_Dan.SDualListSrcCaption;
     SDualListDestCaption : Result := RxResConst_Dan.SDualListDestCaption;
     SClipbrdUnknown : Result := RxResConst_Dan.SClipbrdUnknown;
     SClipbrdEmpty : Result := RxResConst_Dan.SClipbrdEmpty;
     SCustomizeSpeedbar : Result := RxResConst_Dan.SCustomizeSpeedbar;
     SAvailButtons : Result := RxResConst_Dan.SAvailButtons;
     SSpeedbarCategories : Result := RxResConst_Dan.SSpeedbarCategories;
     SSpeedbarEditHint : Result := RxResConst_Dan.SSpeedbarEditHint;
     SParseSyntaxError : Result := RxResConst_Dan.SParseSyntaxError;
     SParseNotCramp : Result := RxResConst_Dan.SParseNotCramp;
     SParseDivideByZero : Result := RxResConst_Dan.SParseDivideByZero;
     SParseSqrError : Result := RxResConst_Dan.SParseSqrError;
     SParseLogError : Result := RxResConst_Dan.SParseLogError;
     SParseInvalidFloatOperation : Result := RxResConst_Dan.SParseInvalidFloatOperation;
     srRXControls : Result := RxResConst_Dan.srRXControls;
     srRXDBAware : Result := RxResConst_Dan.srRXDBAware;
     srRXTools : Result := RxResConst_Dan.srRXTools;
     srStorageDesigner : Result := RxResConst_Dan.srStorageDesigner;
     srProxyEditor : Result := RxResConst_Dan.srProxyEditor;
     srPageProxies : Result := RxResConst_Dan.srPageProxies;
     srProxyName : Result := RxResConst_Dan.srProxyName;
     srPageName : Result := RxResConst_Dan.srPageName;
     srSBItemNotCreate : Result := RxResConst_Dan.srSBItemNotCreate;
     srConfirmSBDelete : Result := RxResConst_Dan.srConfirmSBDelete;
     srSpeedbarDesigner : Result := RxResConst_Dan.srSpeedbarDesigner;
     srNewSectionName : Result := RxResConst_Dan.srNewSectionName;
     srEventNotCreate : Result := RxResConst_Dan.srEventNotCreate;
     srTimerDesigner : Result := RxResConst_Dan.srTimerDesigner;
     srTimerEvents : Result := RxResConst_Dan.srTimerEvents;
     srAniCurFilter : Result := RxResConst_Dan.srAniCurFilter;
     srEditPicture : Result := RxResConst_Dan.srEditPicture;
     srLoadAniCursor : Result := RxResConst_Dan.srLoadAniCursor;
     srLoadIcon : Result := RxResConst_Dan.srLoadIcon;
     srSaveImageList : Result := RxResConst_Dan.srSaveImageList;
     srCaptionDesigner : Result := RxResConst_Dan.srCaptionDesigner;
     srGradientCaptions : Result := RxResConst_Dan.srGradientCaptions;
     srBorrowStructure : Result := RxResConst_Dan.srBorrowStructure;
     ccTimeEditHours : Result := RxResConst_Dan.ccTimeEditHours;
     ccTimeEditMins : Result := RxResConst_Dan.ccTimeEditMins;
     ccTimeEditSecs : Result := RxResConst_Dan.ccTimeEditSecs;
     SColorBlack : Result := RxResConst_Dan.SColorBlack;
     SColorMaroon : Result := RxResConst_Dan.SColorMaroon;
     SColorGreen : Result := RxResConst_Dan.SColorGreen;
     SColorOlive : Result := RxResConst_Dan.SColorOlive;
     SColorNavy : Result := RxResConst_Dan.SColorNavy;
     SColorPurple : Result := RxResConst_Dan.SColorPurple;
     SColorTeal : Result := RxResConst_Dan.SColorTeal;
     SColorGray : Result := RxResConst_Dan.SColorGray;
     SColorSilver : Result := RxResConst_Dan.SColorSilver;
     SColorRed : Result := RxResConst_Dan.SColorRed;
     SColorLime : Result := RxResConst_Dan.SColorLime;
     SColorYellow : Result := RxResConst_Dan.SColorYellow;
     SColorBlue : Result := RxResConst_Dan.SColorBlue;
     SColorFuchsia : Result := RxResConst_Dan.SColorFuchsia;
     SColorAqua : Result := RxResConst_Dan.SColorAqua;
     SColorWhite : Result := RxResConst_Dan.SColorWhite;
     SColorScrollBar : Result := RxResConst_Dan.SColorScrollBar;
     SColorBackground : Result := RxResConst_Dan.SColorBackground;
     SColorActiveCaption : Result := RxResConst_Dan.SColorActiveCaption;
     SColorInactiveCaption : Result := RxResConst_Dan.SColorInactiveCaption;
     SColorMenu : Result := RxResConst_Dan.SColorMenu;
     SColorWindow : Result := RxResConst_Dan.SColorWindow;
     SColorWindowFrame : Result := RxResConst_Dan.SColorWindowFrame;
     SColorMenuText : Result := RxResConst_Dan.SColorMenuText;
     SColorWindowText : Result := RxResConst_Dan.SColorWindowText;
     SColorCaptionText : Result := RxResConst_Dan.SColorCaptionText;
     SColorActiveBorder : Result := RxResConst_Dan.SColorActiveBorder;
     SColorInactiveBorder : Result := RxResConst_Dan.SColorInactiveBorder;
     SColorAppWorkSpace : Result := RxResConst_Dan.SColorAppWorkSpace;
     SColorHighlight : Result := RxResConst_Dan.SColorHighlight;
     SColorHighlightText : Result := RxResConst_Dan.SColorHighlightText;
     SColorBtnFace : Result := RxResConst_Dan.SColorBtnFace;
     SColorBtnShadow : Result := RxResConst_Dan.SColorBtnShadow;
     SColorGrayText : Result := RxResConst_Dan.SColorGrayText;
     SColorBtnText : Result := RxResConst_Dan.SColorBtnText;
     SColorInactiveCaptionText : Result := RxResConst_Dan.SColorInactiveCaptionText;
     SColorBtnHighlight : Result := RxResConst_Dan.SColorBtnHighlight;
     SColor3DDkShadow : Result := RxResConst_Dan.SColor3DDkShadow;
     SColor3DLight : Result := RxResConst_Dan.SColor3DLight;
     SColorInfoText : Result := RxResConst_Dan.SColorInfoText;
     SColorInfoBk : Result := RxResConst_Dan.SColorInfoBk;
     SColorCream : Result := RxResConst_Dan.SColorCream;
     SColorMoneyGreen : Result := RxResConst_Dan.SColorMoneyGreen;
     SColorSkyBlue : Result := RxResConst_Dan.SColorSkyBlue;
     SColorNone : Result := RxResConst_Dan.SColorNone;
     SColorDefault : Result := RxResConst_Dan.SColorDefault;
     SColorCustom : Result := RxResConst_Dan.SColorCustom;
     SMemNoRecords : Result := RxResConst_Dan.SMemNoRecords;
     SInvalidFields : Result := RxResConst_Dan.SInvalidFields;
     SAutoSpeedbarMode : Result := RxResConst_Dan.SAutoSpeedbarMode;
    end;
  end;
  {$ENDIF}
  {$IFDEF RXLANG_Dut}
  function RxResConst_Dut_RxLoadStr(IDMsg: Integer): string;
  begin
    case IDMsg of
     SBrowse : Result := RxResConst_Dut.SBrowse;
     SDefaultFilter : Result := RxResConst_Dut.SDefaultFilter;
     SDateDlgTitle : Result := RxResConst_Dut.SDateDlgTitle;
     SNextYear : Result := RxResConst_Dut.SNextYear;
     SNextMonth : Result := RxResConst_Dut.SNextMonth;
     SPrevYear : Result := RxResConst_Dut.SPrevYear;
     SPrevMonth : Result := RxResConst_Dut.SPrevMonth;
     SNotImplemented : Result := RxResConst_Dut.SNotImplemented;
     SFileNotExec : Result := RxResConst_Dut.SFileNotExec;
     SLoadLibError : Result := RxResConst_Dut.SLoadLibError;
     SDetails : Result := RxResConst_Dut.SDetails;
     SDateOutOfRange : Result := RxResConst_Dut.SDateOutOfRange;
     SDateOutOfMin : Result := RxResConst_Dut.SDateOutOfMin;
     SDateOutOfMax : Result := RxResConst_Dut.SDateOutOfMax;
     SDateMinLimit : Result := RxResConst_Dut.SDateMinLimit;
     SDateMaxLimit : Result := RxResConst_Dut.SDateMaxLimit;
     SLocalDatabase : Result := RxResConst_Dut.SLocalDatabase;
     SRetryLogin : Result := RxResConst_Dut.SRetryLogin;
     SExprNotBoolean : Result := RxResConst_Dut.SExprNotBoolean;
     SExprBadNullTest : Result := RxResConst_Dut.SExprBadNullTest;
     SExprBadField : Result := RxResConst_Dut.SExprBadField;
     SCaptureFilter : Result := RxResConst_Dut.SCaptureFilter;
     SNotCaptureFilter : Result := RxResConst_Dut.SNotCaptureFilter;
     SInactiveData : Result := RxResConst_Dut.SInactiveData;
     SBrowseData : Result := RxResConst_Dut.SBrowseData;
     SEditData : Result := RxResConst_Dut.SEditData;
     SInsertData : Result := RxResConst_Dut.SInsertData;
     SSetKeyData : Result := RxResConst_Dut.SSetKeyData;
     SCalcFieldsData : Result := RxResConst_Dut.SCalcFieldsData;
     SRegistration : Result := RxResConst_Dut.SRegistration;
     SAppTitleLabel : Result := RxResConst_Dut.SAppTitleLabel;
     SHintLabel : Result := RxResConst_Dut.SHintLabel;
     SUserNameLabel : Result := RxResConst_Dut.SUserNameLabel;
     SPasswordLabel : Result := RxResConst_Dut.SPasswordLabel;
     SInvalidUserName : Result := RxResConst_Dut.SInvalidUserName;
     SChangePassword : Result := RxResConst_Dut.SChangePassword;
     SOldPasswordLabel : Result := RxResConst_Dut.SOldPasswordLabel;
     SNewPasswordLabel : Result := RxResConst_Dut.SNewPasswordLabel;
     SConfirmPasswordLabel : Result := RxResConst_Dut.SConfirmPasswordLabel;
     SPasswordChanged : Result := RxResConst_Dut.SPasswordChanged;
     SPasswordNotChanged : Result := RxResConst_Dut.SPasswordNotChanged;
     SPasswordsMismatch : Result := RxResConst_Dut.SPasswordsMismatch;
     SDBExceptCaption : Result := RxResConst_Dut.SDBExceptCaption;
     SBDEErrorLabel : Result := RxResConst_Dut.SBDEErrorLabel;
     SServerErrorLabel : Result := RxResConst_Dut.SServerErrorLabel;
     SErrorMsgLabel : Result := RxResConst_Dut.SErrorMsgLabel;
     SNextButton : Result := RxResConst_Dut.SNextButton;
     SPrevButton : Result := RxResConst_Dut.SPrevButton;
     SExprIncorrect : Result := RxResConst_Dut.SExprIncorrect;
     SExprTermination : Result := RxResConst_Dut.SExprTermination;
     SExprNameError : Result := RxResConst_Dut.SExprNameError;
     SExprStringError : Result := RxResConst_Dut.SExprStringError;
     SExprInvalidChar : Result := RxResConst_Dut.SExprInvalidChar;
     SExprNoRParen : Result := RxResConst_Dut.SExprNoRParen;
     SExprExpected : Result := RxResConst_Dut.SExprExpected;
     SExprBadCompare : Result := RxResConst_Dut.SExprBadCompare;
     SConfirmSave : Result := RxResConst_Dut.SConfirmSave;
     SDatabaseName : Result := RxResConst_Dut.SDatabaseName;
     SUnlockCaption : Result := RxResConst_Dut.SUnlockCaption;
     SUnlockHint : Result := RxResConst_Dut.SUnlockHint;
     SDeleteMultipleRecords : Result := RxResConst_Dut.SDeleteMultipleRecords;
     SGIFImage : Result := RxResConst_Dut.SGIFImage;
     SChangeGIFSize : Result := RxResConst_Dut.SChangeGIFSize;
     SNoGIFData : Result := RxResConst_Dut.SNoGIFData;
     SUnrecognizedGIFExt : Result := RxResConst_Dut.SUnrecognizedGIFExt;
     SWrongGIFColors : Result := RxResConst_Dut.SWrongGIFColors;
     SBadGIFCodeSize : Result := RxResConst_Dut.SBadGIFCodeSize;
     SGIFDecodeError : Result := RxResConst_Dut.SGIFDecodeError;
     SGIFEncodeError : Result := RxResConst_Dut.SGIFEncodeError;
     SGIFVersion : Result := RxResConst_Dut.SGIFVersion;
     SDualListSrcCaption : Result := RxResConst_Dut.SDualListSrcCaption;
     SDualListDestCaption : Result := RxResConst_Dut.SDualListDestCaption;
     SClipbrdUnknown : Result := RxResConst_Dut.SClipbrdUnknown;
     SClipbrdEmpty : Result := RxResConst_Dut.SClipbrdEmpty;
     SCustomizeSpeedbar : Result := RxResConst_Dut.SCustomizeSpeedbar;
     SAvailButtons : Result := RxResConst_Dut.SAvailButtons;
     SSpeedbarCategories : Result := RxResConst_Dut.SSpeedbarCategories;
     SSpeedbarEditHint : Result := RxResConst_Dut.SSpeedbarEditHint;
     SParseSyntaxError : Result := RxResConst_Dut.SParseSyntaxError;
     SParseNotCramp : Result := RxResConst_Dut.SParseNotCramp;
     SParseDivideByZero : Result := RxResConst_Dut.SParseDivideByZero;
     SParseSqrError : Result := RxResConst_Dut.SParseSqrError;
     SParseLogError : Result := RxResConst_Dut.SParseLogError;
     SParseInvalidFloatOperation : Result := RxResConst_Dut.SParseInvalidFloatOperation;
     srRXControls : Result := RxResConst_Dut.srRXControls;
     srRXDBAware : Result := RxResConst_Dut.srRXDBAware;
     srRXTools : Result := RxResConst_Dut.srRXTools;
     srStorageDesigner : Result := RxResConst_Dut.srStorageDesigner;
     srProxyEditor : Result := RxResConst_Dut.srProxyEditor;
     srPageProxies : Result := RxResConst_Dut.srPageProxies;
     srProxyName : Result := RxResConst_Dut.srProxyName;
     srPageName : Result := RxResConst_Dut.srPageName;
     srSBItemNotCreate : Result := RxResConst_Dut.srSBItemNotCreate;
     srConfirmSBDelete : Result := RxResConst_Dut.srConfirmSBDelete;
     srSpeedbarDesigner : Result := RxResConst_Dut.srSpeedbarDesigner;
     srNewSectionName : Result := RxResConst_Dut.srNewSectionName;
     srEventNotCreate : Result := RxResConst_Dut.srEventNotCreate;
     srTimerDesigner : Result := RxResConst_Dut.srTimerDesigner;
     srTimerEvents : Result := RxResConst_Dut.srTimerEvents;
     srAniCurFilter : Result := RxResConst_Dut.srAniCurFilter;
     srEditPicture : Result := RxResConst_Dut.srEditPicture;
     srLoadAniCursor : Result := RxResConst_Dut.srLoadAniCursor;
     srLoadIcon : Result := RxResConst_Dut.srLoadIcon;
     srSaveImageList : Result := RxResConst_Dut.srSaveImageList;
     srCaptionDesigner : Result := RxResConst_Dut.srCaptionDesigner;
     srGradientCaptions : Result := RxResConst_Dut.srGradientCaptions;
     srBorrowStructure : Result := RxResConst_Dut.srBorrowStructure;
     ccTimeEditHours : Result := RxResConst_Dut.ccTimeEditHours;
     ccTimeEditMins : Result := RxResConst_Dut.ccTimeEditMins;
     ccTimeEditSecs : Result := RxResConst_Dut.ccTimeEditSecs;
     SColorBlack : Result := RxResConst_Dut.SColorBlack;
     SColorMaroon : Result := RxResConst_Dut.SColorMaroon;
     SColorGreen : Result := RxResConst_Dut.SColorGreen;
     SColorOlive : Result := RxResConst_Dut.SColorOlive;
     SColorNavy : Result := RxResConst_Dut.SColorNavy;
     SColorPurple : Result := RxResConst_Dut.SColorPurple;
     SColorTeal : Result := RxResConst_Dut.SColorTeal;
     SColorGray : Result := RxResConst_Dut.SColorGray;
     SColorSilver : Result := RxResConst_Dut.SColorSilver;
     SColorRed : Result := RxResConst_Dut.SColorRed;
     SColorLime : Result := RxResConst_Dut.SColorLime;
     SColorYellow : Result := RxResConst_Dut.SColorYellow;
     SColorBlue : Result := RxResConst_Dut.SColorBlue;
     SColorFuchsia : Result := RxResConst_Dut.SColorFuchsia;
     SColorAqua : Result := RxResConst_Dut.SColorAqua;
     SColorWhite : Result := RxResConst_Dut.SColorWhite;
     SColorScrollBar : Result := RxResConst_Dut.SColorScrollBar;
     SColorBackground : Result := RxResConst_Dut.SColorBackground;
     SColorActiveCaption : Result := RxResConst_Dut.SColorActiveCaption;
     SColorInactiveCaption : Result := RxResConst_Dut.SColorInactiveCaption;
     SColorMenu : Result := RxResConst_Dut.SColorMenu;
     SColorWindow : Result := RxResConst_Dut.SColorWindow;
     SColorWindowFrame : Result := RxResConst_Dut.SColorWindowFrame;
     SColorMenuText : Result := RxResConst_Dut.SColorMenuText;
     SColorWindowText : Result := RxResConst_Dut.SColorWindowText;
     SColorCaptionText : Result := RxResConst_Dut.SColorCaptionText;
     SColorActiveBorder : Result := RxResConst_Dut.SColorActiveBorder;
     SColorInactiveBorder : Result := RxResConst_Dut.SColorInactiveBorder;
     SColorAppWorkSpace : Result := RxResConst_Dut.SColorAppWorkSpace;
     SColorHighlight : Result := RxResConst_Dut.SColorHighlight;
     SColorHighlightText : Result := RxResConst_Dut.SColorHighlightText;
     SColorBtnFace : Result := RxResConst_Dut.SColorBtnFace;
     SColorBtnShadow : Result := RxResConst_Dut.SColorBtnShadow;
     SColorGrayText : Result := RxResConst_Dut.SColorGrayText;
     SColorBtnText : Result := RxResConst_Dut.SColorBtnText;
     SColorInactiveCaptionText : Result := RxResConst_Dut.SColorInactiveCaptionText;
     SColorBtnHighlight : Result := RxResConst_Dut.SColorBtnHighlight;
     SColor3DDkShadow : Result := RxResConst_Dut.SColor3DDkShadow;
     SColor3DLight : Result := RxResConst_Dut.SColor3DLight;
     SColorInfoText : Result := RxResConst_Dut.SColorInfoText;
     SColorInfoBk : Result := RxResConst_Dut.SColorInfoBk;
     SColorCream : Result := RxResConst_Dut.SColorCream;
     SColorMoneyGreen : Result := RxResConst_Dut.SColorMoneyGreen;
     SColorSkyBlue : Result := RxResConst_Dut.SColorSkyBlue;
     SColorNone : Result := RxResConst_Dut.SColorNone;
     SColorDefault : Result := RxResConst_Dut.SColorDefault;
     SColorCustom : Result := RxResConst_Dut.SColorCustom;
     SMemNoRecords : Result := RxResConst_Dut.SMemNoRecords;
     SInvalidFields : Result := RxResConst_Dut.SInvalidFields;
     SAutoSpeedbarMode : Result := RxResConst_Dut.SAutoSpeedbarMode;
    end;
  end;
  {$ENDIF}
  {$IFDEF RXLANG_Eng}
  function RxResConst_Eng_RxLoadStr(IDMsg: Integer): string;
  begin
    case IDMsg of
     SBrowse : Result := RxResConst_Eng.SBrowse;
     SDefaultFilter : Result := RxResConst_Eng.SDefaultFilter;
     SDateDlgTitle : Result := RxResConst_Eng.SDateDlgTitle;
     SNextYear : Result := RxResConst_Eng.SNextYear;
     SNextMonth : Result := RxResConst_Eng.SNextMonth;
     SPrevYear : Result := RxResConst_Eng.SPrevYear;
     SPrevMonth : Result := RxResConst_Eng.SPrevMonth;
     SNotImplemented : Result := RxResConst_Eng.SNotImplemented;
     SFileNotExec : Result := RxResConst_Eng.SFileNotExec;
     SLoadLibError : Result := RxResConst_Eng.SLoadLibError;
     SDetails : Result := RxResConst_Eng.SDetails;
     SDateOutOfRange : Result := RxResConst_Eng.SDateOutOfRange;
     SDateOutOfMin : Result := RxResConst_Eng.SDateOutOfMin;
     SDateOutOfMax : Result := RxResConst_Eng.SDateOutOfMax;
     SDateMinLimit : Result := RxResConst_Eng.SDateMinLimit;
     SDateMaxLimit : Result := RxResConst_Eng.SDateMaxLimit;
     SLocalDatabase : Result := RxResConst_Eng.SLocalDatabase;
     SRetryLogin : Result := RxResConst_Eng.SRetryLogin;
     SExprNotBoolean : Result := RxResConst_Eng.SExprNotBoolean;
     SExprBadNullTest : Result := RxResConst_Eng.SExprBadNullTest;
     SExprBadField : Result := RxResConst_Eng.SExprBadField;
     SCaptureFilter : Result := RxResConst_Eng.SCaptureFilter;
     SNotCaptureFilter : Result := RxResConst_Eng.SNotCaptureFilter;
     SInactiveData : Result := RxResConst_Eng.SInactiveData;
     SBrowseData : Result := RxResConst_Eng.SBrowseData;
     SEditData : Result := RxResConst_Eng.SEditData;
     SInsertData : Result := RxResConst_Eng.SInsertData;
     SSetKeyData : Result := RxResConst_Eng.SSetKeyData;
     SCalcFieldsData : Result := RxResConst_Eng.SCalcFieldsData;
     SRegistration : Result := RxResConst_Eng.SRegistration;
     SAppTitleLabel : Result := RxResConst_Eng.SAppTitleLabel;
     SHintLabel : Result := RxResConst_Eng.SHintLabel;
     SUserNameLabel : Result := RxResConst_Eng.SUserNameLabel;
     SPasswordLabel : Result := RxResConst_Eng.SPasswordLabel;
     SInvalidUserName : Result := RxResConst_Eng.SInvalidUserName;
     SChangePassword : Result := RxResConst_Eng.SChangePassword;
     SOldPasswordLabel : Result := RxResConst_Eng.SOldPasswordLabel;
     SNewPasswordLabel : Result := RxResConst_Eng.SNewPasswordLabel;
     SConfirmPasswordLabel : Result := RxResConst_Eng.SConfirmPasswordLabel;
     SPasswordChanged : Result := RxResConst_Eng.SPasswordChanged;
     SPasswordNotChanged : Result := RxResConst_Eng.SPasswordNotChanged;
     SPasswordsMismatch : Result := RxResConst_Eng.SPasswordsMismatch;
     SDBExceptCaption : Result := RxResConst_Eng.SDBExceptCaption;
     SBDEErrorLabel : Result := RxResConst_Eng.SBDEErrorLabel;
     SServerErrorLabel : Result := RxResConst_Eng.SServerErrorLabel;
     SErrorMsgLabel : Result := RxResConst_Eng.SErrorMsgLabel;
     SNextButton : Result := RxResConst_Eng.SNextButton;
     SPrevButton : Result := RxResConst_Eng.SPrevButton;
     SExprIncorrect : Result := RxResConst_Eng.SExprIncorrect;
     SExprTermination : Result := RxResConst_Eng.SExprTermination;
     SExprNameError : Result := RxResConst_Eng.SExprNameError;
     SExprStringError : Result := RxResConst_Eng.SExprStringError;
     SExprInvalidChar : Result := RxResConst_Eng.SExprInvalidChar;
     SExprNoRParen : Result := RxResConst_Eng.SExprNoRParen;
     SExprExpected : Result := RxResConst_Eng.SExprExpected;
     SExprBadCompare : Result := RxResConst_Eng.SExprBadCompare;
     SConfirmSave : Result := RxResConst_Eng.SConfirmSave;
     SDatabaseName : Result := RxResConst_Eng.SDatabaseName;
     SUnlockCaption : Result := RxResConst_Eng.SUnlockCaption;
     SUnlockHint : Result := RxResConst_Eng.SUnlockHint;
     SDeleteMultipleRecords : Result := RxResConst_Eng.SDeleteMultipleRecords;
     SGIFImage : Result := RxResConst_Eng.SGIFImage;
     SChangeGIFSize : Result := RxResConst_Eng.SChangeGIFSize;
     SNoGIFData : Result := RxResConst_Eng.SNoGIFData;
     SUnrecognizedGIFExt : Result := RxResConst_Eng.SUnrecognizedGIFExt;
     SWrongGIFColors : Result := RxResConst_Eng.SWrongGIFColors;
     SBadGIFCodeSize : Result := RxResConst_Eng.SBadGIFCodeSize;
     SGIFDecodeError : Result := RxResConst_Eng.SGIFDecodeError;
     SGIFEncodeError : Result := RxResConst_Eng.SGIFEncodeError;
     SGIFVersion : Result := RxResConst_Eng.SGIFVersion;
     SDualListSrcCaption : Result := RxResConst_Eng.SDualListSrcCaption;
     SDualListDestCaption : Result := RxResConst_Eng.SDualListDestCaption;
     SClipbrdUnknown : Result := RxResConst_Eng.SClipbrdUnknown;
     SClipbrdEmpty : Result := RxResConst_Eng.SClipbrdEmpty;
     SCustomizeSpeedbar : Result := RxResConst_Eng.SCustomizeSpeedbar;
     SAvailButtons : Result := RxResConst_Eng.SAvailButtons;
     SSpeedbarCategories : Result := RxResConst_Eng.SSpeedbarCategories;
     SSpeedbarEditHint : Result := RxResConst_Eng.SSpeedbarEditHint;
     SParseSyntaxError : Result := RxResConst_Eng.SParseSyntaxError;
     SParseNotCramp : Result := RxResConst_Eng.SParseNotCramp;
     SParseDivideByZero : Result := RxResConst_Eng.SParseDivideByZero;
     SParseSqrError : Result := RxResConst_Eng.SParseSqrError;
     SParseLogError : Result := RxResConst_Eng.SParseLogError;
     SParseInvalidFloatOperation : Result := RxResConst_Eng.SParseInvalidFloatOperation;
     srRXControls : Result := RxResConst_Eng.srRXControls;
     srRXDBAware : Result := RxResConst_Eng.srRXDBAware;
     srRXTools : Result := RxResConst_Eng.srRXTools;
     srStorageDesigner : Result := RxResConst_Eng.srStorageDesigner;
     srProxyEditor : Result := RxResConst_Eng.srProxyEditor;
     srPageProxies : Result := RxResConst_Eng.srPageProxies;
     srProxyName : Result := RxResConst_Eng.srProxyName;
     srPageName : Result := RxResConst_Eng.srPageName;
     srSBItemNotCreate : Result := RxResConst_Eng.srSBItemNotCreate;
     srConfirmSBDelete : Result := RxResConst_Eng.srConfirmSBDelete;
     srSpeedbarDesigner : Result := RxResConst_Eng.srSpeedbarDesigner;
     srNewSectionName : Result := RxResConst_Eng.srNewSectionName;
     srEventNotCreate : Result := RxResConst_Eng.srEventNotCreate;
     srTimerDesigner : Result := RxResConst_Eng.srTimerDesigner;
     srTimerEvents : Result := RxResConst_Eng.srTimerEvents;
     srAniCurFilter : Result := RxResConst_Eng.srAniCurFilter;
     srEditPicture : Result := RxResConst_Eng.srEditPicture;
     srLoadAniCursor : Result := RxResConst_Eng.srLoadAniCursor;
     srLoadIcon : Result := RxResConst_Eng.srLoadIcon;
     srSaveImageList : Result := RxResConst_Eng.srSaveImageList;
     srCaptionDesigner : Result := RxResConst_Eng.srCaptionDesigner;
     srGradientCaptions : Result := RxResConst_Eng.srGradientCaptions;
     srBorrowStructure : Result := RxResConst_Eng.srBorrowStructure;
     ccTimeEditHours : Result := RxResConst_Eng.ccTimeEditHours;
     ccTimeEditMins : Result := RxResConst_Eng.ccTimeEditMins;
     ccTimeEditSecs : Result := RxResConst_Eng.ccTimeEditSecs;
     SColorBlack : Result := RxResConst_Eng.SColorBlack;
     SColorMaroon : Result := RxResConst_Eng.SColorMaroon;
     SColorGreen : Result := RxResConst_Eng.SColorGreen;
     SColorOlive : Result := RxResConst_Eng.SColorOlive;
     SColorNavy : Result := RxResConst_Eng.SColorNavy;
     SColorPurple : Result := RxResConst_Eng.SColorPurple;
     SColorTeal : Result := RxResConst_Eng.SColorTeal;
     SColorGray : Result := RxResConst_Eng.SColorGray;
     SColorSilver : Result := RxResConst_Eng.SColorSilver;
     SColorRed : Result := RxResConst_Eng.SColorRed;
     SColorLime : Result := RxResConst_Eng.SColorLime;
     SColorYellow : Result := RxResConst_Eng.SColorYellow;
     SColorBlue : Result := RxResConst_Eng.SColorBlue;
     SColorFuchsia : Result := RxResConst_Eng.SColorFuchsia;
     SColorAqua : Result := RxResConst_Eng.SColorAqua;
     SColorWhite : Result := RxResConst_Eng.SColorWhite;
     SColorScrollBar : Result := RxResConst_Eng.SColorScrollBar;
     SColorBackground : Result := RxResConst_Eng.SColorBackground;
     SColorActiveCaption : Result := RxResConst_Eng.SColorActiveCaption;
     SColorInactiveCaption : Result := RxResConst_Eng.SColorInactiveCaption;
     SColorMenu : Result := RxResConst_Eng.SColorMenu;
     SColorWindow : Result := RxResConst_Eng.SColorWindow;
     SColorWindowFrame : Result := RxResConst_Eng.SColorWindowFrame;
     SColorMenuText : Result := RxResConst_Eng.SColorMenuText;
     SColorWindowText : Result := RxResConst_Eng.SColorWindowText;
     SColorCaptionText : Result := RxResConst_Eng.SColorCaptionText;
     SColorActiveBorder : Result := RxResConst_Eng.SColorActiveBorder;
     SColorInactiveBorder : Result := RxResConst_Eng.SColorInactiveBorder;
     SColorAppWorkSpace : Result := RxResConst_Eng.SColorAppWorkSpace;
     SColorHighlight : Result := RxResConst_Eng.SColorHighlight;
     SColorHighlightText : Result := RxResConst_Eng.SColorHighlightText;
     SColorBtnFace : Result := RxResConst_Eng.SColorBtnFace;
     SColorBtnShadow : Result := RxResConst_Eng.SColorBtnShadow;
     SColorGrayText : Result := RxResConst_Eng.SColorGrayText;
     SColorBtnText : Result := RxResConst_Eng.SColorBtnText;
     SColorInactiveCaptionText : Result := RxResConst_Eng.SColorInactiveCaptionText;
     SColorBtnHighlight : Result := RxResConst_Eng.SColorBtnHighlight;
     SColor3DDkShadow : Result := RxResConst_Eng.SColor3DDkShadow;
     SColor3DLight : Result := RxResConst_Eng.SColor3DLight;
     SColorInfoText : Result := RxResConst_Eng.SColorInfoText;
     SColorInfoBk : Result := RxResConst_Eng.SColorInfoBk;
     SColorCream : Result := RxResConst_Eng.SColorCream;
     SColorMoneyGreen : Result := RxResConst_Eng.SColorMoneyGreen;
     SColorSkyBlue : Result := RxResConst_Eng.SColorSkyBlue;
     SColorNone : Result := RxResConst_Eng.SColorNone;
     SColorDefault : Result := RxResConst_Eng.SColorDefault;
     SColorCustom : Result := RxResConst_Eng.SColorCustom;
     SMemNoRecords : Result := RxResConst_Eng.SMemNoRecords;
     SInvalidFields : Result := RxResConst_Eng.SInvalidFields;
     SAutoSpeedbarMode : Result := RxResConst_Eng.SAutoSpeedbarMode;
    end;
  end;
  {$ENDIF}
  {$IFDEF RXLANG_Fra}
  function RxResConst_Fra_RxLoadStr(IDMsg: Integer): string;
  begin
    case IDMsg of
     SBrowse : Result := RxResConst_Fra.SBrowse;
     SDefaultFilter : Result := RxResConst_Fra.SDefaultFilter;
     SDateDlgTitle : Result := RxResConst_Fra.SDateDlgTitle;
     SNextYear : Result := RxResConst_Fra.SNextYear;
     SNextMonth : Result := RxResConst_Fra.SNextMonth;
     SPrevYear : Result := RxResConst_Fra.SPrevYear;
     SPrevMonth : Result := RxResConst_Fra.SPrevMonth;
     SNotImplemented : Result := RxResConst_Fra.SNotImplemented;
     SFileNotExec : Result := RxResConst_Fra.SFileNotExec;
     SLoadLibError : Result := RxResConst_Fra.SLoadLibError;
     SDetails : Result := RxResConst_Fra.SDetails;
     SDateOutOfRange : Result := RxResConst_Fra.SDateOutOfRange;
     SDateOutOfMin : Result := RxResConst_Fra.SDateOutOfMin;
     SDateOutOfMax : Result := RxResConst_Fra.SDateOutOfMax;
     SDateMinLimit : Result := RxResConst_Fra.SDateMinLimit;
     SDateMaxLimit : Result := RxResConst_Fra.SDateMaxLimit;
     SLocalDatabase : Result := RxResConst_Fra.SLocalDatabase;
     SRetryLogin : Result := RxResConst_Fra.SRetryLogin;
     SExprNotBoolean : Result := RxResConst_Fra.SExprNotBoolean;
     SExprBadNullTest : Result := RxResConst_Fra.SExprBadNullTest;
     SExprBadField : Result := RxResConst_Fra.SExprBadField;
     SCaptureFilter : Result := RxResConst_Fra.SCaptureFilter;
     SNotCaptureFilter : Result := RxResConst_Fra.SNotCaptureFilter;
     SInactiveData : Result := RxResConst_Fra.SInactiveData;
     SBrowseData : Result := RxResConst_Fra.SBrowseData;
     SEditData : Result := RxResConst_Fra.SEditData;
     SInsertData : Result := RxResConst_Fra.SInsertData;
     SSetKeyData : Result := RxResConst_Fra.SSetKeyData;
     SCalcFieldsData : Result := RxResConst_Fra.SCalcFieldsData;
     SRegistration : Result := RxResConst_Fra.SRegistration;
     SAppTitleLabel : Result := RxResConst_Fra.SAppTitleLabel;
     SHintLabel : Result := RxResConst_Fra.SHintLabel;
     SUserNameLabel : Result := RxResConst_Fra.SUserNameLabel;
     SPasswordLabel : Result := RxResConst_Fra.SPasswordLabel;
     SInvalidUserName : Result := RxResConst_Fra.SInvalidUserName;
     SChangePassword : Result := RxResConst_Fra.SChangePassword;
     SOldPasswordLabel : Result := RxResConst_Fra.SOldPasswordLabel;
     SNewPasswordLabel : Result := RxResConst_Fra.SNewPasswordLabel;
     SConfirmPasswordLabel : Result := RxResConst_Fra.SConfirmPasswordLabel;
     SPasswordChanged : Result := RxResConst_Fra.SPasswordChanged;
     SPasswordNotChanged : Result := RxResConst_Fra.SPasswordNotChanged;
     SPasswordsMismatch : Result := RxResConst_Fra.SPasswordsMismatch;
     SDBExceptCaption : Result := RxResConst_Fra.SDBExceptCaption;
     SBDEErrorLabel : Result := RxResConst_Fra.SBDEErrorLabel;
     SServerErrorLabel : Result := RxResConst_Fra.SServerErrorLabel;
     SErrorMsgLabel : Result := RxResConst_Fra.SErrorMsgLabel;
     SNextButton : Result := RxResConst_Fra.SNextButton;
     SPrevButton : Result := RxResConst_Fra.SPrevButton;
     SExprIncorrect : Result := RxResConst_Fra.SExprIncorrect;
     SExprTermination : Result := RxResConst_Fra.SExprTermination;
     SExprNameError : Result := RxResConst_Fra.SExprNameError;
     SExprStringError : Result := RxResConst_Fra.SExprStringError;
     SExprInvalidChar : Result := RxResConst_Fra.SExprInvalidChar;
     SExprNoRParen : Result := RxResConst_Fra.SExprNoRParen;
     SExprExpected : Result := RxResConst_Fra.SExprExpected;
     SExprBadCompare : Result := RxResConst_Fra.SExprBadCompare;
     SConfirmSave : Result := RxResConst_Fra.SConfirmSave;
     SDatabaseName : Result := RxResConst_Fra.SDatabaseName;
     SUnlockCaption : Result := RxResConst_Fra.SUnlockCaption;
     SUnlockHint : Result := RxResConst_Fra.SUnlockHint;
     SDeleteMultipleRecords : Result := RxResConst_Fra.SDeleteMultipleRecords;
     SGIFImage : Result := RxResConst_Fra.SGIFImage;
     SChangeGIFSize : Result := RxResConst_Fra.SChangeGIFSize;
     SNoGIFData : Result := RxResConst_Fra.SNoGIFData;
     SUnrecognizedGIFExt : Result := RxResConst_Fra.SUnrecognizedGIFExt;
     SWrongGIFColors : Result := RxResConst_Fra.SWrongGIFColors;
     SBadGIFCodeSize : Result := RxResConst_Fra.SBadGIFCodeSize;
     SGIFDecodeError : Result := RxResConst_Fra.SGIFDecodeError;
     SGIFEncodeError : Result := RxResConst_Fra.SGIFEncodeError;
     SGIFVersion : Result := RxResConst_Fra.SGIFVersion;
     SDualListSrcCaption : Result := RxResConst_Fra.SDualListSrcCaption;
     SDualListDestCaption : Result := RxResConst_Fra.SDualListDestCaption;
     SClipbrdUnknown : Result := RxResConst_Fra.SClipbrdUnknown;
     SClipbrdEmpty : Result := RxResConst_Fra.SClipbrdEmpty;
     SCustomizeSpeedbar : Result := RxResConst_Fra.SCustomizeSpeedbar;
     SAvailButtons : Result := RxResConst_Fra.SAvailButtons;
     SSpeedbarCategories : Result := RxResConst_Fra.SSpeedbarCategories;
     SSpeedbarEditHint : Result := RxResConst_Fra.SSpeedbarEditHint;
     SParseSyntaxError : Result := RxResConst_Fra.SParseSyntaxError;
     SParseNotCramp : Result := RxResConst_Fra.SParseNotCramp;
     SParseDivideByZero : Result := RxResConst_Fra.SParseDivideByZero;
     SParseSqrError : Result := RxResConst_Fra.SParseSqrError;
     SParseLogError : Result := RxResConst_Fra.SParseLogError;
     SParseInvalidFloatOperation : Result := RxResConst_Fra.SParseInvalidFloatOperation;
     srRXControls : Result := RxResConst_Fra.srRXControls;
     srRXDBAware : Result := RxResConst_Fra.srRXDBAware;
     srRXTools : Result := RxResConst_Fra.srRXTools;
     srStorageDesigner : Result := RxResConst_Fra.srStorageDesigner;
     srProxyEditor : Result := RxResConst_Fra.srProxyEditor;
     srPageProxies : Result := RxResConst_Fra.srPageProxies;
     srProxyName : Result := RxResConst_Fra.srProxyName;
     srPageName : Result := RxResConst_Fra.srPageName;
     srSBItemNotCreate : Result := RxResConst_Fra.srSBItemNotCreate;
     srConfirmSBDelete : Result := RxResConst_Fra.srConfirmSBDelete;
     srSpeedbarDesigner : Result := RxResConst_Fra.srSpeedbarDesigner;
     srNewSectionName : Result := RxResConst_Fra.srNewSectionName;
     srEventNotCreate : Result := RxResConst_Fra.srEventNotCreate;
     srTimerDesigner : Result := RxResConst_Fra.srTimerDesigner;
     srTimerEvents : Result := RxResConst_Fra.srTimerEvents;
     srAniCurFilter : Result := RxResConst_Fra.srAniCurFilter;
     srEditPicture : Result := RxResConst_Fra.srEditPicture;
     srLoadAniCursor : Result := RxResConst_Fra.srLoadAniCursor;
     srLoadIcon : Result := RxResConst_Fra.srLoadIcon;
     srSaveImageList : Result := RxResConst_Fra.srSaveImageList;
     srCaptionDesigner : Result := RxResConst_Fra.srCaptionDesigner;
     srGradientCaptions : Result := RxResConst_Fra.srGradientCaptions;
     srBorrowStructure : Result := RxResConst_Fra.srBorrowStructure;
     ccTimeEditHours : Result := RxResConst_Fra.ccTimeEditHours;
     ccTimeEditMins : Result := RxResConst_Fra.ccTimeEditMins;
     ccTimeEditSecs : Result := RxResConst_Fra.ccTimeEditSecs;
     SColorBlack : Result := RxResConst_Fra.SColorBlack;
     SColorMaroon : Result := RxResConst_Fra.SColorMaroon;
     SColorGreen : Result := RxResConst_Fra.SColorGreen;
     SColorOlive : Result := RxResConst_Fra.SColorOlive;
     SColorNavy : Result := RxResConst_Fra.SColorNavy;
     SColorPurple : Result := RxResConst_Fra.SColorPurple;
     SColorTeal : Result := RxResConst_Fra.SColorTeal;
     SColorGray : Result := RxResConst_Fra.SColorGray;
     SColorSilver : Result := RxResConst_Fra.SColorSilver;
     SColorRed : Result := RxResConst_Fra.SColorRed;
     SColorLime : Result := RxResConst_Fra.SColorLime;
     SColorYellow : Result := RxResConst_Fra.SColorYellow;
     SColorBlue : Result := RxResConst_Fra.SColorBlue;
     SColorFuchsia : Result := RxResConst_Fra.SColorFuchsia;
     SColorAqua : Result := RxResConst_Fra.SColorAqua;
     SColorWhite : Result := RxResConst_Fra.SColorWhite;
     SColorScrollBar : Result := RxResConst_Fra.SColorScrollBar;
     SColorBackground : Result := RxResConst_Fra.SColorBackground;
     SColorActiveCaption : Result := RxResConst_Fra.SColorActiveCaption;
     SColorInactiveCaption : Result := RxResConst_Fra.SColorInactiveCaption;
     SColorMenu : Result := RxResConst_Fra.SColorMenu;
     SColorWindow : Result := RxResConst_Fra.SColorWindow;
     SColorWindowFrame : Result := RxResConst_Fra.SColorWindowFrame;
     SColorMenuText : Result := RxResConst_Fra.SColorMenuText;
     SColorWindowText : Result := RxResConst_Fra.SColorWindowText;
     SColorCaptionText : Result := RxResConst_Fra.SColorCaptionText;
     SColorActiveBorder : Result := RxResConst_Fra.SColorActiveBorder;
     SColorInactiveBorder : Result := RxResConst_Fra.SColorInactiveBorder;
     SColorAppWorkSpace : Result := RxResConst_Fra.SColorAppWorkSpace;
     SColorHighlight : Result := RxResConst_Fra.SColorHighlight;
     SColorHighlightText : Result := RxResConst_Fra.SColorHighlightText;
     SColorBtnFace : Result := RxResConst_Fra.SColorBtnFace;
     SColorBtnShadow : Result := RxResConst_Fra.SColorBtnShadow;
     SColorGrayText : Result := RxResConst_Fra.SColorGrayText;
     SColorBtnText : Result := RxResConst_Fra.SColorBtnText;
     SColorInactiveCaptionText : Result := RxResConst_Fra.SColorInactiveCaptionText;
     SColorBtnHighlight : Result := RxResConst_Fra.SColorBtnHighlight;
     SColor3DDkShadow : Result := RxResConst_Fra.SColor3DDkShadow;
     SColor3DLight : Result := RxResConst_Fra.SColor3DLight;
     SColorInfoText : Result := RxResConst_Fra.SColorInfoText;
     SColorInfoBk : Result := RxResConst_Fra.SColorInfoBk;
     SColorCream : Result := RxResConst_Fra.SColorCream;
     SColorMoneyGreen : Result := RxResConst_Fra.SColorMoneyGreen;
     SColorSkyBlue : Result := RxResConst_Fra.SColorSkyBlue;
     SColorNone : Result := RxResConst_Fra.SColorNone;
     SColorDefault : Result := RxResConst_Fra.SColorDefault;
     SColorCustom : Result := RxResConst_Fra.SColorCustom;
     SMemNoRecords : Result := RxResConst_Fra.SMemNoRecords;
     SInvalidFields : Result := RxResConst_Fra.SInvalidFields;
     SAutoSpeedbarMode : Result := RxResConst_Fra.SAutoSpeedbarMode;
    end;
  end;
  {$ENDIF}
  {$IFDEF RXLANG_Gal}
  function RxResConst_Gal_RxLoadStr(IDMsg: Integer): string;
  begin
    case IDMsg of
     SBrowse : Result := RxResConst_Gal.SBrowse;
     SDefaultFilter : Result := RxResConst_Gal.SDefaultFilter;
     SDateDlgTitle : Result := RxResConst_Gal.SDateDlgTitle;
     SNextYear : Result := RxResConst_Gal.SNextYear;
     SNextMonth : Result := RxResConst_Gal.SNextMonth;
     SPrevYear : Result := RxResConst_Gal.SPrevYear;
     SPrevMonth : Result := RxResConst_Gal.SPrevMonth;
     SNotImplemented : Result := RxResConst_Gal.SNotImplemented;
     SFileNotExec : Result := RxResConst_Gal.SFileNotExec;
     SLoadLibError : Result := RxResConst_Gal.SLoadLibError;
     SDetails : Result := RxResConst_Gal.SDetails;
     SDateOutOfRange : Result := RxResConst_Gal.SDateOutOfRange;
     SDateOutOfMin : Result := RxResConst_Gal.SDateOutOfMin;
     SDateOutOfMax : Result := RxResConst_Gal.SDateOutOfMax;
     SDateMinLimit : Result := RxResConst_Gal.SDateMinLimit;
     SDateMaxLimit : Result := RxResConst_Gal.SDateMaxLimit;
     SLocalDatabase : Result := RxResConst_Gal.SLocalDatabase;
     SRetryLogin : Result := RxResConst_Gal.SRetryLogin;
     SExprNotBoolean : Result := RxResConst_Gal.SExprNotBoolean;
     SExprBadNullTest : Result := RxResConst_Gal.SExprBadNullTest;
     SExprBadField : Result := RxResConst_Gal.SExprBadField;
     SCaptureFilter : Result := RxResConst_Gal.SCaptureFilter;
     SNotCaptureFilter : Result := RxResConst_Gal.SNotCaptureFilter;
     SInactiveData : Result := RxResConst_Gal.SInactiveData;
     SBrowseData : Result := RxResConst_Gal.SBrowseData;
     SEditData : Result := RxResConst_Gal.SEditData;
     SInsertData : Result := RxResConst_Gal.SInsertData;
     SSetKeyData : Result := RxResConst_Gal.SSetKeyData;
     SCalcFieldsData : Result := RxResConst_Gal.SCalcFieldsData;
     SRegistration : Result := RxResConst_Gal.SRegistration;
     SAppTitleLabel : Result := RxResConst_Gal.SAppTitleLabel;
     SHintLabel : Result := RxResConst_Gal.SHintLabel;
     SUserNameLabel : Result := RxResConst_Gal.SUserNameLabel;
     SPasswordLabel : Result := RxResConst_Gal.SPasswordLabel;
     SInvalidUserName : Result := RxResConst_Gal.SInvalidUserName;
     SChangePassword : Result := RxResConst_Gal.SChangePassword;
     SOldPasswordLabel : Result := RxResConst_Gal.SOldPasswordLabel;
     SNewPasswordLabel : Result := RxResConst_Gal.SNewPasswordLabel;
     SConfirmPasswordLabel : Result := RxResConst_Gal.SConfirmPasswordLabel;
     SPasswordChanged : Result := RxResConst_Gal.SPasswordChanged;
     SPasswordNotChanged : Result := RxResConst_Gal.SPasswordNotChanged;
     SPasswordsMismatch : Result := RxResConst_Gal.SPasswordsMismatch;
     SDBExceptCaption : Result := RxResConst_Gal.SDBExceptCaption;
     SBDEErrorLabel : Result := RxResConst_Gal.SBDEErrorLabel;
     SServerErrorLabel : Result := RxResConst_Gal.SServerErrorLabel;
     SErrorMsgLabel : Result := RxResConst_Gal.SErrorMsgLabel;
     SNextButton : Result := RxResConst_Gal.SNextButton;
     SPrevButton : Result := RxResConst_Gal.SPrevButton;
     SExprIncorrect : Result := RxResConst_Gal.SExprIncorrect;
     SExprTermination : Result := RxResConst_Gal.SExprTermination;
     SExprNameError : Result := RxResConst_Gal.SExprNameError;
     SExprStringError : Result := RxResConst_Gal.SExprStringError;
     SExprInvalidChar : Result := RxResConst_Gal.SExprInvalidChar;
     SExprNoRParen : Result := RxResConst_Gal.SExprNoRParen;
     SExprExpected : Result := RxResConst_Gal.SExprExpected;
     SExprBadCompare : Result := RxResConst_Gal.SExprBadCompare;
     SConfirmSave : Result := RxResConst_Gal.SConfirmSave;
     SDatabaseName : Result := RxResConst_Gal.SDatabaseName;
     SUnlockCaption : Result := RxResConst_Gal.SUnlockCaption;
     SUnlockHint : Result := RxResConst_Gal.SUnlockHint;
     SDeleteMultipleRecords : Result := RxResConst_Gal.SDeleteMultipleRecords;
     SGIFImage : Result := RxResConst_Gal.SGIFImage;
     SChangeGIFSize : Result := RxResConst_Gal.SChangeGIFSize;
     SNoGIFData : Result := RxResConst_Gal.SNoGIFData;
     SUnrecognizedGIFExt : Result := RxResConst_Gal.SUnrecognizedGIFExt;
     SWrongGIFColors : Result := RxResConst_Gal.SWrongGIFColors;
     SBadGIFCodeSize : Result := RxResConst_Gal.SBadGIFCodeSize;
     SGIFDecodeError : Result := RxResConst_Gal.SGIFDecodeError;
     SGIFEncodeError : Result := RxResConst_Gal.SGIFEncodeError;
     SGIFVersion : Result := RxResConst_Gal.SGIFVersion;
     SDualListSrcCaption : Result := RxResConst_Gal.SDualListSrcCaption;
     SDualListDestCaption : Result := RxResConst_Gal.SDualListDestCaption;
     SClipbrdUnknown : Result := RxResConst_Gal.SClipbrdUnknown;
     SClipbrdEmpty : Result := RxResConst_Gal.SClipbrdEmpty;
     SCustomizeSpeedbar : Result := RxResConst_Gal.SCustomizeSpeedbar;
     SAvailButtons : Result := RxResConst_Gal.SAvailButtons;
     SSpeedbarCategories : Result := RxResConst_Gal.SSpeedbarCategories;
     SSpeedbarEditHint : Result := RxResConst_Gal.SSpeedbarEditHint;
     SParseSyntaxError : Result := RxResConst_Gal.SParseSyntaxError;
     SParseNotCramp : Result := RxResConst_Gal.SParseNotCramp;
     SParseDivideByZero : Result := RxResConst_Gal.SParseDivideByZero;
     SParseSqrError : Result := RxResConst_Gal.SParseSqrError;
     SParseLogError : Result := RxResConst_Gal.SParseLogError;
     SParseInvalidFloatOperation : Result := RxResConst_Gal.SParseInvalidFloatOperation;
     srRXControls : Result := RxResConst_Gal.srRXControls;
     srRXDBAware : Result := RxResConst_Gal.srRXDBAware;
     srRXTools : Result := RxResConst_Gal.srRXTools;
     srStorageDesigner : Result := RxResConst_Gal.srStorageDesigner;
     srProxyEditor : Result := RxResConst_Gal.srProxyEditor;
     srPageProxies : Result := RxResConst_Gal.srPageProxies;
     srProxyName : Result := RxResConst_Gal.srProxyName;
     srPageName : Result := RxResConst_Gal.srPageName;
     srSBItemNotCreate : Result := RxResConst_Gal.srSBItemNotCreate;
     srConfirmSBDelete : Result := RxResConst_Gal.srConfirmSBDelete;
     srSpeedbarDesigner : Result := RxResConst_Gal.srSpeedbarDesigner;
     srNewSectionName : Result := RxResConst_Gal.srNewSectionName;
     srEventNotCreate : Result := RxResConst_Gal.srEventNotCreate;
     srTimerDesigner : Result := RxResConst_Gal.srTimerDesigner;
     srTimerEvents : Result := RxResConst_Gal.srTimerEvents;
     srAniCurFilter : Result := RxResConst_Gal.srAniCurFilter;
     srEditPicture : Result := RxResConst_Gal.srEditPicture;
     srLoadAniCursor : Result := RxResConst_Gal.srLoadAniCursor;
     srLoadIcon : Result := RxResConst_Gal.srLoadIcon;
     srSaveImageList : Result := RxResConst_Gal.srSaveImageList;
     srCaptionDesigner : Result := RxResConst_Gal.srCaptionDesigner;
     srGradientCaptions : Result := RxResConst_Gal.srGradientCaptions;
     srBorrowStructure : Result := RxResConst_Gal.srBorrowStructure;
     ccTimeEditHours : Result := RxResConst_Gal.ccTimeEditHours;
     ccTimeEditMins : Result := RxResConst_Gal.ccTimeEditMins;
     ccTimeEditSecs : Result := RxResConst_Gal.ccTimeEditSecs;
     SColorBlack : Result := RxResConst_Gal.SColorBlack;
     SColorMaroon : Result := RxResConst_Gal.SColorMaroon;
     SColorGreen : Result := RxResConst_Gal.SColorGreen;
     SColorOlive : Result := RxResConst_Gal.SColorOlive;
     SColorNavy : Result := RxResConst_Gal.SColorNavy;
     SColorPurple : Result := RxResConst_Gal.SColorPurple;
     SColorTeal : Result := RxResConst_Gal.SColorTeal;
     SColorGray : Result := RxResConst_Gal.SColorGray;
     SColorSilver : Result := RxResConst_Gal.SColorSilver;
     SColorRed : Result := RxResConst_Gal.SColorRed;
     SColorLime : Result := RxResConst_Gal.SColorLime;
     SColorYellow : Result := RxResConst_Gal.SColorYellow;
     SColorBlue : Result := RxResConst_Gal.SColorBlue;
     SColorFuchsia : Result := RxResConst_Gal.SColorFuchsia;
     SColorAqua : Result := RxResConst_Gal.SColorAqua;
     SColorWhite : Result := RxResConst_Gal.SColorWhite;
     SColorScrollBar : Result := RxResConst_Gal.SColorScrollBar;
     SColorBackground : Result := RxResConst_Gal.SColorBackground;
     SColorActiveCaption : Result := RxResConst_Gal.SColorActiveCaption;
     SColorInactiveCaption : Result := RxResConst_Gal.SColorInactiveCaption;
     SColorMenu : Result := RxResConst_Gal.SColorMenu;
     SColorWindow : Result := RxResConst_Gal.SColorWindow;
     SColorWindowFrame : Result := RxResConst_Gal.SColorWindowFrame;
     SColorMenuText : Result := RxResConst_Gal.SColorMenuText;
     SColorWindowText : Result := RxResConst_Gal.SColorWindowText;
     SColorCaptionText : Result := RxResConst_Gal.SColorCaptionText;
     SColorActiveBorder : Result := RxResConst_Gal.SColorActiveBorder;
     SColorInactiveBorder : Result := RxResConst_Gal.SColorInactiveBorder;
     SColorAppWorkSpace : Result := RxResConst_Gal.SColorAppWorkSpace;
     SColorHighlight : Result := RxResConst_Gal.SColorHighlight;
     SColorHighlightText : Result := RxResConst_Gal.SColorHighlightText;
     SColorBtnFace : Result := RxResConst_Gal.SColorBtnFace;
     SColorBtnShadow : Result := RxResConst_Gal.SColorBtnShadow;
     SColorGrayText : Result := RxResConst_Gal.SColorGrayText;
     SColorBtnText : Result := RxResConst_Gal.SColorBtnText;
     SColorInactiveCaptionText : Result := RxResConst_Gal.SColorInactiveCaptionText;
     SColorBtnHighlight : Result := RxResConst_Gal.SColorBtnHighlight;
     SColor3DDkShadow : Result := RxResConst_Gal.SColor3DDkShadow;
     SColor3DLight : Result := RxResConst_Gal.SColor3DLight;
     SColorInfoText : Result := RxResConst_Gal.SColorInfoText;
     SColorInfoBk : Result := RxResConst_Gal.SColorInfoBk;
     SColorCream : Result := RxResConst_Gal.SColorCream;
     SColorMoneyGreen : Result := RxResConst_Gal.SColorMoneyGreen;
     SColorSkyBlue : Result := RxResConst_Gal.SColorSkyBlue;
     SColorNone : Result := RxResConst_Gal.SColorNone;
     SColorDefault : Result := RxResConst_Gal.SColorDefault;
     SColorCustom : Result := RxResConst_Gal.SColorCustom;
     SMemNoRecords : Result := RxResConst_Gal.SMemNoRecords;
     SInvalidFields : Result := RxResConst_Gal.SInvalidFields;
     SAutoSpeedbarMode : Result := RxResConst_Gal.SAutoSpeedbarMode;
    end;
  end;
  {$ENDIF}
  {$IFDEF RXLANG_Ger}
  function RxResConst_Ger_RxLoadStr(IDMsg: Integer): string;
  begin
    case IDMsg of
     SBrowse : Result := RxResConst_Ger.SBrowse;
     SDefaultFilter : Result := RxResConst_Ger.SDefaultFilter;
     SDateDlgTitle : Result := RxResConst_Ger.SDateDlgTitle;
     SNextYear : Result := RxResConst_Ger.SNextYear;
     SNextMonth : Result := RxResConst_Ger.SNextMonth;
     SPrevYear : Result := RxResConst_Ger.SPrevYear;
     SPrevMonth : Result := RxResConst_Ger.SPrevMonth;
     SNotImplemented : Result := RxResConst_Ger.SNotImplemented;
     SFileNotExec : Result := RxResConst_Ger.SFileNotExec;
     SLoadLibError : Result := RxResConst_Ger.SLoadLibError;
     SDetails : Result := RxResConst_Ger.SDetails;
     SDateOutOfRange : Result := RxResConst_Ger.SDateOutOfRange;
     SDateOutOfMin : Result := RxResConst_Ger.SDateOutOfMin;
     SDateOutOfMax : Result := RxResConst_Ger.SDateOutOfMax;
     SDateMinLimit : Result := RxResConst_Ger.SDateMinLimit;
     SDateMaxLimit : Result := RxResConst_Ger.SDateMaxLimit;
     SLocalDatabase : Result := RxResConst_Ger.SLocalDatabase;
     SRetryLogin : Result := RxResConst_Ger.SRetryLogin;
     SExprNotBoolean : Result := RxResConst_Ger.SExprNotBoolean;
     SExprBadNullTest : Result := RxResConst_Ger.SExprBadNullTest;
     SExprBadField : Result := RxResConst_Ger.SExprBadField;
     SCaptureFilter : Result := RxResConst_Ger.SCaptureFilter;
     SNotCaptureFilter : Result := RxResConst_Ger.SNotCaptureFilter;
     SInactiveData : Result := RxResConst_Ger.SInactiveData;
     SBrowseData : Result := RxResConst_Ger.SBrowseData;
     SEditData : Result := RxResConst_Ger.SEditData;
     SInsertData : Result := RxResConst_Ger.SInsertData;
     SSetKeyData : Result := RxResConst_Ger.SSetKeyData;
     SCalcFieldsData : Result := RxResConst_Ger.SCalcFieldsData;
     SRegistration : Result := RxResConst_Ger.SRegistration;
     SAppTitleLabel : Result := RxResConst_Ger.SAppTitleLabel;
     SHintLabel : Result := RxResConst_Ger.SHintLabel;
     SUserNameLabel : Result := RxResConst_Ger.SUserNameLabel;
     SPasswordLabel : Result := RxResConst_Ger.SPasswordLabel;
     SInvalidUserName : Result := RxResConst_Ger.SInvalidUserName;
     SChangePassword : Result := RxResConst_Ger.SChangePassword;
     SOldPasswordLabel : Result := RxResConst_Ger.SOldPasswordLabel;
     SNewPasswordLabel : Result := RxResConst_Ger.SNewPasswordLabel;
     SConfirmPasswordLabel : Result := RxResConst_Ger.SConfirmPasswordLabel;
     SPasswordChanged : Result := RxResConst_Ger.SPasswordChanged;
     SPasswordNotChanged : Result := RxResConst_Ger.SPasswordNotChanged;
     SPasswordsMismatch : Result := RxResConst_Ger.SPasswordsMismatch;
     SDBExceptCaption : Result := RxResConst_Ger.SDBExceptCaption;
     SBDEErrorLabel : Result := RxResConst_Ger.SBDEErrorLabel;
     SServerErrorLabel : Result := RxResConst_Ger.SServerErrorLabel;
     SErrorMsgLabel : Result := RxResConst_Ger.SErrorMsgLabel;
     SNextButton : Result := RxResConst_Ger.SNextButton;
     SPrevButton : Result := RxResConst_Ger.SPrevButton;
     SExprIncorrect : Result := RxResConst_Ger.SExprIncorrect;
     SExprTermination : Result := RxResConst_Ger.SExprTermination;
     SExprNameError : Result := RxResConst_Ger.SExprNameError;
     SExprStringError : Result := RxResConst_Ger.SExprStringError;
     SExprInvalidChar : Result := RxResConst_Ger.SExprInvalidChar;
     SExprNoRParen : Result := RxResConst_Ger.SExprNoRParen;
     SExprExpected : Result := RxResConst_Ger.SExprExpected;
     SExprBadCompare : Result := RxResConst_Ger.SExprBadCompare;
     SConfirmSave : Result := RxResConst_Ger.SConfirmSave;
     SDatabaseName : Result := RxResConst_Ger.SDatabaseName;
     SUnlockCaption : Result := RxResConst_Ger.SUnlockCaption;
     SUnlockHint : Result := RxResConst_Ger.SUnlockHint;
     SDeleteMultipleRecords : Result := RxResConst_Ger.SDeleteMultipleRecords;
     SGIFImage : Result := RxResConst_Ger.SGIFImage;
     SChangeGIFSize : Result := RxResConst_Ger.SChangeGIFSize;
     SNoGIFData : Result := RxResConst_Ger.SNoGIFData;
     SUnrecognizedGIFExt : Result := RxResConst_Ger.SUnrecognizedGIFExt;
     SWrongGIFColors : Result := RxResConst_Ger.SWrongGIFColors;
     SBadGIFCodeSize : Result := RxResConst_Ger.SBadGIFCodeSize;
     SGIFDecodeError : Result := RxResConst_Ger.SGIFDecodeError;
     SGIFEncodeError : Result := RxResConst_Ger.SGIFEncodeError;
     SGIFVersion : Result := RxResConst_Ger.SGIFVersion;
     SDualListSrcCaption : Result := RxResConst_Ger.SDualListSrcCaption;
     SDualListDestCaption : Result := RxResConst_Ger.SDualListDestCaption;
     SClipbrdUnknown : Result := RxResConst_Ger.SClipbrdUnknown;
     SClipbrdEmpty : Result := RxResConst_Ger.SClipbrdEmpty;
     SCustomizeSpeedbar : Result := RxResConst_Ger.SCustomizeSpeedbar;
     SAvailButtons : Result := RxResConst_Ger.SAvailButtons;
     SSpeedbarCategories : Result := RxResConst_Ger.SSpeedbarCategories;
     SSpeedbarEditHint : Result := RxResConst_Ger.SSpeedbarEditHint;
     SParseSyntaxError : Result := RxResConst_Ger.SParseSyntaxError;
     SParseNotCramp : Result := RxResConst_Ger.SParseNotCramp;
     SParseDivideByZero : Result := RxResConst_Ger.SParseDivideByZero;
     SParseSqrError : Result := RxResConst_Ger.SParseSqrError;
     SParseLogError : Result := RxResConst_Ger.SParseLogError;
     SParseInvalidFloatOperation : Result := RxResConst_Ger.SParseInvalidFloatOperation;
     srRXControls : Result := RxResConst_Ger.srRXControls;
     srRXDBAware : Result := RxResConst_Ger.srRXDBAware;
     srRXTools : Result := RxResConst_Ger.srRXTools;
     srStorageDesigner : Result := RxResConst_Ger.srStorageDesigner;
     srProxyEditor : Result := RxResConst_Ger.srProxyEditor;
     srPageProxies : Result := RxResConst_Ger.srPageProxies;
     srProxyName : Result := RxResConst_Ger.srProxyName;
     srPageName : Result := RxResConst_Ger.srPageName;
     srSBItemNotCreate : Result := RxResConst_Ger.srSBItemNotCreate;
     srConfirmSBDelete : Result := RxResConst_Ger.srConfirmSBDelete;
     srSpeedbarDesigner : Result := RxResConst_Ger.srSpeedbarDesigner;
     srNewSectionName : Result := RxResConst_Ger.srNewSectionName;
     srEventNotCreate : Result := RxResConst_Ger.srEventNotCreate;
     srTimerDesigner : Result := RxResConst_Ger.srTimerDesigner;
     srTimerEvents : Result := RxResConst_Ger.srTimerEvents;
     srAniCurFilter : Result := RxResConst_Ger.srAniCurFilter;
     srEditPicture : Result := RxResConst_Ger.srEditPicture;
     srLoadAniCursor : Result := RxResConst_Ger.srLoadAniCursor;
     srLoadIcon : Result := RxResConst_Ger.srLoadIcon;
     srSaveImageList : Result := RxResConst_Ger.srSaveImageList;
     srCaptionDesigner : Result := RxResConst_Ger.srCaptionDesigner;
     srGradientCaptions : Result := RxResConst_Ger.srGradientCaptions;
     srBorrowStructure : Result := RxResConst_Ger.srBorrowStructure;
     ccTimeEditHours : Result := RxResConst_Ger.ccTimeEditHours;
     ccTimeEditMins : Result := RxResConst_Ger.ccTimeEditMins;
     ccTimeEditSecs : Result := RxResConst_Ger.ccTimeEditSecs;
     SColorBlack : Result := RxResConst_Ger.SColorBlack;
     SColorMaroon : Result := RxResConst_Ger.SColorMaroon;
     SColorGreen : Result := RxResConst_Ger.SColorGreen;
     SColorOlive : Result := RxResConst_Ger.SColorOlive;
     SColorNavy : Result := RxResConst_Ger.SColorNavy;
     SColorPurple : Result := RxResConst_Ger.SColorPurple;
     SColorTeal : Result := RxResConst_Ger.SColorTeal;
     SColorGray : Result := RxResConst_Ger.SColorGray;
     SColorSilver : Result := RxResConst_Ger.SColorSilver;
     SColorRed : Result := RxResConst_Ger.SColorRed;
     SColorLime : Result := RxResConst_Ger.SColorLime;
     SColorYellow : Result := RxResConst_Ger.SColorYellow;
     SColorBlue : Result := RxResConst_Ger.SColorBlue;
     SColorFuchsia : Result := RxResConst_Ger.SColorFuchsia;
     SColorAqua : Result := RxResConst_Ger.SColorAqua;
     SColorWhite : Result := RxResConst_Ger.SColorWhite;
     SColorScrollBar : Result := RxResConst_Ger.SColorScrollBar;
     SColorBackground : Result := RxResConst_Ger.SColorBackground;
     SColorActiveCaption : Result := RxResConst_Ger.SColorActiveCaption;
     SColorInactiveCaption : Result := RxResConst_Ger.SColorInactiveCaption;
     SColorMenu : Result := RxResConst_Ger.SColorMenu;
     SColorWindow : Result := RxResConst_Ger.SColorWindow;
     SColorWindowFrame : Result := RxResConst_Ger.SColorWindowFrame;
     SColorMenuText : Result := RxResConst_Ger.SColorMenuText;
     SColorWindowText : Result := RxResConst_Ger.SColorWindowText;
     SColorCaptionText : Result := RxResConst_Ger.SColorCaptionText;
     SColorActiveBorder : Result := RxResConst_Ger.SColorActiveBorder;
     SColorInactiveBorder : Result := RxResConst_Ger.SColorInactiveBorder;
     SColorAppWorkSpace : Result := RxResConst_Ger.SColorAppWorkSpace;
     SColorHighlight : Result := RxResConst_Ger.SColorHighlight;
     SColorHighlightText : Result := RxResConst_Ger.SColorHighlightText;
     SColorBtnFace : Result := RxResConst_Ger.SColorBtnFace;
     SColorBtnShadow : Result := RxResConst_Ger.SColorBtnShadow;
     SColorGrayText : Result := RxResConst_Ger.SColorGrayText;
     SColorBtnText : Result := RxResConst_Ger.SColorBtnText;
     SColorInactiveCaptionText : Result := RxResConst_Ger.SColorInactiveCaptionText;
     SColorBtnHighlight : Result := RxResConst_Ger.SColorBtnHighlight;
     SColor3DDkShadow : Result := RxResConst_Ger.SColor3DDkShadow;
     SColor3DLight : Result := RxResConst_Ger.SColor3DLight;
     SColorInfoText : Result := RxResConst_Ger.SColorInfoText;
     SColorInfoBk : Result := RxResConst_Ger.SColorInfoBk;
     SColorCream : Result := RxResConst_Ger.SColorCream;
     SColorMoneyGreen : Result := RxResConst_Ger.SColorMoneyGreen;
     SColorSkyBlue : Result := RxResConst_Ger.SColorSkyBlue;
     SColorNone : Result := RxResConst_Ger.SColorNone;
     SColorDefault : Result := RxResConst_Ger.SColorDefault;
     SColorCustom : Result := RxResConst_Ger.SColorCustom;
     SMemNoRecords : Result := RxResConst_Ger.SMemNoRecords;
     SInvalidFields : Result := RxResConst_Ger.SInvalidFields;
     SAutoSpeedbarMode : Result := RxResConst_Ger.SAutoSpeedbarMode;
    end;
  end;
  {$ENDIF}
  {$IFDEF RXLANG_Gre}
  function RxResConst_Gre_RxLoadStr(IDMsg: Integer): string;
  begin
    case IDMsg of
     SBrowse : Result := RxResConst_Gre.SBrowse;
     SDefaultFilter : Result := RxResConst_Gre.SDefaultFilter;
     SDateDlgTitle : Result := RxResConst_Gre.SDateDlgTitle;
     SNextYear : Result := RxResConst_Gre.SNextYear;
     SNextMonth : Result := RxResConst_Gre.SNextMonth;
     SPrevYear : Result := RxResConst_Gre.SPrevYear;
     SPrevMonth : Result := RxResConst_Gre.SPrevMonth;
     SNotImplemented : Result := RxResConst_Gre.SNotImplemented;
     SFileNotExec : Result := RxResConst_Gre.SFileNotExec;
     SLoadLibError : Result := RxResConst_Gre.SLoadLibError;
     SDetails : Result := RxResConst_Gre.SDetails;
     SDateOutOfRange : Result := RxResConst_Gre.SDateOutOfRange;
     SDateOutOfMin : Result := RxResConst_Gre.SDateOutOfMin;
     SDateOutOfMax : Result := RxResConst_Gre.SDateOutOfMax;
     SDateMinLimit : Result := RxResConst_Gre.SDateMinLimit;
     SDateMaxLimit : Result := RxResConst_Gre.SDateMaxLimit;
     SLocalDatabase : Result := RxResConst_Gre.SLocalDatabase;
     SRetryLogin : Result := RxResConst_Gre.SRetryLogin;
     SExprNotBoolean : Result := RxResConst_Gre.SExprNotBoolean;
     SExprBadNullTest : Result := RxResConst_Gre.SExprBadNullTest;
     SExprBadField : Result := RxResConst_Gre.SExprBadField;
     SCaptureFilter : Result := RxResConst_Gre.SCaptureFilter;
     SNotCaptureFilter : Result := RxResConst_Gre.SNotCaptureFilter;
     SInactiveData : Result := RxResConst_Gre.SInactiveData;
     SBrowseData : Result := RxResConst_Gre.SBrowseData;
     SEditData : Result := RxResConst_Gre.SEditData;
     SInsertData : Result := RxResConst_Gre.SInsertData;
     SSetKeyData : Result := RxResConst_Gre.SSetKeyData;
     SCalcFieldsData : Result := RxResConst_Gre.SCalcFieldsData;
     SRegistration : Result := RxResConst_Gre.SRegistration;
     SAppTitleLabel : Result := RxResConst_Gre.SAppTitleLabel;
     SHintLabel : Result := RxResConst_Gre.SHintLabel;
     SUserNameLabel : Result := RxResConst_Gre.SUserNameLabel;
     SPasswordLabel : Result := RxResConst_Gre.SPasswordLabel;
     SInvalidUserName : Result := RxResConst_Gre.SInvalidUserName;
     SChangePassword : Result := RxResConst_Gre.SChangePassword;
     SOldPasswordLabel : Result := RxResConst_Gre.SOldPasswordLabel;
     SNewPasswordLabel : Result := RxResConst_Gre.SNewPasswordLabel;
     SConfirmPasswordLabel : Result := RxResConst_Gre.SConfirmPasswordLabel;
     SPasswordChanged : Result := RxResConst_Gre.SPasswordChanged;
     SPasswordNotChanged : Result := RxResConst_Gre.SPasswordNotChanged;
     SPasswordsMismatch : Result := RxResConst_Gre.SPasswordsMismatch;
     SDBExceptCaption : Result := RxResConst_Gre.SDBExceptCaption;
     SBDEErrorLabel : Result := RxResConst_Gre.SBDEErrorLabel;
     SServerErrorLabel : Result := RxResConst_Gre.SServerErrorLabel;
     SErrorMsgLabel : Result := RxResConst_Gre.SErrorMsgLabel;
     SNextButton : Result := RxResConst_Gre.SNextButton;
     SPrevButton : Result := RxResConst_Gre.SPrevButton;
     SExprIncorrect : Result := RxResConst_Gre.SExprIncorrect;
     SExprTermination : Result := RxResConst_Gre.SExprTermination;
     SExprNameError : Result := RxResConst_Gre.SExprNameError;
     SExprStringError : Result := RxResConst_Gre.SExprStringError;
     SExprInvalidChar : Result := RxResConst_Gre.SExprInvalidChar;
     SExprNoRParen : Result := RxResConst_Gre.SExprNoRParen;
     SExprExpected : Result := RxResConst_Gre.SExprExpected;
     SExprBadCompare : Result := RxResConst_Gre.SExprBadCompare;
     SConfirmSave : Result := RxResConst_Gre.SConfirmSave;
     SDatabaseName : Result := RxResConst_Gre.SDatabaseName;
     SUnlockCaption : Result := RxResConst_Gre.SUnlockCaption;
     SUnlockHint : Result := RxResConst_Gre.SUnlockHint;
     SDeleteMultipleRecords : Result := RxResConst_Gre.SDeleteMultipleRecords;
     SGIFImage : Result := RxResConst_Gre.SGIFImage;
     SChangeGIFSize : Result := RxResConst_Gre.SChangeGIFSize;
     SNoGIFData : Result := RxResConst_Gre.SNoGIFData;
     SUnrecognizedGIFExt : Result := RxResConst_Gre.SUnrecognizedGIFExt;
     SWrongGIFColors : Result := RxResConst_Gre.SWrongGIFColors;
     SBadGIFCodeSize : Result := RxResConst_Gre.SBadGIFCodeSize;
     SGIFDecodeError : Result := RxResConst_Gre.SGIFDecodeError;
     SGIFEncodeError : Result := RxResConst_Gre.SGIFEncodeError;
     SGIFVersion : Result := RxResConst_Gre.SGIFVersion;
     SDualListSrcCaption : Result := RxResConst_Gre.SDualListSrcCaption;
     SDualListDestCaption : Result := RxResConst_Gre.SDualListDestCaption;
     SClipbrdUnknown : Result := RxResConst_Gre.SClipbrdUnknown;
     SClipbrdEmpty : Result := RxResConst_Gre.SClipbrdEmpty;
     SCustomizeSpeedbar : Result := RxResConst_Gre.SCustomizeSpeedbar;
     SAvailButtons : Result := RxResConst_Gre.SAvailButtons;
     SSpeedbarCategories : Result := RxResConst_Gre.SSpeedbarCategories;
     SSpeedbarEditHint : Result := RxResConst_Gre.SSpeedbarEditHint;
     SParseSyntaxError : Result := RxResConst_Gre.SParseSyntaxError;
     SParseNotCramp : Result := RxResConst_Gre.SParseNotCramp;
     SParseDivideByZero : Result := RxResConst_Gre.SParseDivideByZero;
     SParseSqrError : Result := RxResConst_Gre.SParseSqrError;
     SParseLogError : Result := RxResConst_Gre.SParseLogError;
     SParseInvalidFloatOperation : Result := RxResConst_Gre.SParseInvalidFloatOperation;
     srRXControls : Result := RxResConst_Gre.srRXControls;
     srRXDBAware : Result := RxResConst_Gre.srRXDBAware;
     srRXTools : Result := RxResConst_Gre.srRXTools;
     srStorageDesigner : Result := RxResConst_Gre.srStorageDesigner;
     srProxyEditor : Result := RxResConst_Gre.srProxyEditor;
     srPageProxies : Result := RxResConst_Gre.srPageProxies;
     srProxyName : Result := RxResConst_Gre.srProxyName;
     srPageName : Result := RxResConst_Gre.srPageName;
     srSBItemNotCreate : Result := RxResConst_Gre.srSBItemNotCreate;
     srConfirmSBDelete : Result := RxResConst_Gre.srConfirmSBDelete;
     srSpeedbarDesigner : Result := RxResConst_Gre.srSpeedbarDesigner;
     srNewSectionName : Result := RxResConst_Gre.srNewSectionName;
     srEventNotCreate : Result := RxResConst_Gre.srEventNotCreate;
     srTimerDesigner : Result := RxResConst_Gre.srTimerDesigner;
     srTimerEvents : Result := RxResConst_Gre.srTimerEvents;
     srAniCurFilter : Result := RxResConst_Gre.srAniCurFilter;
     srEditPicture : Result := RxResConst_Gre.srEditPicture;
     srLoadAniCursor : Result := RxResConst_Gre.srLoadAniCursor;
     srLoadIcon : Result := RxResConst_Gre.srLoadIcon;
     srSaveImageList : Result := RxResConst_Gre.srSaveImageList;
     srCaptionDesigner : Result := RxResConst_Gre.srCaptionDesigner;
     srGradientCaptions : Result := RxResConst_Gre.srGradientCaptions;
     srBorrowStructure : Result := RxResConst_Gre.srBorrowStructure;
     ccTimeEditHours : Result := RxResConst_Gre.ccTimeEditHours;
     ccTimeEditMins : Result := RxResConst_Gre.ccTimeEditMins;
     ccTimeEditSecs : Result := RxResConst_Gre.ccTimeEditSecs;
     SColorBlack : Result := RxResConst_Gre.SColorBlack;
     SColorMaroon : Result := RxResConst_Gre.SColorMaroon;
     SColorGreen : Result := RxResConst_Gre.SColorGreen;
     SColorOlive : Result := RxResConst_Gre.SColorOlive;
     SColorNavy : Result := RxResConst_Gre.SColorNavy;
     SColorPurple : Result := RxResConst_Gre.SColorPurple;
     SColorTeal : Result := RxResConst_Gre.SColorTeal;
     SColorGray : Result := RxResConst_Gre.SColorGray;
     SColorSilver : Result := RxResConst_Gre.SColorSilver;
     SColorRed : Result := RxResConst_Gre.SColorRed;
     SColorLime : Result := RxResConst_Gre.SColorLime;
     SColorYellow : Result := RxResConst_Gre.SColorYellow;
     SColorBlue : Result := RxResConst_Gre.SColorBlue;
     SColorFuchsia : Result := RxResConst_Gre.SColorFuchsia;
     SColorAqua : Result := RxResConst_Gre.SColorAqua;
     SColorWhite : Result := RxResConst_Gre.SColorWhite;
     SColorScrollBar : Result := RxResConst_Gre.SColorScrollBar;
     SColorBackground : Result := RxResConst_Gre.SColorBackground;
     SColorActiveCaption : Result := RxResConst_Gre.SColorActiveCaption;
     SColorInactiveCaption : Result := RxResConst_Gre.SColorInactiveCaption;
     SColorMenu : Result := RxResConst_Gre.SColorMenu;
     SColorWindow : Result := RxResConst_Gre.SColorWindow;
     SColorWindowFrame : Result := RxResConst_Gre.SColorWindowFrame;
     SColorMenuText : Result := RxResConst_Gre.SColorMenuText;
     SColorWindowText : Result := RxResConst_Gre.SColorWindowText;
     SColorCaptionText : Result := RxResConst_Gre.SColorCaptionText;
     SColorActiveBorder : Result := RxResConst_Gre.SColorActiveBorder;
     SColorInactiveBorder : Result := RxResConst_Gre.SColorInactiveBorder;
     SColorAppWorkSpace : Result := RxResConst_Gre.SColorAppWorkSpace;
     SColorHighlight : Result := RxResConst_Gre.SColorHighlight;
     SColorHighlightText : Result := RxResConst_Gre.SColorHighlightText;
     SColorBtnFace : Result := RxResConst_Gre.SColorBtnFace;
     SColorBtnShadow : Result := RxResConst_Gre.SColorBtnShadow;
     SColorGrayText : Result := RxResConst_Gre.SColorGrayText;
     SColorBtnText : Result := RxResConst_Gre.SColorBtnText;
     SColorInactiveCaptionText : Result := RxResConst_Gre.SColorInactiveCaptionText;
     SColorBtnHighlight : Result := RxResConst_Gre.SColorBtnHighlight;
     SColor3DDkShadow : Result := RxResConst_Gre.SColor3DDkShadow;
     SColor3DLight : Result := RxResConst_Gre.SColor3DLight;
     SColorInfoText : Result := RxResConst_Gre.SColorInfoText;
     SColorInfoBk : Result := RxResConst_Gre.SColorInfoBk;
     SColorCream : Result := RxResConst_Gre.SColorCream;
     SColorMoneyGreen : Result := RxResConst_Gre.SColorMoneyGreen;
     SColorSkyBlue : Result := RxResConst_Gre.SColorSkyBlue;
     SColorNone : Result := RxResConst_Gre.SColorNone;
     SColorDefault : Result := RxResConst_Gre.SColorDefault;
     SColorCustom : Result := RxResConst_Gre.SColorCustom;
     SMemNoRecords : Result := RxResConst_Gre.SMemNoRecords;
     SInvalidFields : Result := RxResConst_Gre.SInvalidFields;
     SAutoSpeedbarMode : Result := RxResConst_Gre.SAutoSpeedbarMode;
    end;
  end;
  {$ENDIF}
  {$IFDEF RXLANG_Hun}
  function RxResConst_Hun_RxLoadStr(IDMsg: Integer): string;
  begin
    case IDMsg of
     SBrowse : Result := RxResConst_Hun.SBrowse;
     SDefaultFilter : Result := RxResConst_Hun.SDefaultFilter;
     SDateDlgTitle : Result := RxResConst_Hun.SDateDlgTitle;
     SNextYear : Result := RxResConst_Hun.SNextYear;
     SNextMonth : Result := RxResConst_Hun.SNextMonth;
     SPrevYear : Result := RxResConst_Hun.SPrevYear;
     SPrevMonth : Result := RxResConst_Hun.SPrevMonth;
     SNotImplemented : Result := RxResConst_Hun.SNotImplemented;
     SFileNotExec : Result := RxResConst_Hun.SFileNotExec;
     SLoadLibError : Result := RxResConst_Hun.SLoadLibError;
     SDetails : Result := RxResConst_Hun.SDetails;
     SDateOutOfRange : Result := RxResConst_Hun.SDateOutOfRange;
     SDateOutOfMin : Result := RxResConst_Hun.SDateOutOfMin;
     SDateOutOfMax : Result := RxResConst_Hun.SDateOutOfMax;
     SDateMinLimit : Result := RxResConst_Hun.SDateMinLimit;
     SDateMaxLimit : Result := RxResConst_Hun.SDateMaxLimit;
     SLocalDatabase : Result := RxResConst_Hun.SLocalDatabase;
     SRetryLogin : Result := RxResConst_Hun.SRetryLogin;
     SExprNotBoolean : Result := RxResConst_Hun.SExprNotBoolean;
     SExprBadNullTest : Result := RxResConst_Hun.SExprBadNullTest;
     SExprBadField : Result := RxResConst_Hun.SExprBadField;
     SCaptureFilter : Result := RxResConst_Hun.SCaptureFilter;
     SNotCaptureFilter : Result := RxResConst_Hun.SNotCaptureFilter;
     SInactiveData : Result := RxResConst_Hun.SInactiveData;
     SBrowseData : Result := RxResConst_Hun.SBrowseData;
     SEditData : Result := RxResConst_Hun.SEditData;
     SInsertData : Result := RxResConst_Hun.SInsertData;
     SSetKeyData : Result := RxResConst_Hun.SSetKeyData;
     SCalcFieldsData : Result := RxResConst_Hun.SCalcFieldsData;
     SRegistration : Result := RxResConst_Hun.SRegistration;
     SAppTitleLabel : Result := RxResConst_Hun.SAppTitleLabel;
     SHintLabel : Result := RxResConst_Hun.SHintLabel;
     SUserNameLabel : Result := RxResConst_Hun.SUserNameLabel;
     SPasswordLabel : Result := RxResConst_Hun.SPasswordLabel;
     SInvalidUserName : Result := RxResConst_Hun.SInvalidUserName;
     SChangePassword : Result := RxResConst_Hun.SChangePassword;
     SOldPasswordLabel : Result := RxResConst_Hun.SOldPasswordLabel;
     SNewPasswordLabel : Result := RxResConst_Hun.SNewPasswordLabel;
     SConfirmPasswordLabel : Result := RxResConst_Hun.SConfirmPasswordLabel;
     SPasswordChanged : Result := RxResConst_Hun.SPasswordChanged;
     SPasswordNotChanged : Result := RxResConst_Hun.SPasswordNotChanged;
     SPasswordsMismatch : Result := RxResConst_Hun.SPasswordsMismatch;
     SDBExceptCaption : Result := RxResConst_Hun.SDBExceptCaption;
     SBDEErrorLabel : Result := RxResConst_Hun.SBDEErrorLabel;
     SServerErrorLabel : Result := RxResConst_Hun.SServerErrorLabel;
     SErrorMsgLabel : Result := RxResConst_Hun.SErrorMsgLabel;
     SNextButton : Result := RxResConst_Hun.SNextButton;
     SPrevButton : Result := RxResConst_Hun.SPrevButton;
     SExprIncorrect : Result := RxResConst_Hun.SExprIncorrect;
     SExprTermination : Result := RxResConst_Hun.SExprTermination;
     SExprNameError : Result := RxResConst_Hun.SExprNameError;
     SExprStringError : Result := RxResConst_Hun.SExprStringError;
     SExprInvalidChar : Result := RxResConst_Hun.SExprInvalidChar;
     SExprNoRParen : Result := RxResConst_Hun.SExprNoRParen;
     SExprExpected : Result := RxResConst_Hun.SExprExpected;
     SExprBadCompare : Result := RxResConst_Hun.SExprBadCompare;
     SConfirmSave : Result := RxResConst_Hun.SConfirmSave;
     SDatabaseName : Result := RxResConst_Hun.SDatabaseName;
     SUnlockCaption : Result := RxResConst_Hun.SUnlockCaption;
     SUnlockHint : Result := RxResConst_Hun.SUnlockHint;
     SDeleteMultipleRecords : Result := RxResConst_Hun.SDeleteMultipleRecords;
     SGIFImage : Result := RxResConst_Hun.SGIFImage;
     SChangeGIFSize : Result := RxResConst_Hun.SChangeGIFSize;
     SNoGIFData : Result := RxResConst_Hun.SNoGIFData;
     SUnrecognizedGIFExt : Result := RxResConst_Hun.SUnrecognizedGIFExt;
     SWrongGIFColors : Result := RxResConst_Hun.SWrongGIFColors;
     SBadGIFCodeSize : Result := RxResConst_Hun.SBadGIFCodeSize;
     SGIFDecodeError : Result := RxResConst_Hun.SGIFDecodeError;
     SGIFEncodeError : Result := RxResConst_Hun.SGIFEncodeError;
     SGIFVersion : Result := RxResConst_Hun.SGIFVersion;
     SDualListSrcCaption : Result := RxResConst_Hun.SDualListSrcCaption;
     SDualListDestCaption : Result := RxResConst_Hun.SDualListDestCaption;
     SClipbrdUnknown : Result := RxResConst_Hun.SClipbrdUnknown;
     SClipbrdEmpty : Result := RxResConst_Hun.SClipbrdEmpty;
     SCustomizeSpeedbar : Result := RxResConst_Hun.SCustomizeSpeedbar;
     SAvailButtons : Result := RxResConst_Hun.SAvailButtons;
     SSpeedbarCategories : Result := RxResConst_Hun.SSpeedbarCategories;
     SSpeedbarEditHint : Result := RxResConst_Hun.SSpeedbarEditHint;
     SParseSyntaxError : Result := RxResConst_Hun.SParseSyntaxError;
     SParseNotCramp : Result := RxResConst_Hun.SParseNotCramp;
     SParseDivideByZero : Result := RxResConst_Hun.SParseDivideByZero;
     SParseSqrError : Result := RxResConst_Hun.SParseSqrError;
     SParseLogError : Result := RxResConst_Hun.SParseLogError;
     SParseInvalidFloatOperation : Result := RxResConst_Hun.SParseInvalidFloatOperation;
     srRXControls : Result := RxResConst_Hun.srRXControls;
     srRXDBAware : Result := RxResConst_Hun.srRXDBAware;
     srRXTools : Result := RxResConst_Hun.srRXTools;
     srStorageDesigner : Result := RxResConst_Hun.srStorageDesigner;
     srProxyEditor : Result := RxResConst_Hun.srProxyEditor;
     srPageProxies : Result := RxResConst_Hun.srPageProxies;
     srProxyName : Result := RxResConst_Hun.srProxyName;
     srPageName : Result := RxResConst_Hun.srPageName;
     srSBItemNotCreate : Result := RxResConst_Hun.srSBItemNotCreate;
     srConfirmSBDelete : Result := RxResConst_Hun.srConfirmSBDelete;
     srSpeedbarDesigner : Result := RxResConst_Hun.srSpeedbarDesigner;
     srNewSectionName : Result := RxResConst_Hun.srNewSectionName;
     srEventNotCreate : Result := RxResConst_Hun.srEventNotCreate;
     srTimerDesigner : Result := RxResConst_Hun.srTimerDesigner;
     srTimerEvents : Result := RxResConst_Hun.srTimerEvents;
     srAniCurFilter : Result := RxResConst_Hun.srAniCurFilter;
     srEditPicture : Result := RxResConst_Hun.srEditPicture;
     srLoadAniCursor : Result := RxResConst_Hun.srLoadAniCursor;
     srLoadIcon : Result := RxResConst_Hun.srLoadIcon;
     srSaveImageList : Result := RxResConst_Hun.srSaveImageList;
     srCaptionDesigner : Result := RxResConst_Hun.srCaptionDesigner;
     srGradientCaptions : Result := RxResConst_Hun.srGradientCaptions;
     srBorrowStructure : Result := RxResConst_Hun.srBorrowStructure;
     ccTimeEditHours : Result := RxResConst_Hun.ccTimeEditHours;
     ccTimeEditMins : Result := RxResConst_Hun.ccTimeEditMins;
     ccTimeEditSecs : Result := RxResConst_Hun.ccTimeEditSecs;
     SColorBlack : Result := RxResConst_Hun.SColorBlack;
     SColorMaroon : Result := RxResConst_Hun.SColorMaroon;
     SColorGreen : Result := RxResConst_Hun.SColorGreen;
     SColorOlive : Result := RxResConst_Hun.SColorOlive;
     SColorNavy : Result := RxResConst_Hun.SColorNavy;
     SColorPurple : Result := RxResConst_Hun.SColorPurple;
     SColorTeal : Result := RxResConst_Hun.SColorTeal;
     SColorGray : Result := RxResConst_Hun.SColorGray;
     SColorSilver : Result := RxResConst_Hun.SColorSilver;
     SColorRed : Result := RxResConst_Hun.SColorRed;
     SColorLime : Result := RxResConst_Hun.SColorLime;
     SColorYellow : Result := RxResConst_Hun.SColorYellow;
     SColorBlue : Result := RxResConst_Hun.SColorBlue;
     SColorFuchsia : Result := RxResConst_Hun.SColorFuchsia;
     SColorAqua : Result := RxResConst_Hun.SColorAqua;
     SColorWhite : Result := RxResConst_Hun.SColorWhite;
     SColorScrollBar : Result := RxResConst_Hun.SColorScrollBar;
     SColorBackground : Result := RxResConst_Hun.SColorBackground;
     SColorActiveCaption : Result := RxResConst_Hun.SColorActiveCaption;
     SColorInactiveCaption : Result := RxResConst_Hun.SColorInactiveCaption;
     SColorMenu : Result := RxResConst_Hun.SColorMenu;
     SColorWindow : Result := RxResConst_Hun.SColorWindow;
     SColorWindowFrame : Result := RxResConst_Hun.SColorWindowFrame;
     SColorMenuText : Result := RxResConst_Hun.SColorMenuText;
     SColorWindowText : Result := RxResConst_Hun.SColorWindowText;
     SColorCaptionText : Result := RxResConst_Hun.SColorCaptionText;
     SColorActiveBorder : Result := RxResConst_Hun.SColorActiveBorder;
     SColorInactiveBorder : Result := RxResConst_Hun.SColorInactiveBorder;
     SColorAppWorkSpace : Result := RxResConst_Hun.SColorAppWorkSpace;
     SColorHighlight : Result := RxResConst_Hun.SColorHighlight;
     SColorHighlightText : Result := RxResConst_Hun.SColorHighlightText;
     SColorBtnFace : Result := RxResConst_Hun.SColorBtnFace;
     SColorBtnShadow : Result := RxResConst_Hun.SColorBtnShadow;
     SColorGrayText : Result := RxResConst_Hun.SColorGrayText;
     SColorBtnText : Result := RxResConst_Hun.SColorBtnText;
     SColorInactiveCaptionText : Result := RxResConst_Hun.SColorInactiveCaptionText;
     SColorBtnHighlight : Result := RxResConst_Hun.SColorBtnHighlight;
     SColor3DDkShadow : Result := RxResConst_Hun.SColor3DDkShadow;
     SColor3DLight : Result := RxResConst_Hun.SColor3DLight;
     SColorInfoText : Result := RxResConst_Hun.SColorInfoText;
     SColorInfoBk : Result := RxResConst_Hun.SColorInfoBk;
     SColorCream : Result := RxResConst_Hun.SColorCream;
     SColorMoneyGreen : Result := RxResConst_Hun.SColorMoneyGreen;
     SColorSkyBlue : Result := RxResConst_Hun.SColorSkyBlue;
     SColorNone : Result := RxResConst_Hun.SColorNone;
     SColorDefault : Result := RxResConst_Hun.SColorDefault;
     SColorCustom : Result := RxResConst_Hun.SColorCustom;
     SMemNoRecords : Result := RxResConst_Hun.SMemNoRecords;
     SInvalidFields : Result := RxResConst_Hun.SInvalidFields;
     SAutoSpeedbarMode : Result := RxResConst_Hun.SAutoSpeedbarMode;
    end;
  end;
  {$ENDIF}
  {$IFDEF RXLANG_Ita}
  function RxResConst_Ita_RxLoadStr(IDMsg: Integer): string;
  begin
    case IDMsg of
     SBrowse : Result := RxResConst_Ita.SBrowse;
     SDefaultFilter : Result := RxResConst_Ita.SDefaultFilter;
     SDateDlgTitle : Result := RxResConst_Ita.SDateDlgTitle;
     SNextYear : Result := RxResConst_Ita.SNextYear;
     SNextMonth : Result := RxResConst_Ita.SNextMonth;
     SPrevYear : Result := RxResConst_Ita.SPrevYear;
     SPrevMonth : Result := RxResConst_Ita.SPrevMonth;
     SNotImplemented : Result := RxResConst_Ita.SNotImplemented;
     SFileNotExec : Result := RxResConst_Ita.SFileNotExec;
     SLoadLibError : Result := RxResConst_Ita.SLoadLibError;
     SDetails : Result := RxResConst_Ita.SDetails;
     SDateOutOfRange : Result := RxResConst_Ita.SDateOutOfRange;
     SDateOutOfMin : Result := RxResConst_Ita.SDateOutOfMin;
     SDateOutOfMax : Result := RxResConst_Ita.SDateOutOfMax;
     SDateMinLimit : Result := RxResConst_Ita.SDateMinLimit;
     SDateMaxLimit : Result := RxResConst_Ita.SDateMaxLimit;
     SLocalDatabase : Result := RxResConst_Ita.SLocalDatabase;
     SRetryLogin : Result := RxResConst_Ita.SRetryLogin;
     SExprNotBoolean : Result := RxResConst_Ita.SExprNotBoolean;
     SExprBadNullTest : Result := RxResConst_Ita.SExprBadNullTest;
     SExprBadField : Result := RxResConst_Ita.SExprBadField;
     SCaptureFilter : Result := RxResConst_Ita.SCaptureFilter;
     SNotCaptureFilter : Result := RxResConst_Ita.SNotCaptureFilter;
     SInactiveData : Result := RxResConst_Ita.SInactiveData;
     SBrowseData : Result := RxResConst_Ita.SBrowseData;
     SEditData : Result := RxResConst_Ita.SEditData;
     SInsertData : Result := RxResConst_Ita.SInsertData;
     SSetKeyData : Result := RxResConst_Ita.SSetKeyData;
     SCalcFieldsData : Result := RxResConst_Ita.SCalcFieldsData;
     SRegistration : Result := RxResConst_Ita.SRegistration;
     SAppTitleLabel : Result := RxResConst_Ita.SAppTitleLabel;
     SHintLabel : Result := RxResConst_Ita.SHintLabel;
     SUserNameLabel : Result := RxResConst_Ita.SUserNameLabel;
     SPasswordLabel : Result := RxResConst_Ita.SPasswordLabel;
     SInvalidUserName : Result := RxResConst_Ita.SInvalidUserName;
     SChangePassword : Result := RxResConst_Ita.SChangePassword;
     SOldPasswordLabel : Result := RxResConst_Ita.SOldPasswordLabel;
     SNewPasswordLabel : Result := RxResConst_Ita.SNewPasswordLabel;
     SConfirmPasswordLabel : Result := RxResConst_Ita.SConfirmPasswordLabel;
     SPasswordChanged : Result := RxResConst_Ita.SPasswordChanged;
     SPasswordNotChanged : Result := RxResConst_Ita.SPasswordNotChanged;
     SPasswordsMismatch : Result := RxResConst_Ita.SPasswordsMismatch;
     SDBExceptCaption : Result := RxResConst_Ita.SDBExceptCaption;
     SBDEErrorLabel : Result := RxResConst_Ita.SBDEErrorLabel;
     SServerErrorLabel : Result := RxResConst_Ita.SServerErrorLabel;
     SErrorMsgLabel : Result := RxResConst_Ita.SErrorMsgLabel;
     SNextButton : Result := RxResConst_Ita.SNextButton;
     SPrevButton : Result := RxResConst_Ita.SPrevButton;
     SExprIncorrect : Result := RxResConst_Ita.SExprIncorrect;
     SExprTermination : Result := RxResConst_Ita.SExprTermination;
     SExprNameError : Result := RxResConst_Ita.SExprNameError;
     SExprStringError : Result := RxResConst_Ita.SExprStringError;
     SExprInvalidChar : Result := RxResConst_Ita.SExprInvalidChar;
     SExprNoRParen : Result := RxResConst_Ita.SExprNoRParen;
     SExprExpected : Result := RxResConst_Ita.SExprExpected;
     SExprBadCompare : Result := RxResConst_Ita.SExprBadCompare;
     SConfirmSave : Result := RxResConst_Ita.SConfirmSave;
     SDatabaseName : Result := RxResConst_Ita.SDatabaseName;
     SUnlockCaption : Result := RxResConst_Ita.SUnlockCaption;
     SUnlockHint : Result := RxResConst_Ita.SUnlockHint;
     SDeleteMultipleRecords : Result := RxResConst_Ita.SDeleteMultipleRecords;
     SGIFImage : Result := RxResConst_Ita.SGIFImage;
     SChangeGIFSize : Result := RxResConst_Ita.SChangeGIFSize;
     SNoGIFData : Result := RxResConst_Ita.SNoGIFData;
     SUnrecognizedGIFExt : Result := RxResConst_Ita.SUnrecognizedGIFExt;
     SWrongGIFColors : Result := RxResConst_Ita.SWrongGIFColors;
     SBadGIFCodeSize : Result := RxResConst_Ita.SBadGIFCodeSize;
     SGIFDecodeError : Result := RxResConst_Ita.SGIFDecodeError;
     SGIFEncodeError : Result := RxResConst_Ita.SGIFEncodeError;
     SGIFVersion : Result := RxResConst_Ita.SGIFVersion;
     SDualListSrcCaption : Result := RxResConst_Ita.SDualListSrcCaption;
     SDualListDestCaption : Result := RxResConst_Ita.SDualListDestCaption;
     SClipbrdUnknown : Result := RxResConst_Ita.SClipbrdUnknown;
     SClipbrdEmpty : Result := RxResConst_Ita.SClipbrdEmpty;
     SCustomizeSpeedbar : Result := RxResConst_Ita.SCustomizeSpeedbar;
     SAvailButtons : Result := RxResConst_Ita.SAvailButtons;
     SSpeedbarCategories : Result := RxResConst_Ita.SSpeedbarCategories;
     SSpeedbarEditHint : Result := RxResConst_Ita.SSpeedbarEditHint;
     SParseSyntaxError : Result := RxResConst_Ita.SParseSyntaxError;
     SParseNotCramp : Result := RxResConst_Ita.SParseNotCramp;
     SParseDivideByZero : Result := RxResConst_Ita.SParseDivideByZero;
     SParseSqrError : Result := RxResConst_Ita.SParseSqrError;
     SParseLogError : Result := RxResConst_Ita.SParseLogError;
     SParseInvalidFloatOperation : Result := RxResConst_Ita.SParseInvalidFloatOperation;
     srRXControls : Result := RxResConst_Ita.srRXControls;
     srRXDBAware : Result := RxResConst_Ita.srRXDBAware;
     srRXTools : Result := RxResConst_Ita.srRXTools;
     srStorageDesigner : Result := RxResConst_Ita.srStorageDesigner;
     srProxyEditor : Result := RxResConst_Ita.srProxyEditor;
     srPageProxies : Result := RxResConst_Ita.srPageProxies;
     srProxyName : Result := RxResConst_Ita.srProxyName;
     srPageName : Result := RxResConst_Ita.srPageName;
     srSBItemNotCreate : Result := RxResConst_Ita.srSBItemNotCreate;
     srConfirmSBDelete : Result := RxResConst_Ita.srConfirmSBDelete;
     srSpeedbarDesigner : Result := RxResConst_Ita.srSpeedbarDesigner;
     srNewSectionName : Result := RxResConst_Ita.srNewSectionName;
     srEventNotCreate : Result := RxResConst_Ita.srEventNotCreate;
     srTimerDesigner : Result := RxResConst_Ita.srTimerDesigner;
     srTimerEvents : Result := RxResConst_Ita.srTimerEvents;
     srAniCurFilter : Result := RxResConst_Ita.srAniCurFilter;
     srEditPicture : Result := RxResConst_Ita.srEditPicture;
     srLoadAniCursor : Result := RxResConst_Ita.srLoadAniCursor;
     srLoadIcon : Result := RxResConst_Ita.srLoadIcon;
     srSaveImageList : Result := RxResConst_Ita.srSaveImageList;
     srCaptionDesigner : Result := RxResConst_Ita.srCaptionDesigner;
     srGradientCaptions : Result := RxResConst_Ita.srGradientCaptions;
     srBorrowStructure : Result := RxResConst_Ita.srBorrowStructure;
     ccTimeEditHours : Result := RxResConst_Ita.ccTimeEditHours;
     ccTimeEditMins : Result := RxResConst_Ita.ccTimeEditMins;
     ccTimeEditSecs : Result := RxResConst_Ita.ccTimeEditSecs;
     SColorBlack : Result := RxResConst_Ita.SColorBlack;
     SColorMaroon : Result := RxResConst_Ita.SColorMaroon;
     SColorGreen : Result := RxResConst_Ita.SColorGreen;
     SColorOlive : Result := RxResConst_Ita.SColorOlive;
     SColorNavy : Result := RxResConst_Ita.SColorNavy;
     SColorPurple : Result := RxResConst_Ita.SColorPurple;
     SColorTeal : Result := RxResConst_Ita.SColorTeal;
     SColorGray : Result := RxResConst_Ita.SColorGray;
     SColorSilver : Result := RxResConst_Ita.SColorSilver;
     SColorRed : Result := RxResConst_Ita.SColorRed;
     SColorLime : Result := RxResConst_Ita.SColorLime;
     SColorYellow : Result := RxResConst_Ita.SColorYellow;
     SColorBlue : Result := RxResConst_Ita.SColorBlue;
     SColorFuchsia : Result := RxResConst_Ita.SColorFuchsia;
     SColorAqua : Result := RxResConst_Ita.SColorAqua;
     SColorWhite : Result := RxResConst_Ita.SColorWhite;
     SColorScrollBar : Result := RxResConst_Ita.SColorScrollBar;
     SColorBackground : Result := RxResConst_Ita.SColorBackground;
     SColorActiveCaption : Result := RxResConst_Ita.SColorActiveCaption;
     SColorInactiveCaption : Result := RxResConst_Ita.SColorInactiveCaption;
     SColorMenu : Result := RxResConst_Ita.SColorMenu;
     SColorWindow : Result := RxResConst_Ita.SColorWindow;
     SColorWindowFrame : Result := RxResConst_Ita.SColorWindowFrame;
     SColorMenuText : Result := RxResConst_Ita.SColorMenuText;
     SColorWindowText : Result := RxResConst_Ita.SColorWindowText;
     SColorCaptionText : Result := RxResConst_Ita.SColorCaptionText;
     SColorActiveBorder : Result := RxResConst_Ita.SColorActiveBorder;
     SColorInactiveBorder : Result := RxResConst_Ita.SColorInactiveBorder;
     SColorAppWorkSpace : Result := RxResConst_Ita.SColorAppWorkSpace;
     SColorHighlight : Result := RxResConst_Ita.SColorHighlight;
     SColorHighlightText : Result := RxResConst_Ita.SColorHighlightText;
     SColorBtnFace : Result := RxResConst_Ita.SColorBtnFace;
     SColorBtnShadow : Result := RxResConst_Ita.SColorBtnShadow;
     SColorGrayText : Result := RxResConst_Ita.SColorGrayText;
     SColorBtnText : Result := RxResConst_Ita.SColorBtnText;
     SColorInactiveCaptionText : Result := RxResConst_Ita.SColorInactiveCaptionText;
     SColorBtnHighlight : Result := RxResConst_Ita.SColorBtnHighlight;
     SColor3DDkShadow : Result := RxResConst_Ita.SColor3DDkShadow;
     SColor3DLight : Result := RxResConst_Ita.SColor3DLight;
     SColorInfoText : Result := RxResConst_Ita.SColorInfoText;
     SColorInfoBk : Result := RxResConst_Ita.SColorInfoBk;
     SColorCream : Result := RxResConst_Ita.SColorCream;
     SColorMoneyGreen : Result := RxResConst_Ita.SColorMoneyGreen;
     SColorSkyBlue : Result := RxResConst_Ita.SColorSkyBlue;
     SColorNone : Result := RxResConst_Ita.SColorNone;
     SColorDefault : Result := RxResConst_Ita.SColorDefault;
     SColorCustom : Result := RxResConst_Ita.SColorCustom;
     SMemNoRecords : Result := RxResConst_Ita.SMemNoRecords;
     SInvalidFields : Result := RxResConst_Ita.SInvalidFields;
     SAutoSpeedbarMode : Result := RxResConst_Ita.SAutoSpeedbarMode;
    end;
  end;
  {$ENDIF}
  {$IFDEF RXLANG_Kor}
  function RxResConst_Kor_RxLoadStr(IDMsg: Integer): string;
  begin
    case IDMsg of
     SBrowse : Result := RxResConst_Kor.SBrowse;
     SDefaultFilter : Result := RxResConst_Kor.SDefaultFilter;
     SDateDlgTitle : Result := RxResConst_Kor.SDateDlgTitle;
     SNextYear : Result := RxResConst_Kor.SNextYear;
     SNextMonth : Result := RxResConst_Kor.SNextMonth;
     SPrevYear : Result := RxResConst_Kor.SPrevYear;
     SPrevMonth : Result := RxResConst_Kor.SPrevMonth;
     SNotImplemented : Result := RxResConst_Kor.SNotImplemented;
     SFileNotExec : Result := RxResConst_Kor.SFileNotExec;
     SLoadLibError : Result := RxResConst_Kor.SLoadLibError;
     SDetails : Result := RxResConst_Kor.SDetails;
     SDateOutOfRange : Result := RxResConst_Kor.SDateOutOfRange;
     SDateOutOfMin : Result := RxResConst_Kor.SDateOutOfMin;
     SDateOutOfMax : Result := RxResConst_Kor.SDateOutOfMax;
     SDateMinLimit : Result := RxResConst_Kor.SDateMinLimit;
     SDateMaxLimit : Result := RxResConst_Kor.SDateMaxLimit;
     SLocalDatabase : Result := RxResConst_Kor.SLocalDatabase;
     SRetryLogin : Result := RxResConst_Kor.SRetryLogin;
     SExprNotBoolean : Result := RxResConst_Kor.SExprNotBoolean;
     SExprBadNullTest : Result := RxResConst_Kor.SExprBadNullTest;
     SExprBadField : Result := RxResConst_Kor.SExprBadField;
     SCaptureFilter : Result := RxResConst_Kor.SCaptureFilter;
     SNotCaptureFilter : Result := RxResConst_Kor.SNotCaptureFilter;
     SInactiveData : Result := RxResConst_Kor.SInactiveData;
     SBrowseData : Result := RxResConst_Kor.SBrowseData;
     SEditData : Result := RxResConst_Kor.SEditData;
     SInsertData : Result := RxResConst_Kor.SInsertData;
     SSetKeyData : Result := RxResConst_Kor.SSetKeyData;
     SCalcFieldsData : Result := RxResConst_Kor.SCalcFieldsData;
     SRegistration : Result := RxResConst_Kor.SRegistration;
     SAppTitleLabel : Result := RxResConst_Kor.SAppTitleLabel;
     SHintLabel : Result := RxResConst_Kor.SHintLabel;
     SUserNameLabel : Result := RxResConst_Kor.SUserNameLabel;
     SPasswordLabel : Result := RxResConst_Kor.SPasswordLabel;
     SInvalidUserName : Result := RxResConst_Kor.SInvalidUserName;
     SChangePassword : Result := RxResConst_Kor.SChangePassword;
     SOldPasswordLabel : Result := RxResConst_Kor.SOldPasswordLabel;
     SNewPasswordLabel : Result := RxResConst_Kor.SNewPasswordLabel;
     SConfirmPasswordLabel : Result := RxResConst_Kor.SConfirmPasswordLabel;
     SPasswordChanged : Result := RxResConst_Kor.SPasswordChanged;
     SPasswordNotChanged : Result := RxResConst_Kor.SPasswordNotChanged;
     SPasswordsMismatch : Result := RxResConst_Kor.SPasswordsMismatch;
     SDBExceptCaption : Result := RxResConst_Kor.SDBExceptCaption;
     SBDEErrorLabel : Result := RxResConst_Kor.SBDEErrorLabel;
     SServerErrorLabel : Result := RxResConst_Kor.SServerErrorLabel;
     SErrorMsgLabel : Result := RxResConst_Kor.SErrorMsgLabel;
     SNextButton : Result := RxResConst_Kor.SNextButton;
     SPrevButton : Result := RxResConst_Kor.SPrevButton;
     SExprIncorrect : Result := RxResConst_Kor.SExprIncorrect;
     SExprTermination : Result := RxResConst_Kor.SExprTermination;
     SExprNameError : Result := RxResConst_Kor.SExprNameError;
     SExprStringError : Result := RxResConst_Kor.SExprStringError;
     SExprInvalidChar : Result := RxResConst_Kor.SExprInvalidChar;
     SExprNoRParen : Result := RxResConst_Kor.SExprNoRParen;
     SExprExpected : Result := RxResConst_Kor.SExprExpected;
     SExprBadCompare : Result := RxResConst_Kor.SExprBadCompare;
     SConfirmSave : Result := RxResConst_Kor.SConfirmSave;
     SDatabaseName : Result := RxResConst_Kor.SDatabaseName;
     SUnlockCaption : Result := RxResConst_Kor.SUnlockCaption;
     SUnlockHint : Result := RxResConst_Kor.SUnlockHint;
     SDeleteMultipleRecords : Result := RxResConst_Kor.SDeleteMultipleRecords;
     SGIFImage : Result := RxResConst_Kor.SGIFImage;
     SChangeGIFSize : Result := RxResConst_Kor.SChangeGIFSize;
     SNoGIFData : Result := RxResConst_Kor.SNoGIFData;
     SUnrecognizedGIFExt : Result := RxResConst_Kor.SUnrecognizedGIFExt;
     SWrongGIFColors : Result := RxResConst_Kor.SWrongGIFColors;
     SBadGIFCodeSize : Result := RxResConst_Kor.SBadGIFCodeSize;
     SGIFDecodeError : Result := RxResConst_Kor.SGIFDecodeError;
     SGIFEncodeError : Result := RxResConst_Kor.SGIFEncodeError;
     SGIFVersion : Result := RxResConst_Kor.SGIFVersion;
     SDualListSrcCaption : Result := RxResConst_Kor.SDualListSrcCaption;
     SDualListDestCaption : Result := RxResConst_Kor.SDualListDestCaption;
     SClipbrdUnknown : Result := RxResConst_Kor.SClipbrdUnknown;
     SClipbrdEmpty : Result := RxResConst_Kor.SClipbrdEmpty;
     SCustomizeSpeedbar : Result := RxResConst_Kor.SCustomizeSpeedbar;
     SAvailButtons : Result := RxResConst_Kor.SAvailButtons;
     SSpeedbarCategories : Result := RxResConst_Kor.SSpeedbarCategories;
     SSpeedbarEditHint : Result := RxResConst_Kor.SSpeedbarEditHint;
     SParseSyntaxError : Result := RxResConst_Kor.SParseSyntaxError;
     SParseNotCramp : Result := RxResConst_Kor.SParseNotCramp;
     SParseDivideByZero : Result := RxResConst_Kor.SParseDivideByZero;
     SParseSqrError : Result := RxResConst_Kor.SParseSqrError;
     SParseLogError : Result := RxResConst_Kor.SParseLogError;
     SParseInvalidFloatOperation : Result := RxResConst_Kor.SParseInvalidFloatOperation;
     srRXControls : Result := RxResConst_Kor.srRXControls;
     srRXDBAware : Result := RxResConst_Kor.srRXDBAware;
     srRXTools : Result := RxResConst_Kor.srRXTools;
     srStorageDesigner : Result := RxResConst_Kor.srStorageDesigner;
     srProxyEditor : Result := RxResConst_Kor.srProxyEditor;
     srPageProxies : Result := RxResConst_Kor.srPageProxies;
     srProxyName : Result := RxResConst_Kor.srProxyName;
     srPageName : Result := RxResConst_Kor.srPageName;
     srSBItemNotCreate : Result := RxResConst_Kor.srSBItemNotCreate;
     srConfirmSBDelete : Result := RxResConst_Kor.srConfirmSBDelete;
     srSpeedbarDesigner : Result := RxResConst_Kor.srSpeedbarDesigner;
     srNewSectionName : Result := RxResConst_Kor.srNewSectionName;
     srEventNotCreate : Result := RxResConst_Kor.srEventNotCreate;
     srTimerDesigner : Result := RxResConst_Kor.srTimerDesigner;
     srTimerEvents : Result := RxResConst_Kor.srTimerEvents;
     srAniCurFilter : Result := RxResConst_Kor.srAniCurFilter;
     srEditPicture : Result := RxResConst_Kor.srEditPicture;
     srLoadAniCursor : Result := RxResConst_Kor.srLoadAniCursor;
     srLoadIcon : Result := RxResConst_Kor.srLoadIcon;
     srSaveImageList : Result := RxResConst_Kor.srSaveImageList;
     srCaptionDesigner : Result := RxResConst_Kor.srCaptionDesigner;
     srGradientCaptions : Result := RxResConst_Kor.srGradientCaptions;
     srBorrowStructure : Result := RxResConst_Kor.srBorrowStructure;
     ccTimeEditHours : Result := RxResConst_Kor.ccTimeEditHours;
     ccTimeEditMins : Result := RxResConst_Kor.ccTimeEditMins;
     ccTimeEditSecs : Result := RxResConst_Kor.ccTimeEditSecs;
     SColorBlack : Result := RxResConst_Kor.SColorBlack;
     SColorMaroon : Result := RxResConst_Kor.SColorMaroon;
     SColorGreen : Result := RxResConst_Kor.SColorGreen;
     SColorOlive : Result := RxResConst_Kor.SColorOlive;
     SColorNavy : Result := RxResConst_Kor.SColorNavy;
     SColorPurple : Result := RxResConst_Kor.SColorPurple;
     SColorTeal : Result := RxResConst_Kor.SColorTeal;
     SColorGray : Result := RxResConst_Kor.SColorGray;
     SColorSilver : Result := RxResConst_Kor.SColorSilver;
     SColorRed : Result := RxResConst_Kor.SColorRed;
     SColorLime : Result := RxResConst_Kor.SColorLime;
     SColorYellow : Result := RxResConst_Kor.SColorYellow;
     SColorBlue : Result := RxResConst_Kor.SColorBlue;
     SColorFuchsia : Result := RxResConst_Kor.SColorFuchsia;
     SColorAqua : Result := RxResConst_Kor.SColorAqua;
     SColorWhite : Result := RxResConst_Kor.SColorWhite;
     SColorScrollBar : Result := RxResConst_Kor.SColorScrollBar;
     SColorBackground : Result := RxResConst_Kor.SColorBackground;
     SColorActiveCaption : Result := RxResConst_Kor.SColorActiveCaption;
     SColorInactiveCaption : Result := RxResConst_Kor.SColorInactiveCaption;
     SColorMenu : Result := RxResConst_Kor.SColorMenu;
     SColorWindow : Result := RxResConst_Kor.SColorWindow;
     SColorWindowFrame : Result := RxResConst_Kor.SColorWindowFrame;
     SColorMenuText : Result := RxResConst_Kor.SColorMenuText;
     SColorWindowText : Result := RxResConst_Kor.SColorWindowText;
     SColorCaptionText : Result := RxResConst_Kor.SColorCaptionText;
     SColorActiveBorder : Result := RxResConst_Kor.SColorActiveBorder;
     SColorInactiveBorder : Result := RxResConst_Kor.SColorInactiveBorder;
     SColorAppWorkSpace : Result := RxResConst_Kor.SColorAppWorkSpace;
     SColorHighlight : Result := RxResConst_Kor.SColorHighlight;
     SColorHighlightText : Result := RxResConst_Kor.SColorHighlightText;
     SColorBtnFace : Result := RxResConst_Kor.SColorBtnFace;
     SColorBtnShadow : Result := RxResConst_Kor.SColorBtnShadow;
     SColorGrayText : Result := RxResConst_Kor.SColorGrayText;
     SColorBtnText : Result := RxResConst_Kor.SColorBtnText;
     SColorInactiveCaptionText : Result := RxResConst_Kor.SColorInactiveCaptionText;
     SColorBtnHighlight : Result := RxResConst_Kor.SColorBtnHighlight;
     SColor3DDkShadow : Result := RxResConst_Kor.SColor3DDkShadow;
     SColor3DLight : Result := RxResConst_Kor.SColor3DLight;
     SColorInfoText : Result := RxResConst_Kor.SColorInfoText;
     SColorInfoBk : Result := RxResConst_Kor.SColorInfoBk;
     SColorCream : Result := RxResConst_Kor.SColorCream;
     SColorMoneyGreen : Result := RxResConst_Kor.SColorMoneyGreen;
     SColorSkyBlue : Result := RxResConst_Kor.SColorSkyBlue;
     SColorNone : Result := RxResConst_Kor.SColorNone;
     SColorDefault : Result := RxResConst_Kor.SColorDefault;
     SColorCustom : Result := RxResConst_Kor.SColorCustom;
     SMemNoRecords : Result := RxResConst_Kor.SMemNoRecords;
     SInvalidFields : Result := RxResConst_Kor.SInvalidFields;
     SAutoSpeedbarMode : Result := RxResConst_Kor.SAutoSpeedbarMode;
    end;
  end;
  {$ENDIF}
  {$IFDEF RXLANG_Mex}
  function RxResConst_Mex_RxLoadStr(IDMsg: Integer): string;
  begin
    case IDMsg of
     SBrowse : Result := RxResConst_Mex.SBrowse;
     SDefaultFilter : Result := RxResConst_Mex.SDefaultFilter;
     SDateDlgTitle : Result := RxResConst_Mex.SDateDlgTitle;
     SNextYear : Result := RxResConst_Mex.SNextYear;
     SNextMonth : Result := RxResConst_Mex.SNextMonth;
     SPrevYear : Result := RxResConst_Mex.SPrevYear;
     SPrevMonth : Result := RxResConst_Mex.SPrevMonth;
     SNotImplemented : Result := RxResConst_Mex.SNotImplemented;
     SFileNotExec : Result := RxResConst_Mex.SFileNotExec;
     SLoadLibError : Result := RxResConst_Mex.SLoadLibError;
     SDetails : Result := RxResConst_Mex.SDetails;
     SDateOutOfRange : Result := RxResConst_Mex.SDateOutOfRange;
     SDateOutOfMin : Result := RxResConst_Mex.SDateOutOfMin;
     SDateOutOfMax : Result := RxResConst_Mex.SDateOutOfMax;
     SDateMinLimit : Result := RxResConst_Mex.SDateMinLimit;
     SDateMaxLimit : Result := RxResConst_Mex.SDateMaxLimit;
     SLocalDatabase : Result := RxResConst_Mex.SLocalDatabase;
     SRetryLogin : Result := RxResConst_Mex.SRetryLogin;
     SExprNotBoolean : Result := RxResConst_Mex.SExprNotBoolean;
     SExprBadNullTest : Result := RxResConst_Mex.SExprBadNullTest;
     SExprBadField : Result := RxResConst_Mex.SExprBadField;
     SCaptureFilter : Result := RxResConst_Mex.SCaptureFilter;
     SNotCaptureFilter : Result := RxResConst_Mex.SNotCaptureFilter;
     SInactiveData : Result := RxResConst_Mex.SInactiveData;
     SBrowseData : Result := RxResConst_Mex.SBrowseData;
     SEditData : Result := RxResConst_Mex.SEditData;
     SInsertData : Result := RxResConst_Mex.SInsertData;
     SSetKeyData : Result := RxResConst_Mex.SSetKeyData;
     SCalcFieldsData : Result := RxResConst_Mex.SCalcFieldsData;
     SRegistration : Result := RxResConst_Mex.SRegistration;
     SAppTitleLabel : Result := RxResConst_Mex.SAppTitleLabel;
     SHintLabel : Result := RxResConst_Mex.SHintLabel;
     SUserNameLabel : Result := RxResConst_Mex.SUserNameLabel;
     SPasswordLabel : Result := RxResConst_Mex.SPasswordLabel;
     SInvalidUserName : Result := RxResConst_Mex.SInvalidUserName;
     SChangePassword : Result := RxResConst_Mex.SChangePassword;
     SOldPasswordLabel : Result := RxResConst_Mex.SOldPasswordLabel;
     SNewPasswordLabel : Result := RxResConst_Mex.SNewPasswordLabel;
     SConfirmPasswordLabel : Result := RxResConst_Mex.SConfirmPasswordLabel;
     SPasswordChanged : Result := RxResConst_Mex.SPasswordChanged;
     SPasswordNotChanged : Result := RxResConst_Mex.SPasswordNotChanged;
     SPasswordsMismatch : Result := RxResConst_Mex.SPasswordsMismatch;
     SDBExceptCaption : Result := RxResConst_Mex.SDBExceptCaption;
     SBDEErrorLabel : Result := RxResConst_Mex.SBDEErrorLabel;
     SServerErrorLabel : Result := RxResConst_Mex.SServerErrorLabel;
     SErrorMsgLabel : Result := RxResConst_Mex.SErrorMsgLabel;
     SNextButton : Result := RxResConst_Mex.SNextButton;
     SPrevButton : Result := RxResConst_Mex.SPrevButton;
     SExprIncorrect : Result := RxResConst_Mex.SExprIncorrect;
     SExprTermination : Result := RxResConst_Mex.SExprTermination;
     SExprNameError : Result := RxResConst_Mex.SExprNameError;
     SExprStringError : Result := RxResConst_Mex.SExprStringError;
     SExprInvalidChar : Result := RxResConst_Mex.SExprInvalidChar;
     SExprNoRParen : Result := RxResConst_Mex.SExprNoRParen;
     SExprExpected : Result := RxResConst_Mex.SExprExpected;
     SExprBadCompare : Result := RxResConst_Mex.SExprBadCompare;
     SConfirmSave : Result := RxResConst_Mex.SConfirmSave;
     SDatabaseName : Result := RxResConst_Mex.SDatabaseName;
     SUnlockCaption : Result := RxResConst_Mex.SUnlockCaption;
     SUnlockHint : Result := RxResConst_Mex.SUnlockHint;
     SDeleteMultipleRecords : Result := RxResConst_Mex.SDeleteMultipleRecords;
     SGIFImage : Result := RxResConst_Mex.SGIFImage;
     SChangeGIFSize : Result := RxResConst_Mex.SChangeGIFSize;
     SNoGIFData : Result := RxResConst_Mex.SNoGIFData;
     SUnrecognizedGIFExt : Result := RxResConst_Mex.SUnrecognizedGIFExt;
     SWrongGIFColors : Result := RxResConst_Mex.SWrongGIFColors;
     SBadGIFCodeSize : Result := RxResConst_Mex.SBadGIFCodeSize;
     SGIFDecodeError : Result := RxResConst_Mex.SGIFDecodeError;
     SGIFEncodeError : Result := RxResConst_Mex.SGIFEncodeError;
     SGIFVersion : Result := RxResConst_Mex.SGIFVersion;
     SDualListSrcCaption : Result := RxResConst_Mex.SDualListSrcCaption;
     SDualListDestCaption : Result := RxResConst_Mex.SDualListDestCaption;
     SClipbrdUnknown : Result := RxResConst_Mex.SClipbrdUnknown;
     SClipbrdEmpty : Result := RxResConst_Mex.SClipbrdEmpty;
     SCustomizeSpeedbar : Result := RxResConst_Mex.SCustomizeSpeedbar;
     SAvailButtons : Result := RxResConst_Mex.SAvailButtons;
     SSpeedbarCategories : Result := RxResConst_Mex.SSpeedbarCategories;
     SSpeedbarEditHint : Result := RxResConst_Mex.SSpeedbarEditHint;
     SParseSyntaxError : Result := RxResConst_Mex.SParseSyntaxError;
     SParseNotCramp : Result := RxResConst_Mex.SParseNotCramp;
     SParseDivideByZero : Result := RxResConst_Mex.SParseDivideByZero;
     SParseSqrError : Result := RxResConst_Mex.SParseSqrError;
     SParseLogError : Result := RxResConst_Mex.SParseLogError;
     SParseInvalidFloatOperation : Result := RxResConst_Mex.SParseInvalidFloatOperation;
     srRXControls : Result := RxResConst_Mex.srRXControls;
     srRXDBAware : Result := RxResConst_Mex.srRXDBAware;
     srRXTools : Result := RxResConst_Mex.srRXTools;
     srStorageDesigner : Result := RxResConst_Mex.srStorageDesigner;
     srProxyEditor : Result := RxResConst_Mex.srProxyEditor;
     srPageProxies : Result := RxResConst_Mex.srPageProxies;
     srProxyName : Result := RxResConst_Mex.srProxyName;
     srPageName : Result := RxResConst_Mex.srPageName;
     srSBItemNotCreate : Result := RxResConst_Mex.srSBItemNotCreate;
     srConfirmSBDelete : Result := RxResConst_Mex.srConfirmSBDelete;
     srSpeedbarDesigner : Result := RxResConst_Mex.srSpeedbarDesigner;
     srNewSectionName : Result := RxResConst_Mex.srNewSectionName;
     srEventNotCreate : Result := RxResConst_Mex.srEventNotCreate;
     srTimerDesigner : Result := RxResConst_Mex.srTimerDesigner;
     srTimerEvents : Result := RxResConst_Mex.srTimerEvents;
     srAniCurFilter : Result := RxResConst_Mex.srAniCurFilter;
     srEditPicture : Result := RxResConst_Mex.srEditPicture;
     srLoadAniCursor : Result := RxResConst_Mex.srLoadAniCursor;
     srLoadIcon : Result := RxResConst_Mex.srLoadIcon;
     srSaveImageList : Result := RxResConst_Mex.srSaveImageList;
     srCaptionDesigner : Result := RxResConst_Mex.srCaptionDesigner;
     srGradientCaptions : Result := RxResConst_Mex.srGradientCaptions;
     srBorrowStructure : Result := RxResConst_Mex.srBorrowStructure;
     ccTimeEditHours : Result := RxResConst_Mex.ccTimeEditHours;
     ccTimeEditMins : Result := RxResConst_Mex.ccTimeEditMins;
     ccTimeEditSecs : Result := RxResConst_Mex.ccTimeEditSecs;
     SColorBlack : Result := RxResConst_Mex.SColorBlack;
     SColorMaroon : Result := RxResConst_Mex.SColorMaroon;
     SColorGreen : Result := RxResConst_Mex.SColorGreen;
     SColorOlive : Result := RxResConst_Mex.SColorOlive;
     SColorNavy : Result := RxResConst_Mex.SColorNavy;
     SColorPurple : Result := RxResConst_Mex.SColorPurple;
     SColorTeal : Result := RxResConst_Mex.SColorTeal;
     SColorGray : Result := RxResConst_Mex.SColorGray;
     SColorSilver : Result := RxResConst_Mex.SColorSilver;
     SColorRed : Result := RxResConst_Mex.SColorRed;
     SColorLime : Result := RxResConst_Mex.SColorLime;
     SColorYellow : Result := RxResConst_Mex.SColorYellow;
     SColorBlue : Result := RxResConst_Mex.SColorBlue;
     SColorFuchsia : Result := RxResConst_Mex.SColorFuchsia;
     SColorAqua : Result := RxResConst_Mex.SColorAqua;
     SColorWhite : Result := RxResConst_Mex.SColorWhite;
     SColorScrollBar : Result := RxResConst_Mex.SColorScrollBar;
     SColorBackground : Result := RxResConst_Mex.SColorBackground;
     SColorActiveCaption : Result := RxResConst_Mex.SColorActiveCaption;
     SColorInactiveCaption : Result := RxResConst_Mex.SColorInactiveCaption;
     SColorMenu : Result := RxResConst_Mex.SColorMenu;
     SColorWindow : Result := RxResConst_Mex.SColorWindow;
     SColorWindowFrame : Result := RxResConst_Mex.SColorWindowFrame;
     SColorMenuText : Result := RxResConst_Mex.SColorMenuText;
     SColorWindowText : Result := RxResConst_Mex.SColorWindowText;
     SColorCaptionText : Result := RxResConst_Mex.SColorCaptionText;
     SColorActiveBorder : Result := RxResConst_Mex.SColorActiveBorder;
     SColorInactiveBorder : Result := RxResConst_Mex.SColorInactiveBorder;
     SColorAppWorkSpace : Result := RxResConst_Mex.SColorAppWorkSpace;
     SColorHighlight : Result := RxResConst_Mex.SColorHighlight;
     SColorHighlightText : Result := RxResConst_Mex.SColorHighlightText;
     SColorBtnFace : Result := RxResConst_Mex.SColorBtnFace;
     SColorBtnShadow : Result := RxResConst_Mex.SColorBtnShadow;
     SColorGrayText : Result := RxResConst_Mex.SColorGrayText;
     SColorBtnText : Result := RxResConst_Mex.SColorBtnText;
     SColorInactiveCaptionText : Result := RxResConst_Mex.SColorInactiveCaptionText;
     SColorBtnHighlight : Result := RxResConst_Mex.SColorBtnHighlight;
     SColor3DDkShadow : Result := RxResConst_Mex.SColor3DDkShadow;
     SColor3DLight : Result := RxResConst_Mex.SColor3DLight;
     SColorInfoText : Result := RxResConst_Mex.SColorInfoText;
     SColorInfoBk : Result := RxResConst_Mex.SColorInfoBk;
     SColorCream : Result := RxResConst_Mex.SColorCream;
     SColorMoneyGreen : Result := RxResConst_Mex.SColorMoneyGreen;
     SColorSkyBlue : Result := RxResConst_Mex.SColorSkyBlue;
     SColorNone : Result := RxResConst_Mex.SColorNone;
     SColorDefault : Result := RxResConst_Mex.SColorDefault;
     SColorCustom : Result := RxResConst_Mex.SColorCustom;
     SMemNoRecords : Result := RxResConst_Mex.SMemNoRecords;
     SInvalidFields : Result := RxResConst_Mex.SInvalidFields;
     SAutoSpeedbarMode : Result := RxResConst_Mex.SAutoSpeedbarMode;
    end;
  end;
  {$ENDIF}
  {$IFDEF RXLANG_Pol}
  function RxResConst_Pol_RxLoadStr(IDMsg: Integer): string;
  begin
    case IDMsg of
     SBrowse : Result := RxResConst_Pol.SBrowse;
     SDefaultFilter : Result := RxResConst_Pol.SDefaultFilter;
     SDateDlgTitle : Result := RxResConst_Pol.SDateDlgTitle;
     SNextYear : Result := RxResConst_Pol.SNextYear;
     SNextMonth : Result := RxResConst_Pol.SNextMonth;
     SPrevYear : Result := RxResConst_Pol.SPrevYear;
     SPrevMonth : Result := RxResConst_Pol.SPrevMonth;
     SNotImplemented : Result := RxResConst_Pol.SNotImplemented;
     SFileNotExec : Result := RxResConst_Pol.SFileNotExec;
     SLoadLibError : Result := RxResConst_Pol.SLoadLibError;
     SDetails : Result := RxResConst_Pol.SDetails;
     SDateOutOfRange : Result := RxResConst_Pol.SDateOutOfRange;
     SDateOutOfMin : Result := RxResConst_Pol.SDateOutOfMin;
     SDateOutOfMax : Result := RxResConst_Pol.SDateOutOfMax;
     SDateMinLimit : Result := RxResConst_Pol.SDateMinLimit;
     SDateMaxLimit : Result := RxResConst_Pol.SDateMaxLimit;
     SLocalDatabase : Result := RxResConst_Pol.SLocalDatabase;
     SRetryLogin : Result := RxResConst_Pol.SRetryLogin;
     SExprNotBoolean : Result := RxResConst_Pol.SExprNotBoolean;
     SExprBadNullTest : Result := RxResConst_Pol.SExprBadNullTest;
     SExprBadField : Result := RxResConst_Pol.SExprBadField;
     SCaptureFilter : Result := RxResConst_Pol.SCaptureFilter;
     SNotCaptureFilter : Result := RxResConst_Pol.SNotCaptureFilter;
     SInactiveData : Result := RxResConst_Pol.SInactiveData;
     SBrowseData : Result := RxResConst_Pol.SBrowseData;
     SEditData : Result := RxResConst_Pol.SEditData;
     SInsertData : Result := RxResConst_Pol.SInsertData;
     SSetKeyData : Result := RxResConst_Pol.SSetKeyData;
     SCalcFieldsData : Result := RxResConst_Pol.SCalcFieldsData;
     SRegistration : Result := RxResConst_Pol.SRegistration;
     SAppTitleLabel : Result := RxResConst_Pol.SAppTitleLabel;
     SHintLabel : Result := RxResConst_Pol.SHintLabel;
     SUserNameLabel : Result := RxResConst_Pol.SUserNameLabel;
     SPasswordLabel : Result := RxResConst_Pol.SPasswordLabel;
     SInvalidUserName : Result := RxResConst_Pol.SInvalidUserName;
     SChangePassword : Result := RxResConst_Pol.SChangePassword;
     SOldPasswordLabel : Result := RxResConst_Pol.SOldPasswordLabel;
     SNewPasswordLabel : Result := RxResConst_Pol.SNewPasswordLabel;
     SConfirmPasswordLabel : Result := RxResConst_Pol.SConfirmPasswordLabel;
     SPasswordChanged : Result := RxResConst_Pol.SPasswordChanged;
     SPasswordNotChanged : Result := RxResConst_Pol.SPasswordNotChanged;
     SPasswordsMismatch : Result := RxResConst_Pol.SPasswordsMismatch;
     SDBExceptCaption : Result := RxResConst_Pol.SDBExceptCaption;
     SBDEErrorLabel : Result := RxResConst_Pol.SBDEErrorLabel;
     SServerErrorLabel : Result := RxResConst_Pol.SServerErrorLabel;
     SErrorMsgLabel : Result := RxResConst_Pol.SErrorMsgLabel;
     SNextButton : Result := RxResConst_Pol.SNextButton;
     SPrevButton : Result := RxResConst_Pol.SPrevButton;
     SExprIncorrect : Result := RxResConst_Pol.SExprIncorrect;
     SExprTermination : Result := RxResConst_Pol.SExprTermination;
     SExprNameError : Result := RxResConst_Pol.SExprNameError;
     SExprStringError : Result := RxResConst_Pol.SExprStringError;
     SExprInvalidChar : Result := RxResConst_Pol.SExprInvalidChar;
     SExprNoRParen : Result := RxResConst_Pol.SExprNoRParen;
     SExprExpected : Result := RxResConst_Pol.SExprExpected;
     SExprBadCompare : Result := RxResConst_Pol.SExprBadCompare;
     SConfirmSave : Result := RxResConst_Pol.SConfirmSave;
     SDatabaseName : Result := RxResConst_Pol.SDatabaseName;
     SUnlockCaption : Result := RxResConst_Pol.SUnlockCaption;
     SUnlockHint : Result := RxResConst_Pol.SUnlockHint;
     SDeleteMultipleRecords : Result := RxResConst_Pol.SDeleteMultipleRecords;
     SGIFImage : Result := RxResConst_Pol.SGIFImage;
     SChangeGIFSize : Result := RxResConst_Pol.SChangeGIFSize;
     SNoGIFData : Result := RxResConst_Pol.SNoGIFData;
     SUnrecognizedGIFExt : Result := RxResConst_Pol.SUnrecognizedGIFExt;
     SWrongGIFColors : Result := RxResConst_Pol.SWrongGIFColors;
     SBadGIFCodeSize : Result := RxResConst_Pol.SBadGIFCodeSize;
     SGIFDecodeError : Result := RxResConst_Pol.SGIFDecodeError;
     SGIFEncodeError : Result := RxResConst_Pol.SGIFEncodeError;
     SGIFVersion : Result := RxResConst_Pol.SGIFVersion;
     SDualListSrcCaption : Result := RxResConst_Pol.SDualListSrcCaption;
     SDualListDestCaption : Result := RxResConst_Pol.SDualListDestCaption;
     SClipbrdUnknown : Result := RxResConst_Pol.SClipbrdUnknown;
     SClipbrdEmpty : Result := RxResConst_Pol.SClipbrdEmpty;
     SCustomizeSpeedbar : Result := RxResConst_Pol.SCustomizeSpeedbar;
     SAvailButtons : Result := RxResConst_Pol.SAvailButtons;
     SSpeedbarCategories : Result := RxResConst_Pol.SSpeedbarCategories;
     SSpeedbarEditHint : Result := RxResConst_Pol.SSpeedbarEditHint;
     SParseSyntaxError : Result := RxResConst_Pol.SParseSyntaxError;
     SParseNotCramp : Result := RxResConst_Pol.SParseNotCramp;
     SParseDivideByZero : Result := RxResConst_Pol.SParseDivideByZero;
     SParseSqrError : Result := RxResConst_Pol.SParseSqrError;
     SParseLogError : Result := RxResConst_Pol.SParseLogError;
     SParseInvalidFloatOperation : Result := RxResConst_Pol.SParseInvalidFloatOperation;
     srRXControls : Result := RxResConst_Pol.srRXControls;
     srRXDBAware : Result := RxResConst_Pol.srRXDBAware;
     srRXTools : Result := RxResConst_Pol.srRXTools;
     srStorageDesigner : Result := RxResConst_Pol.srStorageDesigner;
     srProxyEditor : Result := RxResConst_Pol.srProxyEditor;
     srPageProxies : Result := RxResConst_Pol.srPageProxies;
     srProxyName : Result := RxResConst_Pol.srProxyName;
     srPageName : Result := RxResConst_Pol.srPageName;
     srSBItemNotCreate : Result := RxResConst_Pol.srSBItemNotCreate;
     srConfirmSBDelete : Result := RxResConst_Pol.srConfirmSBDelete;
     srSpeedbarDesigner : Result := RxResConst_Pol.srSpeedbarDesigner;
     srNewSectionName : Result := RxResConst_Pol.srNewSectionName;
     srEventNotCreate : Result := RxResConst_Pol.srEventNotCreate;
     srTimerDesigner : Result := RxResConst_Pol.srTimerDesigner;
     srTimerEvents : Result := RxResConst_Pol.srTimerEvents;
     srAniCurFilter : Result := RxResConst_Pol.srAniCurFilter;
     srEditPicture : Result := RxResConst_Pol.srEditPicture;
     srLoadAniCursor : Result := RxResConst_Pol.srLoadAniCursor;
     srLoadIcon : Result := RxResConst_Pol.srLoadIcon;
     srSaveImageList : Result := RxResConst_Pol.srSaveImageList;
     srCaptionDesigner : Result := RxResConst_Pol.srCaptionDesigner;
     srGradientCaptions : Result := RxResConst_Pol.srGradientCaptions;
     srBorrowStructure : Result := RxResConst_Pol.srBorrowStructure;
     ccTimeEditHours : Result := RxResConst_Pol.ccTimeEditHours;
     ccTimeEditMins : Result := RxResConst_Pol.ccTimeEditMins;
     ccTimeEditSecs : Result := RxResConst_Pol.ccTimeEditSecs;
     SColorBlack : Result := RxResConst_Pol.SColorBlack;
     SColorMaroon : Result := RxResConst_Pol.SColorMaroon;
     SColorGreen : Result := RxResConst_Pol.SColorGreen;
     SColorOlive : Result := RxResConst_Pol.SColorOlive;
     SColorNavy : Result := RxResConst_Pol.SColorNavy;
     SColorPurple : Result := RxResConst_Pol.SColorPurple;
     SColorTeal : Result := RxResConst_Pol.SColorTeal;
     SColorGray : Result := RxResConst_Pol.SColorGray;
     SColorSilver : Result := RxResConst_Pol.SColorSilver;
     SColorRed : Result := RxResConst_Pol.SColorRed;
     SColorLime : Result := RxResConst_Pol.SColorLime;
     SColorYellow : Result := RxResConst_Pol.SColorYellow;
     SColorBlue : Result := RxResConst_Pol.SColorBlue;
     SColorFuchsia : Result := RxResConst_Pol.SColorFuchsia;
     SColorAqua : Result := RxResConst_Pol.SColorAqua;
     SColorWhite : Result := RxResConst_Pol.SColorWhite;
     SColorScrollBar : Result := RxResConst_Pol.SColorScrollBar;
     SColorBackground : Result := RxResConst_Pol.SColorBackground;
     SColorActiveCaption : Result := RxResConst_Pol.SColorActiveCaption;
     SColorInactiveCaption : Result := RxResConst_Pol.SColorInactiveCaption;
     SColorMenu : Result := RxResConst_Pol.SColorMenu;
     SColorWindow : Result := RxResConst_Pol.SColorWindow;
     SColorWindowFrame : Result := RxResConst_Pol.SColorWindowFrame;
     SColorMenuText : Result := RxResConst_Pol.SColorMenuText;
     SColorWindowText : Result := RxResConst_Pol.SColorWindowText;
     SColorCaptionText : Result := RxResConst_Pol.SColorCaptionText;
     SColorActiveBorder : Result := RxResConst_Pol.SColorActiveBorder;
     SColorInactiveBorder : Result := RxResConst_Pol.SColorInactiveBorder;
     SColorAppWorkSpace : Result := RxResConst_Pol.SColorAppWorkSpace;
     SColorHighlight : Result := RxResConst_Pol.SColorHighlight;
     SColorHighlightText : Result := RxResConst_Pol.SColorHighlightText;
     SColorBtnFace : Result := RxResConst_Pol.SColorBtnFace;
     SColorBtnShadow : Result := RxResConst_Pol.SColorBtnShadow;
     SColorGrayText : Result := RxResConst_Pol.SColorGrayText;
     SColorBtnText : Result := RxResConst_Pol.SColorBtnText;
     SColorInactiveCaptionText : Result := RxResConst_Pol.SColorInactiveCaptionText;
     SColorBtnHighlight : Result := RxResConst_Pol.SColorBtnHighlight;
     SColor3DDkShadow : Result := RxResConst_Pol.SColor3DDkShadow;
     SColor3DLight : Result := RxResConst_Pol.SColor3DLight;
     SColorInfoText : Result := RxResConst_Pol.SColorInfoText;
     SColorInfoBk : Result := RxResConst_Pol.SColorInfoBk;
     SColorCream : Result := RxResConst_Pol.SColorCream;
     SColorMoneyGreen : Result := RxResConst_Pol.SColorMoneyGreen;
     SColorSkyBlue : Result := RxResConst_Pol.SColorSkyBlue;
     SColorNone : Result := RxResConst_Pol.SColorNone;
     SColorDefault : Result := RxResConst_Pol.SColorDefault;
     SColorCustom : Result := RxResConst_Pol.SColorCustom;
     SMemNoRecords : Result := RxResConst_Pol.SMemNoRecords;
     SInvalidFields : Result := RxResConst_Pol.SInvalidFields;
     SAutoSpeedbarMode : Result := RxResConst_Pol.SAutoSpeedbarMode;
    end;
  end;
  {$ENDIF}
  {$IFDEF RXLANG_Por}
  function RxResConst_Por_RxLoadStr(IDMsg: Integer): string;
  begin
    case IDMsg of
     SBrowse : Result := RxResConst_Por.SBrowse;
     SDefaultFilter : Result := RxResConst_Por.SDefaultFilter;
     SDateDlgTitle : Result := RxResConst_Por.SDateDlgTitle;
     SNextYear : Result := RxResConst_Por.SNextYear;
     SNextMonth : Result := RxResConst_Por.SNextMonth;
     SPrevYear : Result := RxResConst_Por.SPrevYear;
     SPrevMonth : Result := RxResConst_Por.SPrevMonth;
     SNotImplemented : Result := RxResConst_Por.SNotImplemented;
     SFileNotExec : Result := RxResConst_Por.SFileNotExec;
     SLoadLibError : Result := RxResConst_Por.SLoadLibError;
     SDetails : Result := RxResConst_Por.SDetails;
     SDateOutOfRange : Result := RxResConst_Por.SDateOutOfRange;
     SDateOutOfMin : Result := RxResConst_Por.SDateOutOfMin;
     SDateOutOfMax : Result := RxResConst_Por.SDateOutOfMax;
     SDateMinLimit : Result := RxResConst_Por.SDateMinLimit;
     SDateMaxLimit : Result := RxResConst_Por.SDateMaxLimit;
     SLocalDatabase : Result := RxResConst_Por.SLocalDatabase;
     SRetryLogin : Result := RxResConst_Por.SRetryLogin;
     SExprNotBoolean : Result := RxResConst_Por.SExprNotBoolean;
     SExprBadNullTest : Result := RxResConst_Por.SExprBadNullTest;
     SExprBadField : Result := RxResConst_Por.SExprBadField;
     SCaptureFilter : Result := RxResConst_Por.SCaptureFilter;
     SNotCaptureFilter : Result := RxResConst_Por.SNotCaptureFilter;
     SInactiveData : Result := RxResConst_Por.SInactiveData;
     SBrowseData : Result := RxResConst_Por.SBrowseData;
     SEditData : Result := RxResConst_Por.SEditData;
     SInsertData : Result := RxResConst_Por.SInsertData;
     SSetKeyData : Result := RxResConst_Por.SSetKeyData;
     SCalcFieldsData : Result := RxResConst_Por.SCalcFieldsData;
     SRegistration : Result := RxResConst_Por.SRegistration;
     SAppTitleLabel : Result := RxResConst_Por.SAppTitleLabel;
     SHintLabel : Result := RxResConst_Por.SHintLabel;
     SUserNameLabel : Result := RxResConst_Por.SUserNameLabel;
     SPasswordLabel : Result := RxResConst_Por.SPasswordLabel;
     SInvalidUserName : Result := RxResConst_Por.SInvalidUserName;
     SChangePassword : Result := RxResConst_Por.SChangePassword;
     SOldPasswordLabel : Result := RxResConst_Por.SOldPasswordLabel;
     SNewPasswordLabel : Result := RxResConst_Por.SNewPasswordLabel;
     SConfirmPasswordLabel : Result := RxResConst_Por.SConfirmPasswordLabel;
     SPasswordChanged : Result := RxResConst_Por.SPasswordChanged;
     SPasswordNotChanged : Result := RxResConst_Por.SPasswordNotChanged;
     SPasswordsMismatch : Result := RxResConst_Por.SPasswordsMismatch;
     SDBExceptCaption : Result := RxResConst_Por.SDBExceptCaption;
     SBDEErrorLabel : Result := RxResConst_Por.SBDEErrorLabel;
     SServerErrorLabel : Result := RxResConst_Por.SServerErrorLabel;
     SErrorMsgLabel : Result := RxResConst_Por.SErrorMsgLabel;
     SNextButton : Result := RxResConst_Por.SNextButton;
     SPrevButton : Result := RxResConst_Por.SPrevButton;
     SExprIncorrect : Result := RxResConst_Por.SExprIncorrect;
     SExprTermination : Result := RxResConst_Por.SExprTermination;
     SExprNameError : Result := RxResConst_Por.SExprNameError;
     SExprStringError : Result := RxResConst_Por.SExprStringError;
     SExprInvalidChar : Result := RxResConst_Por.SExprInvalidChar;
     SExprNoRParen : Result := RxResConst_Por.SExprNoRParen;
     SExprExpected : Result := RxResConst_Por.SExprExpected;
     SExprBadCompare : Result := RxResConst_Por.SExprBadCompare;
     SConfirmSave : Result := RxResConst_Por.SConfirmSave;
     SDatabaseName : Result := RxResConst_Por.SDatabaseName;
     SUnlockCaption : Result := RxResConst_Por.SUnlockCaption;
     SUnlockHint : Result := RxResConst_Por.SUnlockHint;
     SDeleteMultipleRecords : Result := RxResConst_Por.SDeleteMultipleRecords;
     SGIFImage : Result := RxResConst_Por.SGIFImage;
     SChangeGIFSize : Result := RxResConst_Por.SChangeGIFSize;
     SNoGIFData : Result := RxResConst_Por.SNoGIFData;
     SUnrecognizedGIFExt : Result := RxResConst_Por.SUnrecognizedGIFExt;
     SWrongGIFColors : Result := RxResConst_Por.SWrongGIFColors;
     SBadGIFCodeSize : Result := RxResConst_Por.SBadGIFCodeSize;
     SGIFDecodeError : Result := RxResConst_Por.SGIFDecodeError;
     SGIFEncodeError : Result := RxResConst_Por.SGIFEncodeError;
     SGIFVersion : Result := RxResConst_Por.SGIFVersion;
     SDualListSrcCaption : Result := RxResConst_Por.SDualListSrcCaption;
     SDualListDestCaption : Result := RxResConst_Por.SDualListDestCaption;
     SClipbrdUnknown : Result := RxResConst_Por.SClipbrdUnknown;
     SClipbrdEmpty : Result := RxResConst_Por.SClipbrdEmpty;
     SCustomizeSpeedbar : Result := RxResConst_Por.SCustomizeSpeedbar;
     SAvailButtons : Result := RxResConst_Por.SAvailButtons;
     SSpeedbarCategories : Result := RxResConst_Por.SSpeedbarCategories;
     SSpeedbarEditHint : Result := RxResConst_Por.SSpeedbarEditHint;
     SParseSyntaxError : Result := RxResConst_Por.SParseSyntaxError;
     SParseNotCramp : Result := RxResConst_Por.SParseNotCramp;
     SParseDivideByZero : Result := RxResConst_Por.SParseDivideByZero;
     SParseSqrError : Result := RxResConst_Por.SParseSqrError;
     SParseLogError : Result := RxResConst_Por.SParseLogError;
     SParseInvalidFloatOperation : Result := RxResConst_Por.SParseInvalidFloatOperation;
     srRXControls : Result := RxResConst_Por.srRXControls;
     srRXDBAware : Result := RxResConst_Por.srRXDBAware;
     srRXTools : Result := RxResConst_Por.srRXTools;
     srStorageDesigner : Result := RxResConst_Por.srStorageDesigner;
     srProxyEditor : Result := RxResConst_Por.srProxyEditor;
     srPageProxies : Result := RxResConst_Por.srPageProxies;
     srProxyName : Result := RxResConst_Por.srProxyName;
     srPageName : Result := RxResConst_Por.srPageName;
     srSBItemNotCreate : Result := RxResConst_Por.srSBItemNotCreate;
     srConfirmSBDelete : Result := RxResConst_Por.srConfirmSBDelete;
     srSpeedbarDesigner : Result := RxResConst_Por.srSpeedbarDesigner;
     srNewSectionName : Result := RxResConst_Por.srNewSectionName;
     srEventNotCreate : Result := RxResConst_Por.srEventNotCreate;
     srTimerDesigner : Result := RxResConst_Por.srTimerDesigner;
     srTimerEvents : Result := RxResConst_Por.srTimerEvents;
     srAniCurFilter : Result := RxResConst_Por.srAniCurFilter;
     srEditPicture : Result := RxResConst_Por.srEditPicture;
     srLoadAniCursor : Result := RxResConst_Por.srLoadAniCursor;
     srLoadIcon : Result := RxResConst_Por.srLoadIcon;
     srSaveImageList : Result := RxResConst_Por.srSaveImageList;
     srCaptionDesigner : Result := RxResConst_Por.srCaptionDesigner;
     srGradientCaptions : Result := RxResConst_Por.srGradientCaptions;
     srBorrowStructure : Result := RxResConst_Por.srBorrowStructure;
     ccTimeEditHours : Result := RxResConst_Por.ccTimeEditHours;
     ccTimeEditMins : Result := RxResConst_Por.ccTimeEditMins;
     ccTimeEditSecs : Result := RxResConst_Por.ccTimeEditSecs;
     SColorBlack : Result := RxResConst_Por.SColorBlack;
     SColorMaroon : Result := RxResConst_Por.SColorMaroon;
     SColorGreen : Result := RxResConst_Por.SColorGreen;
     SColorOlive : Result := RxResConst_Por.SColorOlive;
     SColorNavy : Result := RxResConst_Por.SColorNavy;
     SColorPurple : Result := RxResConst_Por.SColorPurple;
     SColorTeal : Result := RxResConst_Por.SColorTeal;
     SColorGray : Result := RxResConst_Por.SColorGray;
     SColorSilver : Result := RxResConst_Por.SColorSilver;
     SColorRed : Result := RxResConst_Por.SColorRed;
     SColorLime : Result := RxResConst_Por.SColorLime;
     SColorYellow : Result := RxResConst_Por.SColorYellow;
     SColorBlue : Result := RxResConst_Por.SColorBlue;
     SColorFuchsia : Result := RxResConst_Por.SColorFuchsia;
     SColorAqua : Result := RxResConst_Por.SColorAqua;
     SColorWhite : Result := RxResConst_Por.SColorWhite;
     SColorScrollBar : Result := RxResConst_Por.SColorScrollBar;
     SColorBackground : Result := RxResConst_Por.SColorBackground;
     SColorActiveCaption : Result := RxResConst_Por.SColorActiveCaption;
     SColorInactiveCaption : Result := RxResConst_Por.SColorInactiveCaption;
     SColorMenu : Result := RxResConst_Por.SColorMenu;
     SColorWindow : Result := RxResConst_Por.SColorWindow;
     SColorWindowFrame : Result := RxResConst_Por.SColorWindowFrame;
     SColorMenuText : Result := RxResConst_Por.SColorMenuText;
     SColorWindowText : Result := RxResConst_Por.SColorWindowText;
     SColorCaptionText : Result := RxResConst_Por.SColorCaptionText;
     SColorActiveBorder : Result := RxResConst_Por.SColorActiveBorder;
     SColorInactiveBorder : Result := RxResConst_Por.SColorInactiveBorder;
     SColorAppWorkSpace : Result := RxResConst_Por.SColorAppWorkSpace;
     SColorHighlight : Result := RxResConst_Por.SColorHighlight;
     SColorHighlightText : Result := RxResConst_Por.SColorHighlightText;
     SColorBtnFace : Result := RxResConst_Por.SColorBtnFace;
     SColorBtnShadow : Result := RxResConst_Por.SColorBtnShadow;
     SColorGrayText : Result := RxResConst_Por.SColorGrayText;
     SColorBtnText : Result := RxResConst_Por.SColorBtnText;
     SColorInactiveCaptionText : Result := RxResConst_Por.SColorInactiveCaptionText;
     SColorBtnHighlight : Result := RxResConst_Por.SColorBtnHighlight;
     SColor3DDkShadow : Result := RxResConst_Por.SColor3DDkShadow;
     SColor3DLight : Result := RxResConst_Por.SColor3DLight;
     SColorInfoText : Result := RxResConst_Por.SColorInfoText;
     SColorInfoBk : Result := RxResConst_Por.SColorInfoBk;
     SColorCream : Result := RxResConst_Por.SColorCream;
     SColorMoneyGreen : Result := RxResConst_Por.SColorMoneyGreen;
     SColorSkyBlue : Result := RxResConst_Por.SColorSkyBlue;
     SColorNone : Result := RxResConst_Por.SColorNone;
     SColorDefault : Result := RxResConst_Por.SColorDefault;
     SColorCustom : Result := RxResConst_Por.SColorCustom;
     SMemNoRecords : Result := RxResConst_Por.SMemNoRecords;
     SInvalidFields : Result := RxResConst_Por.SInvalidFields;
     SAutoSpeedbarMode : Result := RxResConst_Por.SAutoSpeedbarMode;
    end;
  end;
  {$ENDIF}
  {$IFDEF RXLANG_Rom}
  function RxResConst_Rom_RxLoadStr(IDMsg: Integer): string;
  begin
    case IDMsg of
     SBrowse : Result := RxResConst_Rom.SBrowse;
     SDefaultFilter : Result := RxResConst_Rom.SDefaultFilter;
     SDateDlgTitle : Result := RxResConst_Rom.SDateDlgTitle;
     SNextYear : Result := RxResConst_Rom.SNextYear;
     SNextMonth : Result := RxResConst_Rom.SNextMonth;
     SPrevYear : Result := RxResConst_Rom.SPrevYear;
     SPrevMonth : Result := RxResConst_Rom.SPrevMonth;
     SNotImplemented : Result := RxResConst_Rom.SNotImplemented;
     SFileNotExec : Result := RxResConst_Rom.SFileNotExec;
     SLoadLibError : Result := RxResConst_Rom.SLoadLibError;
     SDetails : Result := RxResConst_Rom.SDetails;
     SDateOutOfRange : Result := RxResConst_Rom.SDateOutOfRange;
     SDateOutOfMin : Result := RxResConst_Rom.SDateOutOfMin;
     SDateOutOfMax : Result := RxResConst_Rom.SDateOutOfMax;
     SDateMinLimit : Result := RxResConst_Rom.SDateMinLimit;
     SDateMaxLimit : Result := RxResConst_Rom.SDateMaxLimit;
     SLocalDatabase : Result := RxResConst_Rom.SLocalDatabase;
     SRetryLogin : Result := RxResConst_Rom.SRetryLogin;
     SExprNotBoolean : Result := RxResConst_Rom.SExprNotBoolean;
     SExprBadNullTest : Result := RxResConst_Rom.SExprBadNullTest;
     SExprBadField : Result := RxResConst_Rom.SExprBadField;
     SCaptureFilter : Result := RxResConst_Rom.SCaptureFilter;
     SNotCaptureFilter : Result := RxResConst_Rom.SNotCaptureFilter;
     SInactiveData : Result := RxResConst_Rom.SInactiveData;
     SBrowseData : Result := RxResConst_Rom.SBrowseData;
     SEditData : Result := RxResConst_Rom.SEditData;
     SInsertData : Result := RxResConst_Rom.SInsertData;
     SSetKeyData : Result := RxResConst_Rom.SSetKeyData;
     SCalcFieldsData : Result := RxResConst_Rom.SCalcFieldsData;
     SRegistration : Result := RxResConst_Rom.SRegistration;
     SAppTitleLabel : Result := RxResConst_Rom.SAppTitleLabel;
     SHintLabel : Result := RxResConst_Rom.SHintLabel;
     SUserNameLabel : Result := RxResConst_Rom.SUserNameLabel;
     SPasswordLabel : Result := RxResConst_Rom.SPasswordLabel;
     SInvalidUserName : Result := RxResConst_Rom.SInvalidUserName;
     SChangePassword : Result := RxResConst_Rom.SChangePassword;
     SOldPasswordLabel : Result := RxResConst_Rom.SOldPasswordLabel;
     SNewPasswordLabel : Result := RxResConst_Rom.SNewPasswordLabel;
     SConfirmPasswordLabel : Result := RxResConst_Rom.SConfirmPasswordLabel;
     SPasswordChanged : Result := RxResConst_Rom.SPasswordChanged;
     SPasswordNotChanged : Result := RxResConst_Rom.SPasswordNotChanged;
     SPasswordsMismatch : Result := RxResConst_Rom.SPasswordsMismatch;
     SDBExceptCaption : Result := RxResConst_Rom.SDBExceptCaption;
     SBDEErrorLabel : Result := RxResConst_Rom.SBDEErrorLabel;
     SServerErrorLabel : Result := RxResConst_Rom.SServerErrorLabel;
     SErrorMsgLabel : Result := RxResConst_Rom.SErrorMsgLabel;
     SNextButton : Result := RxResConst_Rom.SNextButton;
     SPrevButton : Result := RxResConst_Rom.SPrevButton;
     SExprIncorrect : Result := RxResConst_Rom.SExprIncorrect;
     SExprTermination : Result := RxResConst_Rom.SExprTermination;
     SExprNameError : Result := RxResConst_Rom.SExprNameError;
     SExprStringError : Result := RxResConst_Rom.SExprStringError;
     SExprInvalidChar : Result := RxResConst_Rom.SExprInvalidChar;
     SExprNoRParen : Result := RxResConst_Rom.SExprNoRParen;
     SExprExpected : Result := RxResConst_Rom.SExprExpected;
     SExprBadCompare : Result := RxResConst_Rom.SExprBadCompare;
     SConfirmSave : Result := RxResConst_Rom.SConfirmSave;
     SDatabaseName : Result := RxResConst_Rom.SDatabaseName;
     SUnlockCaption : Result := RxResConst_Rom.SUnlockCaption;
     SUnlockHint : Result := RxResConst_Rom.SUnlockHint;
     SDeleteMultipleRecords : Result := RxResConst_Rom.SDeleteMultipleRecords;
     SGIFImage : Result := RxResConst_Rom.SGIFImage;
     SChangeGIFSize : Result := RxResConst_Rom.SChangeGIFSize;
     SNoGIFData : Result := RxResConst_Rom.SNoGIFData;
     SUnrecognizedGIFExt : Result := RxResConst_Rom.SUnrecognizedGIFExt;
     SWrongGIFColors : Result := RxResConst_Rom.SWrongGIFColors;
     SBadGIFCodeSize : Result := RxResConst_Rom.SBadGIFCodeSize;
     SGIFDecodeError : Result := RxResConst_Rom.SGIFDecodeError;
     SGIFEncodeError : Result := RxResConst_Rom.SGIFEncodeError;
     SGIFVersion : Result := RxResConst_Rom.SGIFVersion;
     SDualListSrcCaption : Result := RxResConst_Rom.SDualListSrcCaption;
     SDualListDestCaption : Result := RxResConst_Rom.SDualListDestCaption;
     SClipbrdUnknown : Result := RxResConst_Rom.SClipbrdUnknown;
     SClipbrdEmpty : Result := RxResConst_Rom.SClipbrdEmpty;
     SCustomizeSpeedbar : Result := RxResConst_Rom.SCustomizeSpeedbar;
     SAvailButtons : Result := RxResConst_Rom.SAvailButtons;
     SSpeedbarCategories : Result := RxResConst_Rom.SSpeedbarCategories;
     SSpeedbarEditHint : Result := RxResConst_Rom.SSpeedbarEditHint;
     SParseSyntaxError : Result := RxResConst_Rom.SParseSyntaxError;
     SParseNotCramp : Result := RxResConst_Rom.SParseNotCramp;
     SParseDivideByZero : Result := RxResConst_Rom.SParseDivideByZero;
     SParseSqrError : Result := RxResConst_Rom.SParseSqrError;
     SParseLogError : Result := RxResConst_Rom.SParseLogError;
     SParseInvalidFloatOperation : Result := RxResConst_Rom.SParseInvalidFloatOperation;
     srRXControls : Result := RxResConst_Rom.srRXControls;
     srRXDBAware : Result := RxResConst_Rom.srRXDBAware;
     srRXTools : Result := RxResConst_Rom.srRXTools;
     srStorageDesigner : Result := RxResConst_Rom.srStorageDesigner;
     srProxyEditor : Result := RxResConst_Rom.srProxyEditor;
     srPageProxies : Result := RxResConst_Rom.srPageProxies;
     srProxyName : Result := RxResConst_Rom.srProxyName;
     srPageName : Result := RxResConst_Rom.srPageName;
     srSBItemNotCreate : Result := RxResConst_Rom.srSBItemNotCreate;
     srConfirmSBDelete : Result := RxResConst_Rom.srConfirmSBDelete;
     srSpeedbarDesigner : Result := RxResConst_Rom.srSpeedbarDesigner;
     srNewSectionName : Result := RxResConst_Rom.srNewSectionName;
     srEventNotCreate : Result := RxResConst_Rom.srEventNotCreate;
     srTimerDesigner : Result := RxResConst_Rom.srTimerDesigner;
     srTimerEvents : Result := RxResConst_Rom.srTimerEvents;
     srAniCurFilter : Result := RxResConst_Rom.srAniCurFilter;
     srEditPicture : Result := RxResConst_Rom.srEditPicture;
     srLoadAniCursor : Result := RxResConst_Rom.srLoadAniCursor;
     srLoadIcon : Result := RxResConst_Rom.srLoadIcon;
     srSaveImageList : Result := RxResConst_Rom.srSaveImageList;
     srCaptionDesigner : Result := RxResConst_Rom.srCaptionDesigner;
     srGradientCaptions : Result := RxResConst_Rom.srGradientCaptions;
     srBorrowStructure : Result := RxResConst_Rom.srBorrowStructure;
     ccTimeEditHours : Result := RxResConst_Rom.ccTimeEditHours;
     ccTimeEditMins : Result := RxResConst_Rom.ccTimeEditMins;
     ccTimeEditSecs : Result := RxResConst_Rom.ccTimeEditSecs;
     SColorBlack : Result := RxResConst_Rom.SColorBlack;
     SColorMaroon : Result := RxResConst_Rom.SColorMaroon;
     SColorGreen : Result := RxResConst_Rom.SColorGreen;
     SColorOlive : Result := RxResConst_Rom.SColorOlive;
     SColorNavy : Result := RxResConst_Rom.SColorNavy;
     SColorPurple : Result := RxResConst_Rom.SColorPurple;
     SColorTeal : Result := RxResConst_Rom.SColorTeal;
     SColorGray : Result := RxResConst_Rom.SColorGray;
     SColorSilver : Result := RxResConst_Rom.SColorSilver;
     SColorRed : Result := RxResConst_Rom.SColorRed;
     SColorLime : Result := RxResConst_Rom.SColorLime;
     SColorYellow : Result := RxResConst_Rom.SColorYellow;
     SColorBlue : Result := RxResConst_Rom.SColorBlue;
     SColorFuchsia : Result := RxResConst_Rom.SColorFuchsia;
     SColorAqua : Result := RxResConst_Rom.SColorAqua;
     SColorWhite : Result := RxResConst_Rom.SColorWhite;
     SColorScrollBar : Result := RxResConst_Rom.SColorScrollBar;
     SColorBackground : Result := RxResConst_Rom.SColorBackground;
     SColorActiveCaption : Result := RxResConst_Rom.SColorActiveCaption;
     SColorInactiveCaption : Result := RxResConst_Rom.SColorInactiveCaption;
     SColorMenu : Result := RxResConst_Rom.SColorMenu;
     SColorWindow : Result := RxResConst_Rom.SColorWindow;
     SColorWindowFrame : Result := RxResConst_Rom.SColorWindowFrame;
     SColorMenuText : Result := RxResConst_Rom.SColorMenuText;
     SColorWindowText : Result := RxResConst_Rom.SColorWindowText;
     SColorCaptionText : Result := RxResConst_Rom.SColorCaptionText;
     SColorActiveBorder : Result := RxResConst_Rom.SColorActiveBorder;
     SColorInactiveBorder : Result := RxResConst_Rom.SColorInactiveBorder;
     SColorAppWorkSpace : Result := RxResConst_Rom.SColorAppWorkSpace;
     SColorHighlight : Result := RxResConst_Rom.SColorHighlight;
     SColorHighlightText : Result := RxResConst_Rom.SColorHighlightText;
     SColorBtnFace : Result := RxResConst_Rom.SColorBtnFace;
     SColorBtnShadow : Result := RxResConst_Rom.SColorBtnShadow;
     SColorGrayText : Result := RxResConst_Rom.SColorGrayText;
     SColorBtnText : Result := RxResConst_Rom.SColorBtnText;
     SColorInactiveCaptionText : Result := RxResConst_Rom.SColorInactiveCaptionText;
     SColorBtnHighlight : Result := RxResConst_Rom.SColorBtnHighlight;
     SColor3DDkShadow : Result := RxResConst_Rom.SColor3DDkShadow;
     SColor3DLight : Result := RxResConst_Rom.SColor3DLight;
     SColorInfoText : Result := RxResConst_Rom.SColorInfoText;
     SColorInfoBk : Result := RxResConst_Rom.SColorInfoBk;
     SColorCream : Result := RxResConst_Rom.SColorCream;
     SColorMoneyGreen : Result := RxResConst_Rom.SColorMoneyGreen;
     SColorSkyBlue : Result := RxResConst_Rom.SColorSkyBlue;
     SColorNone : Result := RxResConst_Rom.SColorNone;
     SColorDefault : Result := RxResConst_Rom.SColorDefault;
     SColorCustom : Result := RxResConst_Rom.SColorCustom;
     SMemNoRecords : Result := RxResConst_Rom.SMemNoRecords;
     SInvalidFields : Result := RxResConst_Rom.SInvalidFields;
     SAutoSpeedbarMode : Result := RxResConst_Rom.SAutoSpeedbarMode;
    end;
  end;
  {$ENDIF}
  {$IFDEF RXLANG_Rus}
  function RxResConst_Rus_RxLoadStr(IDMsg: Integer): string;
  begin
    case IDMsg of
     SBrowse : Result := RxResConst_Rus.SBrowse;
     SDefaultFilter : Result := RxResConst_Rus.SDefaultFilter;
     SDateDlgTitle : Result := RxResConst_Rus.SDateDlgTitle;
     SNextYear : Result := RxResConst_Rus.SNextYear;
     SNextMonth : Result := RxResConst_Rus.SNextMonth;
     SPrevYear : Result := RxResConst_Rus.SPrevYear;
     SPrevMonth : Result := RxResConst_Rus.SPrevMonth;
     SNotImplemented : Result := RxResConst_Rus.SNotImplemented;
     SFileNotExec : Result := RxResConst_Rus.SFileNotExec;
     SLoadLibError : Result := RxResConst_Rus.SLoadLibError;
     SDetails : Result := RxResConst_Rus.SDetails;
     SDateOutOfRange : Result := RxResConst_Rus.SDateOutOfRange;
     SDateOutOfMin : Result := RxResConst_Rus.SDateOutOfMin;
     SDateOutOfMax : Result := RxResConst_Rus.SDateOutOfMax;
     SDateMinLimit : Result := RxResConst_Rus.SDateMinLimit;
     SDateMaxLimit : Result := RxResConst_Rus.SDateMaxLimit;
     SLocalDatabase : Result := RxResConst_Rus.SLocalDatabase;
     SRetryLogin : Result := RxResConst_Rus.SRetryLogin;
     SExprNotBoolean : Result := RxResConst_Rus.SExprNotBoolean;
     SExprBadNullTest : Result := RxResConst_Rus.SExprBadNullTest;
     SExprBadField : Result := RxResConst_Rus.SExprBadField;
     SCaptureFilter : Result := RxResConst_Rus.SCaptureFilter;
     SNotCaptureFilter : Result := RxResConst_Rus.SNotCaptureFilter;
     SInactiveData : Result := RxResConst_Rus.SInactiveData;
     SBrowseData : Result := RxResConst_Rus.SBrowseData;
     SEditData : Result := RxResConst_Rus.SEditData;
     SInsertData : Result := RxResConst_Rus.SInsertData;
     SSetKeyData : Result := RxResConst_Rus.SSetKeyData;
     SCalcFieldsData : Result := RxResConst_Rus.SCalcFieldsData;
     SRegistration : Result := RxResConst_Rus.SRegistration;
     SAppTitleLabel : Result := RxResConst_Rus.SAppTitleLabel;
     SHintLabel : Result := RxResConst_Rus.SHintLabel;
     SUserNameLabel : Result := RxResConst_Rus.SUserNameLabel;
     SPasswordLabel : Result := RxResConst_Rus.SPasswordLabel;
     SInvalidUserName : Result := RxResConst_Rus.SInvalidUserName;
     SChangePassword : Result := RxResConst_Rus.SChangePassword;
     SOldPasswordLabel : Result := RxResConst_Rus.SOldPasswordLabel;
     SNewPasswordLabel : Result := RxResConst_Rus.SNewPasswordLabel;
     SConfirmPasswordLabel : Result := RxResConst_Rus.SConfirmPasswordLabel;
     SPasswordChanged : Result := RxResConst_Rus.SPasswordChanged;
     SPasswordNotChanged : Result := RxResConst_Rus.SPasswordNotChanged;
     SPasswordsMismatch : Result := RxResConst_Rus.SPasswordsMismatch;
     SDBExceptCaption : Result := RxResConst_Rus.SDBExceptCaption;
     SBDEErrorLabel : Result := RxResConst_Rus.SBDEErrorLabel;
     SServerErrorLabel : Result := RxResConst_Rus.SServerErrorLabel;
     SErrorMsgLabel : Result := RxResConst_Rus.SErrorMsgLabel;
     SNextButton : Result := RxResConst_Rus.SNextButton;
     SPrevButton : Result := RxResConst_Rus.SPrevButton;
     SExprIncorrect : Result := RxResConst_Rus.SExprIncorrect;
     SExprTermination : Result := RxResConst_Rus.SExprTermination;
     SExprNameError : Result := RxResConst_Rus.SExprNameError;
     SExprStringError : Result := RxResConst_Rus.SExprStringError;
     SExprInvalidChar : Result := RxResConst_Rus.SExprInvalidChar;
     SExprNoRParen : Result := RxResConst_Rus.SExprNoRParen;
     SExprExpected : Result := RxResConst_Rus.SExprExpected;
     SExprBadCompare : Result := RxResConst_Rus.SExprBadCompare;
     SConfirmSave : Result := RxResConst_Rus.SConfirmSave;
     SDatabaseName : Result := RxResConst_Rus.SDatabaseName;
     SUnlockCaption : Result := RxResConst_Rus.SUnlockCaption;
     SUnlockHint : Result := RxResConst_Rus.SUnlockHint;
     SDeleteMultipleRecords : Result := RxResConst_Rus.SDeleteMultipleRecords;
     SGIFImage : Result := RxResConst_Rus.SGIFImage;
     SChangeGIFSize : Result := RxResConst_Rus.SChangeGIFSize;
     SNoGIFData : Result := RxResConst_Rus.SNoGIFData;
     SUnrecognizedGIFExt : Result := RxResConst_Rus.SUnrecognizedGIFExt;
     SWrongGIFColors : Result := RxResConst_Rus.SWrongGIFColors;
     SBadGIFCodeSize : Result := RxResConst_Rus.SBadGIFCodeSize;
     SGIFDecodeError : Result := RxResConst_Rus.SGIFDecodeError;
     SGIFEncodeError : Result := RxResConst_Rus.SGIFEncodeError;
     SGIFVersion : Result := RxResConst_Rus.SGIFVersion;
     SDualListSrcCaption : Result := RxResConst_Rus.SDualListSrcCaption;
     SDualListDestCaption : Result := RxResConst_Rus.SDualListDestCaption;
     SClipbrdUnknown : Result := RxResConst_Rus.SClipbrdUnknown;
     SClipbrdEmpty : Result := RxResConst_Rus.SClipbrdEmpty;
     SCustomizeSpeedbar : Result := RxResConst_Rus.SCustomizeSpeedbar;
     SAvailButtons : Result := RxResConst_Rus.SAvailButtons;
     SSpeedbarCategories : Result := RxResConst_Rus.SSpeedbarCategories;
     SSpeedbarEditHint : Result := RxResConst_Rus.SSpeedbarEditHint;
     SParseSyntaxError : Result := RxResConst_Rus.SParseSyntaxError;
     SParseNotCramp : Result := RxResConst_Rus.SParseNotCramp;
     SParseDivideByZero : Result := RxResConst_Rus.SParseDivideByZero;
     SParseSqrError : Result := RxResConst_Rus.SParseSqrError;
     SParseLogError : Result := RxResConst_Rus.SParseLogError;
     SParseInvalidFloatOperation : Result := RxResConst_Rus.SParseInvalidFloatOperation;
     srRXControls : Result := RxResConst_Rus.srRXControls;
     srRXDBAware : Result := RxResConst_Rus.srRXDBAware;
     srRXTools : Result := RxResConst_Rus.srRXTools;
     srStorageDesigner : Result := RxResConst_Rus.srStorageDesigner;
     srProxyEditor : Result := RxResConst_Rus.srProxyEditor;
     srPageProxies : Result := RxResConst_Rus.srPageProxies;
     srProxyName : Result := RxResConst_Rus.srProxyName;
     srPageName : Result := RxResConst_Rus.srPageName;
     srSBItemNotCreate : Result := RxResConst_Rus.srSBItemNotCreate;
     srConfirmSBDelete : Result := RxResConst_Rus.srConfirmSBDelete;
     srSpeedbarDesigner : Result := RxResConst_Rus.srSpeedbarDesigner;
     srNewSectionName : Result := RxResConst_Rus.srNewSectionName;
     srEventNotCreate : Result := RxResConst_Rus.srEventNotCreate;
     srTimerDesigner : Result := RxResConst_Rus.srTimerDesigner;
     srTimerEvents : Result := RxResConst_Rus.srTimerEvents;
     srAniCurFilter : Result := RxResConst_Rus.srAniCurFilter;
     srEditPicture : Result := RxResConst_Rus.srEditPicture;
     srLoadAniCursor : Result := RxResConst_Rus.srLoadAniCursor;
     srLoadIcon : Result := RxResConst_Rus.srLoadIcon;
     srSaveImageList : Result := RxResConst_Rus.srSaveImageList;
     srCaptionDesigner : Result := RxResConst_Rus.srCaptionDesigner;
     srGradientCaptions : Result := RxResConst_Rus.srGradientCaptions;
     srBorrowStructure : Result := RxResConst_Rus.srBorrowStructure;
     ccTimeEditHours : Result := RxResConst_Rus.ccTimeEditHours;
     ccTimeEditMins : Result := RxResConst_Rus.ccTimeEditMins;
     ccTimeEditSecs : Result := RxResConst_Rus.ccTimeEditSecs;
     SColorBlack : Result := RxResConst_Rus.SColorBlack;
     SColorMaroon : Result := RxResConst_Rus.SColorMaroon;
     SColorGreen : Result := RxResConst_Rus.SColorGreen;
     SColorOlive : Result := RxResConst_Rus.SColorOlive;
     SColorNavy : Result := RxResConst_Rus.SColorNavy;
     SColorPurple : Result := RxResConst_Rus.SColorPurple;
     SColorTeal : Result := RxResConst_Rus.SColorTeal;
     SColorGray : Result := RxResConst_Rus.SColorGray;
     SColorSilver : Result := RxResConst_Rus.SColorSilver;
     SColorRed : Result := RxResConst_Rus.SColorRed;
     SColorLime : Result := RxResConst_Rus.SColorLime;
     SColorYellow : Result := RxResConst_Rus.SColorYellow;
     SColorBlue : Result := RxResConst_Rus.SColorBlue;
     SColorFuchsia : Result := RxResConst_Rus.SColorFuchsia;
     SColorAqua : Result := RxResConst_Rus.SColorAqua;
     SColorWhite : Result := RxResConst_Rus.SColorWhite;
     SColorScrollBar : Result := RxResConst_Rus.SColorScrollBar;
     SColorBackground : Result := RxResConst_Rus.SColorBackground;
     SColorActiveCaption : Result := RxResConst_Rus.SColorActiveCaption;
     SColorInactiveCaption : Result := RxResConst_Rus.SColorInactiveCaption;
     SColorMenu : Result := RxResConst_Rus.SColorMenu;
     SColorWindow : Result := RxResConst_Rus.SColorWindow;
     SColorWindowFrame : Result := RxResConst_Rus.SColorWindowFrame;
     SColorMenuText : Result := RxResConst_Rus.SColorMenuText;
     SColorWindowText : Result := RxResConst_Rus.SColorWindowText;
     SColorCaptionText : Result := RxResConst_Rus.SColorCaptionText;
     SColorActiveBorder : Result := RxResConst_Rus.SColorActiveBorder;
     SColorInactiveBorder : Result := RxResConst_Rus.SColorInactiveBorder;
     SColorAppWorkSpace : Result := RxResConst_Rus.SColorAppWorkSpace;
     SColorHighlight : Result := RxResConst_Rus.SColorHighlight;
     SColorHighlightText : Result := RxResConst_Rus.SColorHighlightText;
     SColorBtnFace : Result := RxResConst_Rus.SColorBtnFace;
     SColorBtnShadow : Result := RxResConst_Rus.SColorBtnShadow;
     SColorGrayText : Result := RxResConst_Rus.SColorGrayText;
     SColorBtnText : Result := RxResConst_Rus.SColorBtnText;
     SColorInactiveCaptionText : Result := RxResConst_Rus.SColorInactiveCaptionText;
     SColorBtnHighlight : Result := RxResConst_Rus.SColorBtnHighlight;
     SColor3DDkShadow : Result := RxResConst_Rus.SColor3DDkShadow;
     SColor3DLight : Result := RxResConst_Rus.SColor3DLight;
     SColorInfoText : Result := RxResConst_Rus.SColorInfoText;
     SColorInfoBk : Result := RxResConst_Rus.SColorInfoBk;
     SColorCream : Result := RxResConst_Rus.SColorCream;
     SColorMoneyGreen : Result := RxResConst_Rus.SColorMoneyGreen;
     SColorSkyBlue : Result := RxResConst_Rus.SColorSkyBlue;
     SColorNone : Result := RxResConst_Rus.SColorNone;
     SColorDefault : Result := RxResConst_Rus.SColorDefault;
     SColorCustom : Result := RxResConst_Rus.SColorCustom;
     SMemNoRecords : Result := RxResConst_Rus.SMemNoRecords;
     SInvalidFields : Result := RxResConst_Rus.SInvalidFields;
     SAutoSpeedbarMode : Result := RxResConst_Rus.SAutoSpeedbarMode;
    end;
  end;
  {$ENDIF}
  {$IFDEF RXLANG_Slo}
  function RxResConst_Slo_RxLoadStr(IDMsg: Integer): string;
  begin
    case IDMsg of
     SBrowse : Result := RxResConst_Slo.SBrowse;
     SDefaultFilter : Result := RxResConst_Slo.SDefaultFilter;
     SDateDlgTitle : Result := RxResConst_Slo.SDateDlgTitle;
     SNextYear : Result := RxResConst_Slo.SNextYear;
     SNextMonth : Result := RxResConst_Slo.SNextMonth;
     SPrevYear : Result := RxResConst_Slo.SPrevYear;
     SPrevMonth : Result := RxResConst_Slo.SPrevMonth;
     SNotImplemented : Result := RxResConst_Slo.SNotImplemented;
     SFileNotExec : Result := RxResConst_Slo.SFileNotExec;
     SLoadLibError : Result := RxResConst_Slo.SLoadLibError;
     SDetails : Result := RxResConst_Slo.SDetails;
     SDateOutOfRange : Result := RxResConst_Slo.SDateOutOfRange;
     SDateOutOfMin : Result := RxResConst_Slo.SDateOutOfMin;
     SDateOutOfMax : Result := RxResConst_Slo.SDateOutOfMax;
     SDateMinLimit : Result := RxResConst_Slo.SDateMinLimit;
     SDateMaxLimit : Result := RxResConst_Slo.SDateMaxLimit;
     SLocalDatabase : Result := RxResConst_Slo.SLocalDatabase;
     SRetryLogin : Result := RxResConst_Slo.SRetryLogin;
     SExprNotBoolean : Result := RxResConst_Slo.SExprNotBoolean;
     SExprBadNullTest : Result := RxResConst_Slo.SExprBadNullTest;
     SExprBadField : Result := RxResConst_Slo.SExprBadField;
     SCaptureFilter : Result := RxResConst_Slo.SCaptureFilter;
     SNotCaptureFilter : Result := RxResConst_Slo.SNotCaptureFilter;
     SInactiveData : Result := RxResConst_Slo.SInactiveData;
     SBrowseData : Result := RxResConst_Slo.SBrowseData;
     SEditData : Result := RxResConst_Slo.SEditData;
     SInsertData : Result := RxResConst_Slo.SInsertData;
     SSetKeyData : Result := RxResConst_Slo.SSetKeyData;
     SCalcFieldsData : Result := RxResConst_Slo.SCalcFieldsData;
     SRegistration : Result := RxResConst_Slo.SRegistration;
     SAppTitleLabel : Result := RxResConst_Slo.SAppTitleLabel;
     SHintLabel : Result := RxResConst_Slo.SHintLabel;
     SUserNameLabel : Result := RxResConst_Slo.SUserNameLabel;
     SPasswordLabel : Result := RxResConst_Slo.SPasswordLabel;
     SInvalidUserName : Result := RxResConst_Slo.SInvalidUserName;
     SChangePassword : Result := RxResConst_Slo.SChangePassword;
     SOldPasswordLabel : Result := RxResConst_Slo.SOldPasswordLabel;
     SNewPasswordLabel : Result := RxResConst_Slo.SNewPasswordLabel;
     SConfirmPasswordLabel : Result := RxResConst_Slo.SConfirmPasswordLabel;
     SPasswordChanged : Result := RxResConst_Slo.SPasswordChanged;
     SPasswordNotChanged : Result := RxResConst_Slo.SPasswordNotChanged;
     SPasswordsMismatch : Result := RxResConst_Slo.SPasswordsMismatch;
     SDBExceptCaption : Result := RxResConst_Slo.SDBExceptCaption;
     SBDEErrorLabel : Result := RxResConst_Slo.SBDEErrorLabel;
     SServerErrorLabel : Result := RxResConst_Slo.SServerErrorLabel;
     SErrorMsgLabel : Result := RxResConst_Slo.SErrorMsgLabel;
     SNextButton : Result := RxResConst_Slo.SNextButton;
     SPrevButton : Result := RxResConst_Slo.SPrevButton;
     SExprIncorrect : Result := RxResConst_Slo.SExprIncorrect;
     SExprTermination : Result := RxResConst_Slo.SExprTermination;
     SExprNameError : Result := RxResConst_Slo.SExprNameError;
     SExprStringError : Result := RxResConst_Slo.SExprStringError;
     SExprInvalidChar : Result := RxResConst_Slo.SExprInvalidChar;
     SExprNoRParen : Result := RxResConst_Slo.SExprNoRParen;
     SExprExpected : Result := RxResConst_Slo.SExprExpected;
     SExprBadCompare : Result := RxResConst_Slo.SExprBadCompare;
     SConfirmSave : Result := RxResConst_Slo.SConfirmSave;
     SDatabaseName : Result := RxResConst_Slo.SDatabaseName;
     SUnlockCaption : Result := RxResConst_Slo.SUnlockCaption;
     SUnlockHint : Result := RxResConst_Slo.SUnlockHint;
     SDeleteMultipleRecords : Result := RxResConst_Slo.SDeleteMultipleRecords;
     SGIFImage : Result := RxResConst_Slo.SGIFImage;
     SChangeGIFSize : Result := RxResConst_Slo.SChangeGIFSize;
     SNoGIFData : Result := RxResConst_Slo.SNoGIFData;
     SUnrecognizedGIFExt : Result := RxResConst_Slo.SUnrecognizedGIFExt;
     SWrongGIFColors : Result := RxResConst_Slo.SWrongGIFColors;
     SBadGIFCodeSize : Result := RxResConst_Slo.SBadGIFCodeSize;
     SGIFDecodeError : Result := RxResConst_Slo.SGIFDecodeError;
     SGIFEncodeError : Result := RxResConst_Slo.SGIFEncodeError;
     SGIFVersion : Result := RxResConst_Slo.SGIFVersion;
     SDualListSrcCaption : Result := RxResConst_Slo.SDualListSrcCaption;
     SDualListDestCaption : Result := RxResConst_Slo.SDualListDestCaption;
     SClipbrdUnknown : Result := RxResConst_Slo.SClipbrdUnknown;
     SClipbrdEmpty : Result := RxResConst_Slo.SClipbrdEmpty;
     SCustomizeSpeedbar : Result := RxResConst_Slo.SCustomizeSpeedbar;
     SAvailButtons : Result := RxResConst_Slo.SAvailButtons;
     SSpeedbarCategories : Result := RxResConst_Slo.SSpeedbarCategories;
     SSpeedbarEditHint : Result := RxResConst_Slo.SSpeedbarEditHint;
     SParseSyntaxError : Result := RxResConst_Slo.SParseSyntaxError;
     SParseNotCramp : Result := RxResConst_Slo.SParseNotCramp;
     SParseDivideByZero : Result := RxResConst_Slo.SParseDivideByZero;
     SParseSqrError : Result := RxResConst_Slo.SParseSqrError;
     SParseLogError : Result := RxResConst_Slo.SParseLogError;
     SParseInvalidFloatOperation : Result := RxResConst_Slo.SParseInvalidFloatOperation;
     srRXControls : Result := RxResConst_Slo.srRXControls;
     srRXDBAware : Result := RxResConst_Slo.srRXDBAware;
     srRXTools : Result := RxResConst_Slo.srRXTools;
     srStorageDesigner : Result := RxResConst_Slo.srStorageDesigner;
     srProxyEditor : Result := RxResConst_Slo.srProxyEditor;
     srPageProxies : Result := RxResConst_Slo.srPageProxies;
     srProxyName : Result := RxResConst_Slo.srProxyName;
     srPageName : Result := RxResConst_Slo.srPageName;
     srSBItemNotCreate : Result := RxResConst_Slo.srSBItemNotCreate;
     srConfirmSBDelete : Result := RxResConst_Slo.srConfirmSBDelete;
     srSpeedbarDesigner : Result := RxResConst_Slo.srSpeedbarDesigner;
     srNewSectionName : Result := RxResConst_Slo.srNewSectionName;
     srEventNotCreate : Result := RxResConst_Slo.srEventNotCreate;
     srTimerDesigner : Result := RxResConst_Slo.srTimerDesigner;
     srTimerEvents : Result := RxResConst_Slo.srTimerEvents;
     srAniCurFilter : Result := RxResConst_Slo.srAniCurFilter;
     srEditPicture : Result := RxResConst_Slo.srEditPicture;
     srLoadAniCursor : Result := RxResConst_Slo.srLoadAniCursor;
     srLoadIcon : Result := RxResConst_Slo.srLoadIcon;
     srSaveImageList : Result := RxResConst_Slo.srSaveImageList;
     srCaptionDesigner : Result := RxResConst_Slo.srCaptionDesigner;
     srGradientCaptions : Result := RxResConst_Slo.srGradientCaptions;
     srBorrowStructure : Result := RxResConst_Slo.srBorrowStructure;
     ccTimeEditHours : Result := RxResConst_Slo.ccTimeEditHours;
     ccTimeEditMins : Result := RxResConst_Slo.ccTimeEditMins;
     ccTimeEditSecs : Result := RxResConst_Slo.ccTimeEditSecs;
     SColorBlack : Result := RxResConst_Slo.SColorBlack;
     SColorMaroon : Result := RxResConst_Slo.SColorMaroon;
     SColorGreen : Result := RxResConst_Slo.SColorGreen;
     SColorOlive : Result := RxResConst_Slo.SColorOlive;
     SColorNavy : Result := RxResConst_Slo.SColorNavy;
     SColorPurple : Result := RxResConst_Slo.SColorPurple;
     SColorTeal : Result := RxResConst_Slo.SColorTeal;
     SColorGray : Result := RxResConst_Slo.SColorGray;
     SColorSilver : Result := RxResConst_Slo.SColorSilver;
     SColorRed : Result := RxResConst_Slo.SColorRed;
     SColorLime : Result := RxResConst_Slo.SColorLime;
     SColorYellow : Result := RxResConst_Slo.SColorYellow;
     SColorBlue : Result := RxResConst_Slo.SColorBlue;
     SColorFuchsia : Result := RxResConst_Slo.SColorFuchsia;
     SColorAqua : Result := RxResConst_Slo.SColorAqua;
     SColorWhite : Result := RxResConst_Slo.SColorWhite;
     SColorScrollBar : Result := RxResConst_Slo.SColorScrollBar;
     SColorBackground : Result := RxResConst_Slo.SColorBackground;
     SColorActiveCaption : Result := RxResConst_Slo.SColorActiveCaption;
     SColorInactiveCaption : Result := RxResConst_Slo.SColorInactiveCaption;
     SColorMenu : Result := RxResConst_Slo.SColorMenu;
     SColorWindow : Result := RxResConst_Slo.SColorWindow;
     SColorWindowFrame : Result := RxResConst_Slo.SColorWindowFrame;
     SColorMenuText : Result := RxResConst_Slo.SColorMenuText;
     SColorWindowText : Result := RxResConst_Slo.SColorWindowText;
     SColorCaptionText : Result := RxResConst_Slo.SColorCaptionText;
     SColorActiveBorder : Result := RxResConst_Slo.SColorActiveBorder;
     SColorInactiveBorder : Result := RxResConst_Slo.SColorInactiveBorder;
     SColorAppWorkSpace : Result := RxResConst_Slo.SColorAppWorkSpace;
     SColorHighlight : Result := RxResConst_Slo.SColorHighlight;
     SColorHighlightText : Result := RxResConst_Slo.SColorHighlightText;
     SColorBtnFace : Result := RxResConst_Slo.SColorBtnFace;
     SColorBtnShadow : Result := RxResConst_Slo.SColorBtnShadow;
     SColorGrayText : Result := RxResConst_Slo.SColorGrayText;
     SColorBtnText : Result := RxResConst_Slo.SColorBtnText;
     SColorInactiveCaptionText : Result := RxResConst_Slo.SColorInactiveCaptionText;
     SColorBtnHighlight : Result := RxResConst_Slo.SColorBtnHighlight;
     SColor3DDkShadow : Result := RxResConst_Slo.SColor3DDkShadow;
     SColor3DLight : Result := RxResConst_Slo.SColor3DLight;
     SColorInfoText : Result := RxResConst_Slo.SColorInfoText;
     SColorInfoBk : Result := RxResConst_Slo.SColorInfoBk;
     SColorCream : Result := RxResConst_Slo.SColorCream;
     SColorMoneyGreen : Result := RxResConst_Slo.SColorMoneyGreen;
     SColorSkyBlue : Result := RxResConst_Slo.SColorSkyBlue;
     SColorNone : Result := RxResConst_Slo.SColorNone;
     SColorDefault : Result := RxResConst_Slo.SColorDefault;
     SColorCustom : Result := RxResConst_Slo.SColorCustom;
     SMemNoRecords : Result := RxResConst_Slo.SMemNoRecords;
     SInvalidFields : Result := RxResConst_Slo.SInvalidFields;
     SAutoSpeedbarMode : Result := RxResConst_Slo.SAutoSpeedbarMode;
    end;
  end;
  {$ENDIF}
  {$IFDEF RXLANG_Spa}
  function RxResConst_Spa_RxLoadStr(IDMsg: Integer): string;
  begin
    case IDMsg of
     SBrowse : Result := RxResConst_Spa.SBrowse;
     SDefaultFilter : Result := RxResConst_Spa.SDefaultFilter;
     SDateDlgTitle : Result := RxResConst_Spa.SDateDlgTitle;
     SNextYear : Result := RxResConst_Spa.SNextYear;
     SNextMonth : Result := RxResConst_Spa.SNextMonth;
     SPrevYear : Result := RxResConst_Spa.SPrevYear;
     SPrevMonth : Result := RxResConst_Spa.SPrevMonth;
     SNotImplemented : Result := RxResConst_Spa.SNotImplemented;
     SFileNotExec : Result := RxResConst_Spa.SFileNotExec;
     SLoadLibError : Result := RxResConst_Spa.SLoadLibError;
     SDetails : Result := RxResConst_Spa.SDetails;
     SDateOutOfRange : Result := RxResConst_Spa.SDateOutOfRange;
     SDateOutOfMin : Result := RxResConst_Spa.SDateOutOfMin;
     SDateOutOfMax : Result := RxResConst_Spa.SDateOutOfMax;
     SDateMinLimit : Result := RxResConst_Spa.SDateMinLimit;
     SDateMaxLimit : Result := RxResConst_Spa.SDateMaxLimit;
     SLocalDatabase : Result := RxResConst_Spa.SLocalDatabase;
     SRetryLogin : Result := RxResConst_Spa.SRetryLogin;
     SExprNotBoolean : Result := RxResConst_Spa.SExprNotBoolean;
     SExprBadNullTest : Result := RxResConst_Spa.SExprBadNullTest;
     SExprBadField : Result := RxResConst_Spa.SExprBadField;
     SCaptureFilter : Result := RxResConst_Spa.SCaptureFilter;
     SNotCaptureFilter : Result := RxResConst_Spa.SNotCaptureFilter;
     SInactiveData : Result := RxResConst_Spa.SInactiveData;
     SBrowseData : Result := RxResConst_Spa.SBrowseData;
     SEditData : Result := RxResConst_Spa.SEditData;
     SInsertData : Result := RxResConst_Spa.SInsertData;
     SSetKeyData : Result := RxResConst_Spa.SSetKeyData;
     SCalcFieldsData : Result := RxResConst_Spa.SCalcFieldsData;
     SRegistration : Result := RxResConst_Spa.SRegistration;
     SAppTitleLabel : Result := RxResConst_Spa.SAppTitleLabel;
     SHintLabel : Result := RxResConst_Spa.SHintLabel;
     SUserNameLabel : Result := RxResConst_Spa.SUserNameLabel;
     SPasswordLabel : Result := RxResConst_Spa.SPasswordLabel;
     SInvalidUserName : Result := RxResConst_Spa.SInvalidUserName;
     SChangePassword : Result := RxResConst_Spa.SChangePassword;
     SOldPasswordLabel : Result := RxResConst_Spa.SOldPasswordLabel;
     SNewPasswordLabel : Result := RxResConst_Spa.SNewPasswordLabel;
     SConfirmPasswordLabel : Result := RxResConst_Spa.SConfirmPasswordLabel;
     SPasswordChanged : Result := RxResConst_Spa.SPasswordChanged;
     SPasswordNotChanged : Result := RxResConst_Spa.SPasswordNotChanged;
     SPasswordsMismatch : Result := RxResConst_Spa.SPasswordsMismatch;
     SDBExceptCaption : Result := RxResConst_Spa.SDBExceptCaption;
     SBDEErrorLabel : Result := RxResConst_Spa.SBDEErrorLabel;
     SServerErrorLabel : Result := RxResConst_Spa.SServerErrorLabel;
     SErrorMsgLabel : Result := RxResConst_Spa.SErrorMsgLabel;
     SNextButton : Result := RxResConst_Spa.SNextButton;
     SPrevButton : Result := RxResConst_Spa.SPrevButton;
     SExprIncorrect : Result := RxResConst_Spa.SExprIncorrect;
     SExprTermination : Result := RxResConst_Spa.SExprTermination;
     SExprNameError : Result := RxResConst_Spa.SExprNameError;
     SExprStringError : Result := RxResConst_Spa.SExprStringError;
     SExprInvalidChar : Result := RxResConst_Spa.SExprInvalidChar;
     SExprNoRParen : Result := RxResConst_Spa.SExprNoRParen;
     SExprExpected : Result := RxResConst_Spa.SExprExpected;
     SExprBadCompare : Result := RxResConst_Spa.SExprBadCompare;
     SConfirmSave : Result := RxResConst_Spa.SConfirmSave;
     SDatabaseName : Result := RxResConst_Spa.SDatabaseName;
     SUnlockCaption : Result := RxResConst_Spa.SUnlockCaption;
     SUnlockHint : Result := RxResConst_Spa.SUnlockHint;
     SDeleteMultipleRecords : Result := RxResConst_Spa.SDeleteMultipleRecords;
     SGIFImage : Result := RxResConst_Spa.SGIFImage;
     SChangeGIFSize : Result := RxResConst_Spa.SChangeGIFSize;
     SNoGIFData : Result := RxResConst_Spa.SNoGIFData;
     SUnrecognizedGIFExt : Result := RxResConst_Spa.SUnrecognizedGIFExt;
     SWrongGIFColors : Result := RxResConst_Spa.SWrongGIFColors;
     SBadGIFCodeSize : Result := RxResConst_Spa.SBadGIFCodeSize;
     SGIFDecodeError : Result := RxResConst_Spa.SGIFDecodeError;
     SGIFEncodeError : Result := RxResConst_Spa.SGIFEncodeError;
     SGIFVersion : Result := RxResConst_Spa.SGIFVersion;
     SDualListSrcCaption : Result := RxResConst_Spa.SDualListSrcCaption;
     SDualListDestCaption : Result := RxResConst_Spa.SDualListDestCaption;
     SClipbrdUnknown : Result := RxResConst_Spa.SClipbrdUnknown;
     SClipbrdEmpty : Result := RxResConst_Spa.SClipbrdEmpty;
     SCustomizeSpeedbar : Result := RxResConst_Spa.SCustomizeSpeedbar;
     SAvailButtons : Result := RxResConst_Spa.SAvailButtons;
     SSpeedbarCategories : Result := RxResConst_Spa.SSpeedbarCategories;
     SSpeedbarEditHint : Result := RxResConst_Spa.SSpeedbarEditHint;
     SParseSyntaxError : Result := RxResConst_Spa.SParseSyntaxError;
     SParseNotCramp : Result := RxResConst_Spa.SParseNotCramp;
     SParseDivideByZero : Result := RxResConst_Spa.SParseDivideByZero;
     SParseSqrError : Result := RxResConst_Spa.SParseSqrError;
     SParseLogError : Result := RxResConst_Spa.SParseLogError;
     SParseInvalidFloatOperation : Result := RxResConst_Spa.SParseInvalidFloatOperation;
     srRXControls : Result := RxResConst_Spa.srRXControls;
     srRXDBAware : Result := RxResConst_Spa.srRXDBAware;
     srRXTools : Result := RxResConst_Spa.srRXTools;
     srStorageDesigner : Result := RxResConst_Spa.srStorageDesigner;
     srProxyEditor : Result := RxResConst_Spa.srProxyEditor;
     srPageProxies : Result := RxResConst_Spa.srPageProxies;
     srProxyName : Result := RxResConst_Spa.srProxyName;
     srPageName : Result := RxResConst_Spa.srPageName;
     srSBItemNotCreate : Result := RxResConst_Spa.srSBItemNotCreate;
     srConfirmSBDelete : Result := RxResConst_Spa.srConfirmSBDelete;
     srSpeedbarDesigner : Result := RxResConst_Spa.srSpeedbarDesigner;
     srNewSectionName : Result := RxResConst_Spa.srNewSectionName;
     srEventNotCreate : Result := RxResConst_Spa.srEventNotCreate;
     srTimerDesigner : Result := RxResConst_Spa.srTimerDesigner;
     srTimerEvents : Result := RxResConst_Spa.srTimerEvents;
     srAniCurFilter : Result := RxResConst_Spa.srAniCurFilter;
     srEditPicture : Result := RxResConst_Spa.srEditPicture;
     srLoadAniCursor : Result := RxResConst_Spa.srLoadAniCursor;
     srLoadIcon : Result := RxResConst_Spa.srLoadIcon;
     srSaveImageList : Result := RxResConst_Spa.srSaveImageList;
     srCaptionDesigner : Result := RxResConst_Spa.srCaptionDesigner;
     srGradientCaptions : Result := RxResConst_Spa.srGradientCaptions;
     srBorrowStructure : Result := RxResConst_Spa.srBorrowStructure;
     ccTimeEditHours : Result := RxResConst_Spa.ccTimeEditHours;
     ccTimeEditMins : Result := RxResConst_Spa.ccTimeEditMins;
     ccTimeEditSecs : Result := RxResConst_Spa.ccTimeEditSecs;
     SColorBlack : Result := RxResConst_Spa.SColorBlack;
     SColorMaroon : Result := RxResConst_Spa.SColorMaroon;
     SColorGreen : Result := RxResConst_Spa.SColorGreen;
     SColorOlive : Result := RxResConst_Spa.SColorOlive;
     SColorNavy : Result := RxResConst_Spa.SColorNavy;
     SColorPurple : Result := RxResConst_Spa.SColorPurple;
     SColorTeal : Result := RxResConst_Spa.SColorTeal;
     SColorGray : Result := RxResConst_Spa.SColorGray;
     SColorSilver : Result := RxResConst_Spa.SColorSilver;
     SColorRed : Result := RxResConst_Spa.SColorRed;
     SColorLime : Result := RxResConst_Spa.SColorLime;
     SColorYellow : Result := RxResConst_Spa.SColorYellow;
     SColorBlue : Result := RxResConst_Spa.SColorBlue;
     SColorFuchsia : Result := RxResConst_Spa.SColorFuchsia;
     SColorAqua : Result := RxResConst_Spa.SColorAqua;
     SColorWhite : Result := RxResConst_Spa.SColorWhite;
     SColorScrollBar : Result := RxResConst_Spa.SColorScrollBar;
     SColorBackground : Result := RxResConst_Spa.SColorBackground;
     SColorActiveCaption : Result := RxResConst_Spa.SColorActiveCaption;
     SColorInactiveCaption : Result := RxResConst_Spa.SColorInactiveCaption;
     SColorMenu : Result := RxResConst_Spa.SColorMenu;
     SColorWindow : Result := RxResConst_Spa.SColorWindow;
     SColorWindowFrame : Result := RxResConst_Spa.SColorWindowFrame;
     SColorMenuText : Result := RxResConst_Spa.SColorMenuText;
     SColorWindowText : Result := RxResConst_Spa.SColorWindowText;
     SColorCaptionText : Result := RxResConst_Spa.SColorCaptionText;
     SColorActiveBorder : Result := RxResConst_Spa.SColorActiveBorder;
     SColorInactiveBorder : Result := RxResConst_Spa.SColorInactiveBorder;
     SColorAppWorkSpace : Result := RxResConst_Spa.SColorAppWorkSpace;
     SColorHighlight : Result := RxResConst_Spa.SColorHighlight;
     SColorHighlightText : Result := RxResConst_Spa.SColorHighlightText;
     SColorBtnFace : Result := RxResConst_Spa.SColorBtnFace;
     SColorBtnShadow : Result := RxResConst_Spa.SColorBtnShadow;
     SColorGrayText : Result := RxResConst_Spa.SColorGrayText;
     SColorBtnText : Result := RxResConst_Spa.SColorBtnText;
     SColorInactiveCaptionText : Result := RxResConst_Spa.SColorInactiveCaptionText;
     SColorBtnHighlight : Result := RxResConst_Spa.SColorBtnHighlight;
     SColor3DDkShadow : Result := RxResConst_Spa.SColor3DDkShadow;
     SColor3DLight : Result := RxResConst_Spa.SColor3DLight;
     SColorInfoText : Result := RxResConst_Spa.SColorInfoText;
     SColorInfoBk : Result := RxResConst_Spa.SColorInfoBk;
     SColorCream : Result := RxResConst_Spa.SColorCream;
     SColorMoneyGreen : Result := RxResConst_Spa.SColorMoneyGreen;
     SColorSkyBlue : Result := RxResConst_Spa.SColorSkyBlue;
     SColorNone : Result := RxResConst_Spa.SColorNone;
     SColorDefault : Result := RxResConst_Spa.SColorDefault;
     SColorCustom : Result := RxResConst_Spa.SColorCustom;
     SMemNoRecords : Result := RxResConst_Spa.SMemNoRecords;
     SInvalidFields : Result := RxResConst_Spa.SInvalidFields;
     SAutoSpeedbarMode : Result := RxResConst_Spa.SAutoSpeedbarMode;
    end;
  end;
  {$ENDIF}
  {$IFDEF RXLANG_Srb}
  function RxResConst_Srb_RxLoadStr(IDMsg: Integer): string;
  begin
    case IDMsg of
     SBrowse : Result := RxResConst_Srb.SBrowse;
     SDefaultFilter : Result := RxResConst_Srb.SDefaultFilter;
     SDateDlgTitle : Result := RxResConst_Srb.SDateDlgTitle;
     SNextYear : Result := RxResConst_Srb.SNextYear;
     SNextMonth : Result := RxResConst_Srb.SNextMonth;
     SPrevYear : Result := RxResConst_Srb.SPrevYear;
     SPrevMonth : Result := RxResConst_Srb.SPrevMonth;
     SNotImplemented : Result := RxResConst_Srb.SNotImplemented;
     SFileNotExec : Result := RxResConst_Srb.SFileNotExec;
     SLoadLibError : Result := RxResConst_Srb.SLoadLibError;
     SDetails : Result := RxResConst_Srb.SDetails;
     SDateOutOfRange : Result := RxResConst_Srb.SDateOutOfRange;
     SDateOutOfMin : Result := RxResConst_Srb.SDateOutOfMin;
     SDateOutOfMax : Result := RxResConst_Srb.SDateOutOfMax;
     SDateMinLimit : Result := RxResConst_Srb.SDateMinLimit;
     SDateMaxLimit : Result := RxResConst_Srb.SDateMaxLimit;
     SLocalDatabase : Result := RxResConst_Srb.SLocalDatabase;
     SRetryLogin : Result := RxResConst_Srb.SRetryLogin;
     SExprNotBoolean : Result := RxResConst_Srb.SExprNotBoolean;
     SExprBadNullTest : Result := RxResConst_Srb.SExprBadNullTest;
     SExprBadField : Result := RxResConst_Srb.SExprBadField;
     SCaptureFilter : Result := RxResConst_Srb.SCaptureFilter;
     SNotCaptureFilter : Result := RxResConst_Srb.SNotCaptureFilter;
     SInactiveData : Result := RxResConst_Srb.SInactiveData;
     SBrowseData : Result := RxResConst_Srb.SBrowseData;
     SEditData : Result := RxResConst_Srb.SEditData;
     SInsertData : Result := RxResConst_Srb.SInsertData;
     SSetKeyData : Result := RxResConst_Srb.SSetKeyData;
     SCalcFieldsData : Result := RxResConst_Srb.SCalcFieldsData;
     SRegistration : Result := RxResConst_Srb.SRegistration;
     SAppTitleLabel : Result := RxResConst_Srb.SAppTitleLabel;
     SHintLabel : Result := RxResConst_Srb.SHintLabel;
     SUserNameLabel : Result := RxResConst_Srb.SUserNameLabel;
     SPasswordLabel : Result := RxResConst_Srb.SPasswordLabel;
     SInvalidUserName : Result := RxResConst_Srb.SInvalidUserName;
     SChangePassword : Result := RxResConst_Srb.SChangePassword;
     SOldPasswordLabel : Result := RxResConst_Srb.SOldPasswordLabel;
     SNewPasswordLabel : Result := RxResConst_Srb.SNewPasswordLabel;
     SConfirmPasswordLabel : Result := RxResConst_Srb.SConfirmPasswordLabel;
     SPasswordChanged : Result := RxResConst_Srb.SPasswordChanged;
     SPasswordNotChanged : Result := RxResConst_Srb.SPasswordNotChanged;
     SPasswordsMismatch : Result := RxResConst_Srb.SPasswordsMismatch;
     SDBExceptCaption : Result := RxResConst_Srb.SDBExceptCaption;
     SBDEErrorLabel : Result := RxResConst_Srb.SBDEErrorLabel;
     SServerErrorLabel : Result := RxResConst_Srb.SServerErrorLabel;
     SErrorMsgLabel : Result := RxResConst_Srb.SErrorMsgLabel;
     SNextButton : Result := RxResConst_Srb.SNextButton;
     SPrevButton : Result := RxResConst_Srb.SPrevButton;
     SExprIncorrect : Result := RxResConst_Srb.SExprIncorrect;
     SExprTermination : Result := RxResConst_Srb.SExprTermination;
     SExprNameError : Result := RxResConst_Srb.SExprNameError;
     SExprStringError : Result := RxResConst_Srb.SExprStringError;
     SExprInvalidChar : Result := RxResConst_Srb.SExprInvalidChar;
     SExprNoRParen : Result := RxResConst_Srb.SExprNoRParen;
     SExprExpected : Result := RxResConst_Srb.SExprExpected;
     SExprBadCompare : Result := RxResConst_Srb.SExprBadCompare;
     SConfirmSave : Result := RxResConst_Srb.SConfirmSave;
     SDatabaseName : Result := RxResConst_Srb.SDatabaseName;
     SUnlockCaption : Result := RxResConst_Srb.SUnlockCaption;
     SUnlockHint : Result := RxResConst_Srb.SUnlockHint;
     SDeleteMultipleRecords : Result := RxResConst_Srb.SDeleteMultipleRecords;
     SGIFImage : Result := RxResConst_Srb.SGIFImage;
     SChangeGIFSize : Result := RxResConst_Srb.SChangeGIFSize;
     SNoGIFData : Result := RxResConst_Srb.SNoGIFData;
     SUnrecognizedGIFExt : Result := RxResConst_Srb.SUnrecognizedGIFExt;
     SWrongGIFColors : Result := RxResConst_Srb.SWrongGIFColors;
     SBadGIFCodeSize : Result := RxResConst_Srb.SBadGIFCodeSize;
     SGIFDecodeError : Result := RxResConst_Srb.SGIFDecodeError;
     SGIFEncodeError : Result := RxResConst_Srb.SGIFEncodeError;
     SGIFVersion : Result := RxResConst_Srb.SGIFVersion;
     SDualListSrcCaption : Result := RxResConst_Srb.SDualListSrcCaption;
     SDualListDestCaption : Result := RxResConst_Srb.SDualListDestCaption;
     SClipbrdUnknown : Result := RxResConst_Srb.SClipbrdUnknown;
     SClipbrdEmpty : Result := RxResConst_Srb.SClipbrdEmpty;
     SCustomizeSpeedbar : Result := RxResConst_Srb.SCustomizeSpeedbar;
     SAvailButtons : Result := RxResConst_Srb.SAvailButtons;
     SSpeedbarCategories : Result := RxResConst_Srb.SSpeedbarCategories;
     SSpeedbarEditHint : Result := RxResConst_Srb.SSpeedbarEditHint;
     SParseSyntaxError : Result := RxResConst_Srb.SParseSyntaxError;
     SParseNotCramp : Result := RxResConst_Srb.SParseNotCramp;
     SParseDivideByZero : Result := RxResConst_Srb.SParseDivideByZero;
     SParseSqrError : Result := RxResConst_Srb.SParseSqrError;
     SParseLogError : Result := RxResConst_Srb.SParseLogError;
     SParseInvalidFloatOperation : Result := RxResConst_Srb.SParseInvalidFloatOperation;
     srRXControls : Result := RxResConst_Srb.srRXControls;
     srRXDBAware : Result := RxResConst_Srb.srRXDBAware;
     srRXTools : Result := RxResConst_Srb.srRXTools;
     srStorageDesigner : Result := RxResConst_Srb.srStorageDesigner;
     srProxyEditor : Result := RxResConst_Srb.srProxyEditor;
     srPageProxies : Result := RxResConst_Srb.srPageProxies;
     srProxyName : Result := RxResConst_Srb.srProxyName;
     srPageName : Result := RxResConst_Srb.srPageName;
     srSBItemNotCreate : Result := RxResConst_Srb.srSBItemNotCreate;
     srConfirmSBDelete : Result := RxResConst_Srb.srConfirmSBDelete;
     srSpeedbarDesigner : Result := RxResConst_Srb.srSpeedbarDesigner;
     srNewSectionName : Result := RxResConst_Srb.srNewSectionName;
     srEventNotCreate : Result := RxResConst_Srb.srEventNotCreate;
     srTimerDesigner : Result := RxResConst_Srb.srTimerDesigner;
     srTimerEvents : Result := RxResConst_Srb.srTimerEvents;
     srAniCurFilter : Result := RxResConst_Srb.srAniCurFilter;
     srEditPicture : Result := RxResConst_Srb.srEditPicture;
     srLoadAniCursor : Result := RxResConst_Srb.srLoadAniCursor;
     srLoadIcon : Result := RxResConst_Srb.srLoadIcon;
     srSaveImageList : Result := RxResConst_Srb.srSaveImageList;
     srCaptionDesigner : Result := RxResConst_Srb.srCaptionDesigner;
     srGradientCaptions : Result := RxResConst_Srb.srGradientCaptions;
     srBorrowStructure : Result := RxResConst_Srb.srBorrowStructure;
     ccTimeEditHours : Result := RxResConst_Srb.ccTimeEditHours;
     ccTimeEditMins : Result := RxResConst_Srb.ccTimeEditMins;
     ccTimeEditSecs : Result := RxResConst_Srb.ccTimeEditSecs;
     SColorBlack : Result := RxResConst_Srb.SColorBlack;
     SColorMaroon : Result := RxResConst_Srb.SColorMaroon;
     SColorGreen : Result := RxResConst_Srb.SColorGreen;
     SColorOlive : Result := RxResConst_Srb.SColorOlive;
     SColorNavy : Result := RxResConst_Srb.SColorNavy;
     SColorPurple : Result := RxResConst_Srb.SColorPurple;
     SColorTeal : Result := RxResConst_Srb.SColorTeal;
     SColorGray : Result := RxResConst_Srb.SColorGray;
     SColorSilver : Result := RxResConst_Srb.SColorSilver;
     SColorRed : Result := RxResConst_Srb.SColorRed;
     SColorLime : Result := RxResConst_Srb.SColorLime;
     SColorYellow : Result := RxResConst_Srb.SColorYellow;
     SColorBlue : Result := RxResConst_Srb.SColorBlue;
     SColorFuchsia : Result := RxResConst_Srb.SColorFuchsia;
     SColorAqua : Result := RxResConst_Srb.SColorAqua;
     SColorWhite : Result := RxResConst_Srb.SColorWhite;
     SColorScrollBar : Result := RxResConst_Srb.SColorScrollBar;
     SColorBackground : Result := RxResConst_Srb.SColorBackground;
     SColorActiveCaption : Result := RxResConst_Srb.SColorActiveCaption;
     SColorInactiveCaption : Result := RxResConst_Srb.SColorInactiveCaption;
     SColorMenu : Result := RxResConst_Srb.SColorMenu;
     SColorWindow : Result := RxResConst_Srb.SColorWindow;
     SColorWindowFrame : Result := RxResConst_Srb.SColorWindowFrame;
     SColorMenuText : Result := RxResConst_Srb.SColorMenuText;
     SColorWindowText : Result := RxResConst_Srb.SColorWindowText;
     SColorCaptionText : Result := RxResConst_Srb.SColorCaptionText;
     SColorActiveBorder : Result := RxResConst_Srb.SColorActiveBorder;
     SColorInactiveBorder : Result := RxResConst_Srb.SColorInactiveBorder;
     SColorAppWorkSpace : Result := RxResConst_Srb.SColorAppWorkSpace;
     SColorHighlight : Result := RxResConst_Srb.SColorHighlight;
     SColorHighlightText : Result := RxResConst_Srb.SColorHighlightText;
     SColorBtnFace : Result := RxResConst_Srb.SColorBtnFace;
     SColorBtnShadow : Result := RxResConst_Srb.SColorBtnShadow;
     SColorGrayText : Result := RxResConst_Srb.SColorGrayText;
     SColorBtnText : Result := RxResConst_Srb.SColorBtnText;
     SColorInactiveCaptionText : Result := RxResConst_Srb.SColorInactiveCaptionText;
     SColorBtnHighlight : Result := RxResConst_Srb.SColorBtnHighlight;
     SColor3DDkShadow : Result := RxResConst_Srb.SColor3DDkShadow;
     SColor3DLight : Result := RxResConst_Srb.SColor3DLight;
     SColorInfoText : Result := RxResConst_Srb.SColorInfoText;
     SColorInfoBk : Result := RxResConst_Srb.SColorInfoBk;
     SColorCream : Result := RxResConst_Srb.SColorCream;
     SColorMoneyGreen : Result := RxResConst_Srb.SColorMoneyGreen;
     SColorSkyBlue : Result := RxResConst_Srb.SColorSkyBlue;
     SColorNone : Result := RxResConst_Srb.SColorNone;
     SColorDefault : Result := RxResConst_Srb.SColorDefault;
     SColorCustom : Result := RxResConst_Srb.SColorCustom;
     SMemNoRecords : Result := RxResConst_Srb.SMemNoRecords;
     SInvalidFields : Result := RxResConst_Srb.SInvalidFields;
     SAutoSpeedbarMode : Result := RxResConst_Srb.SAutoSpeedbarMode;
    end;
  end;
  {$ENDIF}
  {$IFDEF RXLANG_Svk}
  function RxResConst_Svk_RxLoadStr(IDMsg: Integer): string;
  begin
    case IDMsg of
     SBrowse : Result := RxResConst_Svk.SBrowse;
     SDefaultFilter : Result := RxResConst_Svk.SDefaultFilter;
     SDateDlgTitle : Result := RxResConst_Svk.SDateDlgTitle;
     SNextYear : Result := RxResConst_Svk.SNextYear;
     SNextMonth : Result := RxResConst_Svk.SNextMonth;
     SPrevYear : Result := RxResConst_Svk.SPrevYear;
     SPrevMonth : Result := RxResConst_Svk.SPrevMonth;
     SNotImplemented : Result := RxResConst_Svk.SNotImplemented;
     SFileNotExec : Result := RxResConst_Svk.SFileNotExec;
     SLoadLibError : Result := RxResConst_Svk.SLoadLibError;
     SDetails : Result := RxResConst_Svk.SDetails;
     SDateOutOfRange : Result := RxResConst_Svk.SDateOutOfRange;
     SDateOutOfMin : Result := RxResConst_Svk.SDateOutOfMin;
     SDateOutOfMax : Result := RxResConst_Svk.SDateOutOfMax;
     SDateMinLimit : Result := RxResConst_Svk.SDateMinLimit;
     SDateMaxLimit : Result := RxResConst_Svk.SDateMaxLimit;
     SLocalDatabase : Result := RxResConst_Svk.SLocalDatabase;
     SRetryLogin : Result := RxResConst_Svk.SRetryLogin;
     SExprNotBoolean : Result := RxResConst_Svk.SExprNotBoolean;
     SExprBadNullTest : Result := RxResConst_Svk.SExprBadNullTest;
     SExprBadField : Result := RxResConst_Svk.SExprBadField;
     SCaptureFilter : Result := RxResConst_Svk.SCaptureFilter;
     SNotCaptureFilter : Result := RxResConst_Svk.SNotCaptureFilter;
     SInactiveData : Result := RxResConst_Svk.SInactiveData;
     SBrowseData : Result := RxResConst_Svk.SBrowseData;
     SEditData : Result := RxResConst_Svk.SEditData;
     SInsertData : Result := RxResConst_Svk.SInsertData;
     SSetKeyData : Result := RxResConst_Svk.SSetKeyData;
     SCalcFieldsData : Result := RxResConst_Svk.SCalcFieldsData;
     SRegistration : Result := RxResConst_Svk.SRegistration;
     SAppTitleLabel : Result := RxResConst_Svk.SAppTitleLabel;
     SHintLabel : Result := RxResConst_Svk.SHintLabel;
     SUserNameLabel : Result := RxResConst_Svk.SUserNameLabel;
     SPasswordLabel : Result := RxResConst_Svk.SPasswordLabel;
     SInvalidUserName : Result := RxResConst_Svk.SInvalidUserName;
     SChangePassword : Result := RxResConst_Svk.SChangePassword;
     SOldPasswordLabel : Result := RxResConst_Svk.SOldPasswordLabel;
     SNewPasswordLabel : Result := RxResConst_Svk.SNewPasswordLabel;
     SConfirmPasswordLabel : Result := RxResConst_Svk.SConfirmPasswordLabel;
     SPasswordChanged : Result := RxResConst_Svk.SPasswordChanged;
     SPasswordNotChanged : Result := RxResConst_Svk.SPasswordNotChanged;
     SPasswordsMismatch : Result := RxResConst_Svk.SPasswordsMismatch;
     SDBExceptCaption : Result := RxResConst_Svk.SDBExceptCaption;
     SBDEErrorLabel : Result := RxResConst_Svk.SBDEErrorLabel;
     SServerErrorLabel : Result := RxResConst_Svk.SServerErrorLabel;
     SErrorMsgLabel : Result := RxResConst_Svk.SErrorMsgLabel;
     SNextButton : Result := RxResConst_Svk.SNextButton;
     SPrevButton : Result := RxResConst_Svk.SPrevButton;
     SExprIncorrect : Result := RxResConst_Svk.SExprIncorrect;
     SExprTermination : Result := RxResConst_Svk.SExprTermination;
     SExprNameError : Result := RxResConst_Svk.SExprNameError;
     SExprStringError : Result := RxResConst_Svk.SExprStringError;
     SExprInvalidChar : Result := RxResConst_Svk.SExprInvalidChar;
     SExprNoRParen : Result := RxResConst_Svk.SExprNoRParen;
     SExprExpected : Result := RxResConst_Svk.SExprExpected;
     SExprBadCompare : Result := RxResConst_Svk.SExprBadCompare;
     SConfirmSave : Result := RxResConst_Svk.SConfirmSave;
     SDatabaseName : Result := RxResConst_Svk.SDatabaseName;
     SUnlockCaption : Result := RxResConst_Svk.SUnlockCaption;
     SUnlockHint : Result := RxResConst_Svk.SUnlockHint;
     SDeleteMultipleRecords : Result := RxResConst_Svk.SDeleteMultipleRecords;
     SGIFImage : Result := RxResConst_Svk.SGIFImage;
     SChangeGIFSize : Result := RxResConst_Svk.SChangeGIFSize;
     SNoGIFData : Result := RxResConst_Svk.SNoGIFData;
     SUnrecognizedGIFExt : Result := RxResConst_Svk.SUnrecognizedGIFExt;
     SWrongGIFColors : Result := RxResConst_Svk.SWrongGIFColors;
     SBadGIFCodeSize : Result := RxResConst_Svk.SBadGIFCodeSize;
     SGIFDecodeError : Result := RxResConst_Svk.SGIFDecodeError;
     SGIFEncodeError : Result := RxResConst_Svk.SGIFEncodeError;
     SGIFVersion : Result := RxResConst_Svk.SGIFVersion;
     SDualListSrcCaption : Result := RxResConst_Svk.SDualListSrcCaption;
     SDualListDestCaption : Result := RxResConst_Svk.SDualListDestCaption;
     SClipbrdUnknown : Result := RxResConst_Svk.SClipbrdUnknown;
     SClipbrdEmpty : Result := RxResConst_Svk.SClipbrdEmpty;
     SCustomizeSpeedbar : Result := RxResConst_Svk.SCustomizeSpeedbar;
     SAvailButtons : Result := RxResConst_Svk.SAvailButtons;
     SSpeedbarCategories : Result := RxResConst_Svk.SSpeedbarCategories;
     SSpeedbarEditHint : Result := RxResConst_Svk.SSpeedbarEditHint;
     SParseSyntaxError : Result := RxResConst_Svk.SParseSyntaxError;
     SParseNotCramp : Result := RxResConst_Svk.SParseNotCramp;
     SParseDivideByZero : Result := RxResConst_Svk.SParseDivideByZero;
     SParseSqrError : Result := RxResConst_Svk.SParseSqrError;
     SParseLogError : Result := RxResConst_Svk.SParseLogError;
     SParseInvalidFloatOperation : Result := RxResConst_Svk.SParseInvalidFloatOperation;
     srRXControls : Result := RxResConst_Svk.srRXControls;
     srRXDBAware : Result := RxResConst_Svk.srRXDBAware;
     srRXTools : Result := RxResConst_Svk.srRXTools;
     srStorageDesigner : Result := RxResConst_Svk.srStorageDesigner;
     srProxyEditor : Result := RxResConst_Svk.srProxyEditor;
     srPageProxies : Result := RxResConst_Svk.srPageProxies;
     srProxyName : Result := RxResConst_Svk.srProxyName;
     srPageName : Result := RxResConst_Svk.srPageName;
     srSBItemNotCreate : Result := RxResConst_Svk.srSBItemNotCreate;
     srConfirmSBDelete : Result := RxResConst_Svk.srConfirmSBDelete;
     srSpeedbarDesigner : Result := RxResConst_Svk.srSpeedbarDesigner;
     srNewSectionName : Result := RxResConst_Svk.srNewSectionName;
     srEventNotCreate : Result := RxResConst_Svk.srEventNotCreate;
     srTimerDesigner : Result := RxResConst_Svk.srTimerDesigner;
     srTimerEvents : Result := RxResConst_Svk.srTimerEvents;
     srAniCurFilter : Result := RxResConst_Svk.srAniCurFilter;
     srEditPicture : Result := RxResConst_Svk.srEditPicture;
     srLoadAniCursor : Result := RxResConst_Svk.srLoadAniCursor;
     srLoadIcon : Result := RxResConst_Svk.srLoadIcon;
     srSaveImageList : Result := RxResConst_Svk.srSaveImageList;
     srCaptionDesigner : Result := RxResConst_Svk.srCaptionDesigner;
     srGradientCaptions : Result := RxResConst_Svk.srGradientCaptions;
     srBorrowStructure : Result := RxResConst_Svk.srBorrowStructure;
     ccTimeEditHours : Result := RxResConst_Svk.ccTimeEditHours;
     ccTimeEditMins : Result := RxResConst_Svk.ccTimeEditMins;
     ccTimeEditSecs : Result := RxResConst_Svk.ccTimeEditSecs;
     SColorBlack : Result := RxResConst_Svk.SColorBlack;
     SColorMaroon : Result := RxResConst_Svk.SColorMaroon;
     SColorGreen : Result := RxResConst_Svk.SColorGreen;
     SColorOlive : Result := RxResConst_Svk.SColorOlive;
     SColorNavy : Result := RxResConst_Svk.SColorNavy;
     SColorPurple : Result := RxResConst_Svk.SColorPurple;
     SColorTeal : Result := RxResConst_Svk.SColorTeal;
     SColorGray : Result := RxResConst_Svk.SColorGray;
     SColorSilver : Result := RxResConst_Svk.SColorSilver;
     SColorRed : Result := RxResConst_Svk.SColorRed;
     SColorLime : Result := RxResConst_Svk.SColorLime;
     SColorYellow : Result := RxResConst_Svk.SColorYellow;
     SColorBlue : Result := RxResConst_Svk.SColorBlue;
     SColorFuchsia : Result := RxResConst_Svk.SColorFuchsia;
     SColorAqua : Result := RxResConst_Svk.SColorAqua;
     SColorWhite : Result := RxResConst_Svk.SColorWhite;
     SColorScrollBar : Result := RxResConst_Svk.SColorScrollBar;
     SColorBackground : Result := RxResConst_Svk.SColorBackground;
     SColorActiveCaption : Result := RxResConst_Svk.SColorActiveCaption;
     SColorInactiveCaption : Result := RxResConst_Svk.SColorInactiveCaption;
     SColorMenu : Result := RxResConst_Svk.SColorMenu;
     SColorWindow : Result := RxResConst_Svk.SColorWindow;
     SColorWindowFrame : Result := RxResConst_Svk.SColorWindowFrame;
     SColorMenuText : Result := RxResConst_Svk.SColorMenuText;
     SColorWindowText : Result := RxResConst_Svk.SColorWindowText;
     SColorCaptionText : Result := RxResConst_Svk.SColorCaptionText;
     SColorActiveBorder : Result := RxResConst_Svk.SColorActiveBorder;
     SColorInactiveBorder : Result := RxResConst_Svk.SColorInactiveBorder;
     SColorAppWorkSpace : Result := RxResConst_Svk.SColorAppWorkSpace;
     SColorHighlight : Result := RxResConst_Svk.SColorHighlight;
     SColorHighlightText : Result := RxResConst_Svk.SColorHighlightText;
     SColorBtnFace : Result := RxResConst_Svk.SColorBtnFace;
     SColorBtnShadow : Result := RxResConst_Svk.SColorBtnShadow;
     SColorGrayText : Result := RxResConst_Svk.SColorGrayText;
     SColorBtnText : Result := RxResConst_Svk.SColorBtnText;
     SColorInactiveCaptionText : Result := RxResConst_Svk.SColorInactiveCaptionText;
     SColorBtnHighlight : Result := RxResConst_Svk.SColorBtnHighlight;
     SColor3DDkShadow : Result := RxResConst_Svk.SColor3DDkShadow;
     SColor3DLight : Result := RxResConst_Svk.SColor3DLight;
     SColorInfoText : Result := RxResConst_Svk.SColorInfoText;
     SColorInfoBk : Result := RxResConst_Svk.SColorInfoBk;
     SColorCream : Result := RxResConst_Svk.SColorCream;
     SColorMoneyGreen : Result := RxResConst_Svk.SColorMoneyGreen;
     SColorSkyBlue : Result := RxResConst_Svk.SColorSkyBlue;
     SColorNone : Result := RxResConst_Svk.SColorNone;
     SColorDefault : Result := RxResConst_Svk.SColorDefault;
     SColorCustom : Result := RxResConst_Svk.SColorCustom;
     SMemNoRecords : Result := RxResConst_Svk.SMemNoRecords;
     SInvalidFields : Result := RxResConst_Svk.SInvalidFields;
     SAutoSpeedbarMode : Result := RxResConst_Svk.SAutoSpeedbarMode;
    end;
  end;
  {$ENDIF}
  {$IFDEF RXLANG_Swe}
  function RxResConst_Swe_RxLoadStr(IDMsg: Integer): string;
  begin
    case IDMsg of
     SBrowse : Result := RxResConst_Swe.SBrowse;
     SDefaultFilter : Result := RxResConst_Swe.SDefaultFilter;
     SDateDlgTitle : Result := RxResConst_Swe.SDateDlgTitle;
     SNextYear : Result := RxResConst_Swe.SNextYear;
     SNextMonth : Result := RxResConst_Swe.SNextMonth;
     SPrevYear : Result := RxResConst_Swe.SPrevYear;
     SPrevMonth : Result := RxResConst_Swe.SPrevMonth;
     SNotImplemented : Result := RxResConst_Swe.SNotImplemented;
     SFileNotExec : Result := RxResConst_Swe.SFileNotExec;
     SLoadLibError : Result := RxResConst_Swe.SLoadLibError;
     SDetails : Result := RxResConst_Swe.SDetails;
     SDateOutOfRange : Result := RxResConst_Swe.SDateOutOfRange;
     SDateOutOfMin : Result := RxResConst_Swe.SDateOutOfMin;
     SDateOutOfMax : Result := RxResConst_Swe.SDateOutOfMax;
     SDateMinLimit : Result := RxResConst_Swe.SDateMinLimit;
     SDateMaxLimit : Result := RxResConst_Swe.SDateMaxLimit;
     SLocalDatabase : Result := RxResConst_Swe.SLocalDatabase;
     SRetryLogin : Result := RxResConst_Swe.SRetryLogin;
     SExprNotBoolean : Result := RxResConst_Swe.SExprNotBoolean;
     SExprBadNullTest : Result := RxResConst_Swe.SExprBadNullTest;
     SExprBadField : Result := RxResConst_Swe.SExprBadField;
     SCaptureFilter : Result := RxResConst_Swe.SCaptureFilter;
     SNotCaptureFilter : Result := RxResConst_Swe.SNotCaptureFilter;
     SInactiveData : Result := RxResConst_Swe.SInactiveData;
     SBrowseData : Result := RxResConst_Swe.SBrowseData;
     SEditData : Result := RxResConst_Swe.SEditData;
     SInsertData : Result := RxResConst_Swe.SInsertData;
     SSetKeyData : Result := RxResConst_Swe.SSetKeyData;
     SCalcFieldsData : Result := RxResConst_Swe.SCalcFieldsData;
     SRegistration : Result := RxResConst_Swe.SRegistration;
     SAppTitleLabel : Result := RxResConst_Swe.SAppTitleLabel;
     SHintLabel : Result := RxResConst_Swe.SHintLabel;
     SUserNameLabel : Result := RxResConst_Swe.SUserNameLabel;
     SPasswordLabel : Result := RxResConst_Swe.SPasswordLabel;
     SInvalidUserName : Result := RxResConst_Swe.SInvalidUserName;
     SChangePassword : Result := RxResConst_Swe.SChangePassword;
     SOldPasswordLabel : Result := RxResConst_Swe.SOldPasswordLabel;
     SNewPasswordLabel : Result := RxResConst_Swe.SNewPasswordLabel;
     SConfirmPasswordLabel : Result := RxResConst_Swe.SConfirmPasswordLabel;
     SPasswordChanged : Result := RxResConst_Swe.SPasswordChanged;
     SPasswordNotChanged : Result := RxResConst_Swe.SPasswordNotChanged;
     SPasswordsMismatch : Result := RxResConst_Swe.SPasswordsMismatch;
     SDBExceptCaption : Result := RxResConst_Swe.SDBExceptCaption;
     SBDEErrorLabel : Result := RxResConst_Swe.SBDEErrorLabel;
     SServerErrorLabel : Result := RxResConst_Swe.SServerErrorLabel;
     SErrorMsgLabel : Result := RxResConst_Swe.SErrorMsgLabel;
     SNextButton : Result := RxResConst_Swe.SNextButton;
     SPrevButton : Result := RxResConst_Swe.SPrevButton;
     SExprIncorrect : Result := RxResConst_Swe.SExprIncorrect;
     SExprTermination : Result := RxResConst_Swe.SExprTermination;
     SExprNameError : Result := RxResConst_Swe.SExprNameError;
     SExprStringError : Result := RxResConst_Swe.SExprStringError;
     SExprInvalidChar : Result := RxResConst_Swe.SExprInvalidChar;
     SExprNoRParen : Result := RxResConst_Swe.SExprNoRParen;
     SExprExpected : Result := RxResConst_Swe.SExprExpected;
     SExprBadCompare : Result := RxResConst_Swe.SExprBadCompare;
     SConfirmSave : Result := RxResConst_Swe.SConfirmSave;
     SDatabaseName : Result := RxResConst_Swe.SDatabaseName;
     SUnlockCaption : Result := RxResConst_Swe.SUnlockCaption;
     SUnlockHint : Result := RxResConst_Swe.SUnlockHint;
     SDeleteMultipleRecords : Result := RxResConst_Swe.SDeleteMultipleRecords;
     SGIFImage : Result := RxResConst_Swe.SGIFImage;
     SChangeGIFSize : Result := RxResConst_Swe.SChangeGIFSize;
     SNoGIFData : Result := RxResConst_Swe.SNoGIFData;
     SUnrecognizedGIFExt : Result := RxResConst_Swe.SUnrecognizedGIFExt;
     SWrongGIFColors : Result := RxResConst_Swe.SWrongGIFColors;
     SBadGIFCodeSize : Result := RxResConst_Swe.SBadGIFCodeSize;
     SGIFDecodeError : Result := RxResConst_Swe.SGIFDecodeError;
     SGIFEncodeError : Result := RxResConst_Swe.SGIFEncodeError;
     SGIFVersion : Result := RxResConst_Swe.SGIFVersion;
     SDualListSrcCaption : Result := RxResConst_Swe.SDualListSrcCaption;
     SDualListDestCaption : Result := RxResConst_Swe.SDualListDestCaption;
     SClipbrdUnknown : Result := RxResConst_Swe.SClipbrdUnknown;
     SClipbrdEmpty : Result := RxResConst_Swe.SClipbrdEmpty;
     SCustomizeSpeedbar : Result := RxResConst_Swe.SCustomizeSpeedbar;
     SAvailButtons : Result := RxResConst_Swe.SAvailButtons;
     SSpeedbarCategories : Result := RxResConst_Swe.SSpeedbarCategories;
     SSpeedbarEditHint : Result := RxResConst_Swe.SSpeedbarEditHint;
     SParseSyntaxError : Result := RxResConst_Swe.SParseSyntaxError;
     SParseNotCramp : Result := RxResConst_Swe.SParseNotCramp;
     SParseDivideByZero : Result := RxResConst_Swe.SParseDivideByZero;
     SParseSqrError : Result := RxResConst_Swe.SParseSqrError;
     SParseLogError : Result := RxResConst_Swe.SParseLogError;
     SParseInvalidFloatOperation : Result := RxResConst_Swe.SParseInvalidFloatOperation;
     srRXControls : Result := RxResConst_Swe.srRXControls;
     srRXDBAware : Result := RxResConst_Swe.srRXDBAware;
     srRXTools : Result := RxResConst_Swe.srRXTools;
     srStorageDesigner : Result := RxResConst_Swe.srStorageDesigner;
     srProxyEditor : Result := RxResConst_Swe.srProxyEditor;
     srPageProxies : Result := RxResConst_Swe.srPageProxies;
     srProxyName : Result := RxResConst_Swe.srProxyName;
     srPageName : Result := RxResConst_Swe.srPageName;
     srSBItemNotCreate : Result := RxResConst_Swe.srSBItemNotCreate;
     srConfirmSBDelete : Result := RxResConst_Swe.srConfirmSBDelete;
     srSpeedbarDesigner : Result := RxResConst_Swe.srSpeedbarDesigner;
     srNewSectionName : Result := RxResConst_Swe.srNewSectionName;
     srEventNotCreate : Result := RxResConst_Swe.srEventNotCreate;
     srTimerDesigner : Result := RxResConst_Swe.srTimerDesigner;
     srTimerEvents : Result := RxResConst_Swe.srTimerEvents;
     srAniCurFilter : Result := RxResConst_Swe.srAniCurFilter;
     srEditPicture : Result := RxResConst_Swe.srEditPicture;
     srLoadAniCursor : Result := RxResConst_Swe.srLoadAniCursor;
     srLoadIcon : Result := RxResConst_Swe.srLoadIcon;
     srSaveImageList : Result := RxResConst_Swe.srSaveImageList;
     srCaptionDesigner : Result := RxResConst_Swe.srCaptionDesigner;
     srGradientCaptions : Result := RxResConst_Swe.srGradientCaptions;
     srBorrowStructure : Result := RxResConst_Swe.srBorrowStructure;
     ccTimeEditHours : Result := RxResConst_Swe.ccTimeEditHours;
     ccTimeEditMins : Result := RxResConst_Swe.ccTimeEditMins;
     ccTimeEditSecs : Result := RxResConst_Swe.ccTimeEditSecs;
     SColorBlack : Result := RxResConst_Swe.SColorBlack;
     SColorMaroon : Result := RxResConst_Swe.SColorMaroon;
     SColorGreen : Result := RxResConst_Swe.SColorGreen;
     SColorOlive : Result := RxResConst_Swe.SColorOlive;
     SColorNavy : Result := RxResConst_Swe.SColorNavy;
     SColorPurple : Result := RxResConst_Swe.SColorPurple;
     SColorTeal : Result := RxResConst_Swe.SColorTeal;
     SColorGray : Result := RxResConst_Swe.SColorGray;
     SColorSilver : Result := RxResConst_Swe.SColorSilver;
     SColorRed : Result := RxResConst_Swe.SColorRed;
     SColorLime : Result := RxResConst_Swe.SColorLime;
     SColorYellow : Result := RxResConst_Swe.SColorYellow;
     SColorBlue : Result := RxResConst_Swe.SColorBlue;
     SColorFuchsia : Result := RxResConst_Swe.SColorFuchsia;
     SColorAqua : Result := RxResConst_Swe.SColorAqua;
     SColorWhite : Result := RxResConst_Swe.SColorWhite;
     SColorScrollBar : Result := RxResConst_Swe.SColorScrollBar;
     SColorBackground : Result := RxResConst_Swe.SColorBackground;
     SColorActiveCaption : Result := RxResConst_Swe.SColorActiveCaption;
     SColorInactiveCaption : Result := RxResConst_Swe.SColorInactiveCaption;
     SColorMenu : Result := RxResConst_Swe.SColorMenu;
     SColorWindow : Result := RxResConst_Swe.SColorWindow;
     SColorWindowFrame : Result := RxResConst_Swe.SColorWindowFrame;
     SColorMenuText : Result := RxResConst_Swe.SColorMenuText;
     SColorWindowText : Result := RxResConst_Swe.SColorWindowText;
     SColorCaptionText : Result := RxResConst_Swe.SColorCaptionText;
     SColorActiveBorder : Result := RxResConst_Swe.SColorActiveBorder;
     SColorInactiveBorder : Result := RxResConst_Swe.SColorInactiveBorder;
     SColorAppWorkSpace : Result := RxResConst_Swe.SColorAppWorkSpace;
     SColorHighlight : Result := RxResConst_Swe.SColorHighlight;
     SColorHighlightText : Result := RxResConst_Swe.SColorHighlightText;
     SColorBtnFace : Result := RxResConst_Swe.SColorBtnFace;
     SColorBtnShadow : Result := RxResConst_Swe.SColorBtnShadow;
     SColorGrayText : Result := RxResConst_Swe.SColorGrayText;
     SColorBtnText : Result := RxResConst_Swe.SColorBtnText;
     SColorInactiveCaptionText : Result := RxResConst_Swe.SColorInactiveCaptionText;
     SColorBtnHighlight : Result := RxResConst_Swe.SColorBtnHighlight;
     SColor3DDkShadow : Result := RxResConst_Swe.SColor3DDkShadow;
     SColor3DLight : Result := RxResConst_Swe.SColor3DLight;
     SColorInfoText : Result := RxResConst_Swe.SColorInfoText;
     SColorInfoBk : Result := RxResConst_Swe.SColorInfoBk;
     SColorCream : Result := RxResConst_Swe.SColorCream;
     SColorMoneyGreen : Result := RxResConst_Swe.SColorMoneyGreen;
     SColorSkyBlue : Result := RxResConst_Swe.SColorSkyBlue;
     SColorNone : Result := RxResConst_Swe.SColorNone;
     SColorDefault : Result := RxResConst_Swe.SColorDefault;
     SColorCustom : Result := RxResConst_Swe.SColorCustom;
     SMemNoRecords : Result := RxResConst_Swe.SMemNoRecords;
     SInvalidFields : Result := RxResConst_Swe.SInvalidFields;
     SAutoSpeedbarMode : Result := RxResConst_Swe.SAutoSpeedbarMode;
    end;
  end;
  {$ENDIF}
  {$IFDEF RXLANG_Tur}
  function RxResConst_Tur_RxLoadStr(IDMsg: Integer): string;
  begin
    case IDMsg of
     SBrowse : Result := RxResConst_Tur.SBrowse;
     SDefaultFilter : Result := RxResConst_Tur.SDefaultFilter;
     SDateDlgTitle : Result := RxResConst_Tur.SDateDlgTitle;
     SNextYear : Result := RxResConst_Tur.SNextYear;
     SNextMonth : Result := RxResConst_Tur.SNextMonth;
     SPrevYear : Result := RxResConst_Tur.SPrevYear;
     SPrevMonth : Result := RxResConst_Tur.SPrevMonth;
     SNotImplemented : Result := RxResConst_Tur.SNotImplemented;
     SFileNotExec : Result := RxResConst_Tur.SFileNotExec;
     SLoadLibError : Result := RxResConst_Tur.SLoadLibError;
     SDetails : Result := RxResConst_Tur.SDetails;
     SDateOutOfRange : Result := RxResConst_Tur.SDateOutOfRange;
     SDateOutOfMin : Result := RxResConst_Tur.SDateOutOfMin;
     SDateOutOfMax : Result := RxResConst_Tur.SDateOutOfMax;
     SDateMinLimit : Result := RxResConst_Tur.SDateMinLimit;
     SDateMaxLimit : Result := RxResConst_Tur.SDateMaxLimit;
     SLocalDatabase : Result := RxResConst_Tur.SLocalDatabase;
     SRetryLogin : Result := RxResConst_Tur.SRetryLogin;
     SExprNotBoolean : Result := RxResConst_Tur.SExprNotBoolean;
     SExprBadNullTest : Result := RxResConst_Tur.SExprBadNullTest;
     SExprBadField : Result := RxResConst_Tur.SExprBadField;
     SCaptureFilter : Result := RxResConst_Tur.SCaptureFilter;
     SNotCaptureFilter : Result := RxResConst_Tur.SNotCaptureFilter;
     SInactiveData : Result := RxResConst_Tur.SInactiveData;
     SBrowseData : Result := RxResConst_Tur.SBrowseData;
     SEditData : Result := RxResConst_Tur.SEditData;
     SInsertData : Result := RxResConst_Tur.SInsertData;
     SSetKeyData : Result := RxResConst_Tur.SSetKeyData;
     SCalcFieldsData : Result := RxResConst_Tur.SCalcFieldsData;
     SRegistration : Result := RxResConst_Tur.SRegistration;
     SAppTitleLabel : Result := RxResConst_Tur.SAppTitleLabel;
     SHintLabel : Result := RxResConst_Tur.SHintLabel;
     SUserNameLabel : Result := RxResConst_Tur.SUserNameLabel;
     SPasswordLabel : Result := RxResConst_Tur.SPasswordLabel;
     SInvalidUserName : Result := RxResConst_Tur.SInvalidUserName;
     SChangePassword : Result := RxResConst_Tur.SChangePassword;
     SOldPasswordLabel : Result := RxResConst_Tur.SOldPasswordLabel;
     SNewPasswordLabel : Result := RxResConst_Tur.SNewPasswordLabel;
     SConfirmPasswordLabel : Result := RxResConst_Tur.SConfirmPasswordLabel;
     SPasswordChanged : Result := RxResConst_Tur.SPasswordChanged;
     SPasswordNotChanged : Result := RxResConst_Tur.SPasswordNotChanged;
     SPasswordsMismatch : Result := RxResConst_Tur.SPasswordsMismatch;
     SDBExceptCaption : Result := RxResConst_Tur.SDBExceptCaption;
     SBDEErrorLabel : Result := RxResConst_Tur.SBDEErrorLabel;
     SServerErrorLabel : Result := RxResConst_Tur.SServerErrorLabel;
     SErrorMsgLabel : Result := RxResConst_Tur.SErrorMsgLabel;
     SNextButton : Result := RxResConst_Tur.SNextButton;
     SPrevButton : Result := RxResConst_Tur.SPrevButton;
     SExprIncorrect : Result := RxResConst_Tur.SExprIncorrect;
     SExprTermination : Result := RxResConst_Tur.SExprTermination;
     SExprNameError : Result := RxResConst_Tur.SExprNameError;
     SExprStringError : Result := RxResConst_Tur.SExprStringError;
     SExprInvalidChar : Result := RxResConst_Tur.SExprInvalidChar;
     SExprNoRParen : Result := RxResConst_Tur.SExprNoRParen;
     SExprExpected : Result := RxResConst_Tur.SExprExpected;
     SExprBadCompare : Result := RxResConst_Tur.SExprBadCompare;
     SConfirmSave : Result := RxResConst_Tur.SConfirmSave;
     SDatabaseName : Result := RxResConst_Tur.SDatabaseName;
     SUnlockCaption : Result := RxResConst_Tur.SUnlockCaption;
     SUnlockHint : Result := RxResConst_Tur.SUnlockHint;
     SDeleteMultipleRecords : Result := RxResConst_Tur.SDeleteMultipleRecords;
     SGIFImage : Result := RxResConst_Tur.SGIFImage;
     SChangeGIFSize : Result := RxResConst_Tur.SChangeGIFSize;
     SNoGIFData : Result := RxResConst_Tur.SNoGIFData;
     SUnrecognizedGIFExt : Result := RxResConst_Tur.SUnrecognizedGIFExt;
     SWrongGIFColors : Result := RxResConst_Tur.SWrongGIFColors;
     SBadGIFCodeSize : Result := RxResConst_Tur.SBadGIFCodeSize;
     SGIFDecodeError : Result := RxResConst_Tur.SGIFDecodeError;
     SGIFEncodeError : Result := RxResConst_Tur.SGIFEncodeError;
     SGIFVersion : Result := RxResConst_Tur.SGIFVersion;
     SDualListSrcCaption : Result := RxResConst_Tur.SDualListSrcCaption;
     SDualListDestCaption : Result := RxResConst_Tur.SDualListDestCaption;
     SClipbrdUnknown : Result := RxResConst_Tur.SClipbrdUnknown;
     SClipbrdEmpty : Result := RxResConst_Tur.SClipbrdEmpty;
     SCustomizeSpeedbar : Result := RxResConst_Tur.SCustomizeSpeedbar;
     SAvailButtons : Result := RxResConst_Tur.SAvailButtons;
     SSpeedbarCategories : Result := RxResConst_Tur.SSpeedbarCategories;
     SSpeedbarEditHint : Result := RxResConst_Tur.SSpeedbarEditHint;
     SParseSyntaxError : Result := RxResConst_Tur.SParseSyntaxError;
     SParseNotCramp : Result := RxResConst_Tur.SParseNotCramp;
     SParseDivideByZero : Result := RxResConst_Tur.SParseDivideByZero;
     SParseSqrError : Result := RxResConst_Tur.SParseSqrError;
     SParseLogError : Result := RxResConst_Tur.SParseLogError;
     SParseInvalidFloatOperation : Result := RxResConst_Tur.SParseInvalidFloatOperation;
     srRXControls : Result := RxResConst_Tur.srRXControls;
     srRXDBAware : Result := RxResConst_Tur.srRXDBAware;
     srRXTools : Result := RxResConst_Tur.srRXTools;
     srStorageDesigner : Result := RxResConst_Tur.srStorageDesigner;
     srProxyEditor : Result := RxResConst_Tur.srProxyEditor;
     srPageProxies : Result := RxResConst_Tur.srPageProxies;
     srProxyName : Result := RxResConst_Tur.srProxyName;
     srPageName : Result := RxResConst_Tur.srPageName;
     srSBItemNotCreate : Result := RxResConst_Tur.srSBItemNotCreate;
     srConfirmSBDelete : Result := RxResConst_Tur.srConfirmSBDelete;
     srSpeedbarDesigner : Result := RxResConst_Tur.srSpeedbarDesigner;
     srNewSectionName : Result := RxResConst_Tur.srNewSectionName;
     srEventNotCreate : Result := RxResConst_Tur.srEventNotCreate;
     srTimerDesigner : Result := RxResConst_Tur.srTimerDesigner;
     srTimerEvents : Result := RxResConst_Tur.srTimerEvents;
     srAniCurFilter : Result := RxResConst_Tur.srAniCurFilter;
     srEditPicture : Result := RxResConst_Tur.srEditPicture;
     srLoadAniCursor : Result := RxResConst_Tur.srLoadAniCursor;
     srLoadIcon : Result := RxResConst_Tur.srLoadIcon;
     srSaveImageList : Result := RxResConst_Tur.srSaveImageList;
     srCaptionDesigner : Result := RxResConst_Tur.srCaptionDesigner;
     srGradientCaptions : Result := RxResConst_Tur.srGradientCaptions;
     srBorrowStructure : Result := RxResConst_Tur.srBorrowStructure;
     ccTimeEditHours : Result := RxResConst_Tur.ccTimeEditHours;
     ccTimeEditMins : Result := RxResConst_Tur.ccTimeEditMins;
     ccTimeEditSecs : Result := RxResConst_Tur.ccTimeEditSecs;
     SColorBlack : Result := RxResConst_Tur.SColorBlack;
     SColorMaroon : Result := RxResConst_Tur.SColorMaroon;
     SColorGreen : Result := RxResConst_Tur.SColorGreen;
     SColorOlive : Result := RxResConst_Tur.SColorOlive;
     SColorNavy : Result := RxResConst_Tur.SColorNavy;
     SColorPurple : Result := RxResConst_Tur.SColorPurple;
     SColorTeal : Result := RxResConst_Tur.SColorTeal;
     SColorGray : Result := RxResConst_Tur.SColorGray;
     SColorSilver : Result := RxResConst_Tur.SColorSilver;
     SColorRed : Result := RxResConst_Tur.SColorRed;
     SColorLime : Result := RxResConst_Tur.SColorLime;
     SColorYellow : Result := RxResConst_Tur.SColorYellow;
     SColorBlue : Result := RxResConst_Tur.SColorBlue;
     SColorFuchsia : Result := RxResConst_Tur.SColorFuchsia;
     SColorAqua : Result := RxResConst_Tur.SColorAqua;
     SColorWhite : Result := RxResConst_Tur.SColorWhite;
     SColorScrollBar : Result := RxResConst_Tur.SColorScrollBar;
     SColorBackground : Result := RxResConst_Tur.SColorBackground;
     SColorActiveCaption : Result := RxResConst_Tur.SColorActiveCaption;
     SColorInactiveCaption : Result := RxResConst_Tur.SColorInactiveCaption;
     SColorMenu : Result := RxResConst_Tur.SColorMenu;
     SColorWindow : Result := RxResConst_Tur.SColorWindow;
     SColorWindowFrame : Result := RxResConst_Tur.SColorWindowFrame;
     SColorMenuText : Result := RxResConst_Tur.SColorMenuText;
     SColorWindowText : Result := RxResConst_Tur.SColorWindowText;
     SColorCaptionText : Result := RxResConst_Tur.SColorCaptionText;
     SColorActiveBorder : Result := RxResConst_Tur.SColorActiveBorder;
     SColorInactiveBorder : Result := RxResConst_Tur.SColorInactiveBorder;
     SColorAppWorkSpace : Result := RxResConst_Tur.SColorAppWorkSpace;
     SColorHighlight : Result := RxResConst_Tur.SColorHighlight;
     SColorHighlightText : Result := RxResConst_Tur.SColorHighlightText;
     SColorBtnFace : Result := RxResConst_Tur.SColorBtnFace;
     SColorBtnShadow : Result := RxResConst_Tur.SColorBtnShadow;
     SColorGrayText : Result := RxResConst_Tur.SColorGrayText;
     SColorBtnText : Result := RxResConst_Tur.SColorBtnText;
     SColorInactiveCaptionText : Result := RxResConst_Tur.SColorInactiveCaptionText;
     SColorBtnHighlight : Result := RxResConst_Tur.SColorBtnHighlight;
     SColor3DDkShadow : Result := RxResConst_Tur.SColor3DDkShadow;
     SColor3DLight : Result := RxResConst_Tur.SColor3DLight;
     SColorInfoText : Result := RxResConst_Tur.SColorInfoText;
     SColorInfoBk : Result := RxResConst_Tur.SColorInfoBk;
     SColorCream : Result := RxResConst_Tur.SColorCream;
     SColorMoneyGreen : Result := RxResConst_Tur.SColorMoneyGreen;
     SColorSkyBlue : Result := RxResConst_Tur.SColorSkyBlue;
     SColorNone : Result := RxResConst_Tur.SColorNone;
     SColorDefault : Result := RxResConst_Tur.SColorDefault;
     SColorCustom : Result := RxResConst_Tur.SColorCustom;
     SMemNoRecords : Result := RxResConst_Tur.SMemNoRecords;
     SInvalidFields : Result := RxResConst_Tur.SInvalidFields;
     SAutoSpeedbarMode : Result := RxResConst_Tur.SAutoSpeedbarMode;
    end;
  end;
  {$ENDIF}
  {$IFDEF RXLANG_Ukr}
  function RxResConst_Ukr_RxLoadStr(IDMsg: Integer): string;
  begin
    case IDMsg of
     SBrowse : Result := RxResConst_Ukr.SBrowse;
     SDefaultFilter : Result := RxResConst_Ukr.SDefaultFilter;
     SDateDlgTitle : Result := RxResConst_Ukr.SDateDlgTitle;
     SNextYear : Result := RxResConst_Ukr.SNextYear;
     SNextMonth : Result := RxResConst_Ukr.SNextMonth;
     SPrevYear : Result := RxResConst_Ukr.SPrevYear;
     SPrevMonth : Result := RxResConst_Ukr.SPrevMonth;
     SNotImplemented : Result := RxResConst_Ukr.SNotImplemented;
     SFileNotExec : Result := RxResConst_Ukr.SFileNotExec;
     SLoadLibError : Result := RxResConst_Ukr.SLoadLibError;
     SDetails : Result := RxResConst_Ukr.SDetails;
     SDateOutOfRange : Result := RxResConst_Ukr.SDateOutOfRange;
     SDateOutOfMin : Result := RxResConst_Ukr.SDateOutOfMin;
     SDateOutOfMax : Result := RxResConst_Ukr.SDateOutOfMax;
     SDateMinLimit : Result := RxResConst_Ukr.SDateMinLimit;
     SDateMaxLimit : Result := RxResConst_Ukr.SDateMaxLimit;
     SLocalDatabase : Result := RxResConst_Ukr.SLocalDatabase;
     SRetryLogin : Result := RxResConst_Ukr.SRetryLogin;
     SExprNotBoolean : Result := RxResConst_Ukr.SExprNotBoolean;
     SExprBadNullTest : Result := RxResConst_Ukr.SExprBadNullTest;
     SExprBadField : Result := RxResConst_Ukr.SExprBadField;
     SCaptureFilter : Result := RxResConst_Ukr.SCaptureFilter;
     SNotCaptureFilter : Result := RxResConst_Ukr.SNotCaptureFilter;
     SInactiveData : Result := RxResConst_Ukr.SInactiveData;
     SBrowseData : Result := RxResConst_Ukr.SBrowseData;
     SEditData : Result := RxResConst_Ukr.SEditData;
     SInsertData : Result := RxResConst_Ukr.SInsertData;
     SSetKeyData : Result := RxResConst_Ukr.SSetKeyData;
     SCalcFieldsData : Result := RxResConst_Ukr.SCalcFieldsData;
     SRegistration : Result := RxResConst_Ukr.SRegistration;
     SAppTitleLabel : Result := RxResConst_Ukr.SAppTitleLabel;
     SHintLabel : Result := RxResConst_Ukr.SHintLabel;
     SUserNameLabel : Result := RxResConst_Ukr.SUserNameLabel;
     SPasswordLabel : Result := RxResConst_Ukr.SPasswordLabel;
     SInvalidUserName : Result := RxResConst_Ukr.SInvalidUserName;
     SChangePassword : Result := RxResConst_Ukr.SChangePassword;
     SOldPasswordLabel : Result := RxResConst_Ukr.SOldPasswordLabel;
     SNewPasswordLabel : Result := RxResConst_Ukr.SNewPasswordLabel;
     SConfirmPasswordLabel : Result := RxResConst_Ukr.SConfirmPasswordLabel;
     SPasswordChanged : Result := RxResConst_Ukr.SPasswordChanged;
     SPasswordNotChanged : Result := RxResConst_Ukr.SPasswordNotChanged;
     SPasswordsMismatch : Result := RxResConst_Ukr.SPasswordsMismatch;
     SDBExceptCaption : Result := RxResConst_Ukr.SDBExceptCaption;
     SBDEErrorLabel : Result := RxResConst_Ukr.SBDEErrorLabel;
     SServerErrorLabel : Result := RxResConst_Ukr.SServerErrorLabel;
     SErrorMsgLabel : Result := RxResConst_Ukr.SErrorMsgLabel;
     SNextButton : Result := RxResConst_Ukr.SNextButton;
     SPrevButton : Result := RxResConst_Ukr.SPrevButton;
     SExprIncorrect : Result := RxResConst_Ukr.SExprIncorrect;
     SExprTermination : Result := RxResConst_Ukr.SExprTermination;
     SExprNameError : Result := RxResConst_Ukr.SExprNameError;
     SExprStringError : Result := RxResConst_Ukr.SExprStringError;
     SExprInvalidChar : Result := RxResConst_Ukr.SExprInvalidChar;
     SExprNoRParen : Result := RxResConst_Ukr.SExprNoRParen;
     SExprExpected : Result := RxResConst_Ukr.SExprExpected;
     SExprBadCompare : Result := RxResConst_Ukr.SExprBadCompare;
     SConfirmSave : Result := RxResConst_Ukr.SConfirmSave;
     SDatabaseName : Result := RxResConst_Ukr.SDatabaseName;
     SUnlockCaption : Result := RxResConst_Ukr.SUnlockCaption;
     SUnlockHint : Result := RxResConst_Ukr.SUnlockHint;
     SDeleteMultipleRecords : Result := RxResConst_Ukr.SDeleteMultipleRecords;
     SGIFImage : Result := RxResConst_Ukr.SGIFImage;
     SChangeGIFSize : Result := RxResConst_Ukr.SChangeGIFSize;
     SNoGIFData : Result := RxResConst_Ukr.SNoGIFData;
     SUnrecognizedGIFExt : Result := RxResConst_Ukr.SUnrecognizedGIFExt;
     SWrongGIFColors : Result := RxResConst_Ukr.SWrongGIFColors;
     SBadGIFCodeSize : Result := RxResConst_Ukr.SBadGIFCodeSize;
     SGIFDecodeError : Result := RxResConst_Ukr.SGIFDecodeError;
     SGIFEncodeError : Result := RxResConst_Ukr.SGIFEncodeError;
     SGIFVersion : Result := RxResConst_Ukr.SGIFVersion;
     SDualListSrcCaption : Result := RxResConst_Ukr.SDualListSrcCaption;
     SDualListDestCaption : Result := RxResConst_Ukr.SDualListDestCaption;
     SClipbrdUnknown : Result := RxResConst_Ukr.SClipbrdUnknown;
     SClipbrdEmpty : Result := RxResConst_Ukr.SClipbrdEmpty;
     SCustomizeSpeedbar : Result := RxResConst_Ukr.SCustomizeSpeedbar;
     SAvailButtons : Result := RxResConst_Ukr.SAvailButtons;
     SSpeedbarCategories : Result := RxResConst_Ukr.SSpeedbarCategories;
     SSpeedbarEditHint : Result := RxResConst_Ukr.SSpeedbarEditHint;
     SParseSyntaxError : Result := RxResConst_Ukr.SParseSyntaxError;
     SParseNotCramp : Result := RxResConst_Ukr.SParseNotCramp;
     SParseDivideByZero : Result := RxResConst_Ukr.SParseDivideByZero;
     SParseSqrError : Result := RxResConst_Ukr.SParseSqrError;
     SParseLogError : Result := RxResConst_Ukr.SParseLogError;
     SParseInvalidFloatOperation : Result := RxResConst_Ukr.SParseInvalidFloatOperation;
     srRXControls : Result := RxResConst_Ukr.srRXControls;
     srRXDBAware : Result := RxResConst_Ukr.srRXDBAware;
     srRXTools : Result := RxResConst_Ukr.srRXTools;
     srStorageDesigner : Result := RxResConst_Ukr.srStorageDesigner;
     srProxyEditor : Result := RxResConst_Ukr.srProxyEditor;
     srPageProxies : Result := RxResConst_Ukr.srPageProxies;
     srProxyName : Result := RxResConst_Ukr.srProxyName;
     srPageName : Result := RxResConst_Ukr.srPageName;
     srSBItemNotCreate : Result := RxResConst_Ukr.srSBItemNotCreate;
     srConfirmSBDelete : Result := RxResConst_Ukr.srConfirmSBDelete;
     srSpeedbarDesigner : Result := RxResConst_Ukr.srSpeedbarDesigner;
     srNewSectionName : Result := RxResConst_Ukr.srNewSectionName;
     srEventNotCreate : Result := RxResConst_Ukr.srEventNotCreate;
     srTimerDesigner : Result := RxResConst_Ukr.srTimerDesigner;
     srTimerEvents : Result := RxResConst_Ukr.srTimerEvents;
     srAniCurFilter : Result := RxResConst_Ukr.srAniCurFilter;
     srEditPicture : Result := RxResConst_Ukr.srEditPicture;
     srLoadAniCursor : Result := RxResConst_Ukr.srLoadAniCursor;
     srLoadIcon : Result := RxResConst_Ukr.srLoadIcon;
     srSaveImageList : Result := RxResConst_Ukr.srSaveImageList;
     srCaptionDesigner : Result := RxResConst_Ukr.srCaptionDesigner;
     srGradientCaptions : Result := RxResConst_Ukr.srGradientCaptions;
     srBorrowStructure : Result := RxResConst_Ukr.srBorrowStructure;
     ccTimeEditHours : Result := RxResConst_Ukr.ccTimeEditHours;
     ccTimeEditMins : Result := RxResConst_Ukr.ccTimeEditMins;
     ccTimeEditSecs : Result := RxResConst_Ukr.ccTimeEditSecs;
     SColorBlack : Result := RxResConst_Ukr.SColorBlack;
     SColorMaroon : Result := RxResConst_Ukr.SColorMaroon;
     SColorGreen : Result := RxResConst_Ukr.SColorGreen;
     SColorOlive : Result := RxResConst_Ukr.SColorOlive;
     SColorNavy : Result := RxResConst_Ukr.SColorNavy;
     SColorPurple : Result := RxResConst_Ukr.SColorPurple;
     SColorTeal : Result := RxResConst_Ukr.SColorTeal;
     SColorGray : Result := RxResConst_Ukr.SColorGray;
     SColorSilver : Result := RxResConst_Ukr.SColorSilver;
     SColorRed : Result := RxResConst_Ukr.SColorRed;
     SColorLime : Result := RxResConst_Ukr.SColorLime;
     SColorYellow : Result := RxResConst_Ukr.SColorYellow;
     SColorBlue : Result := RxResConst_Ukr.SColorBlue;
     SColorFuchsia : Result := RxResConst_Ukr.SColorFuchsia;
     SColorAqua : Result := RxResConst_Ukr.SColorAqua;
     SColorWhite : Result := RxResConst_Ukr.SColorWhite;
     SColorScrollBar : Result := RxResConst_Ukr.SColorScrollBar;
     SColorBackground : Result := RxResConst_Ukr.SColorBackground;
     SColorActiveCaption : Result := RxResConst_Ukr.SColorActiveCaption;
     SColorInactiveCaption : Result := RxResConst_Ukr.SColorInactiveCaption;
     SColorMenu : Result := RxResConst_Ukr.SColorMenu;
     SColorWindow : Result := RxResConst_Ukr.SColorWindow;
     SColorWindowFrame : Result := RxResConst_Ukr.SColorWindowFrame;
     SColorMenuText : Result := RxResConst_Ukr.SColorMenuText;
     SColorWindowText : Result := RxResConst_Ukr.SColorWindowText;
     SColorCaptionText : Result := RxResConst_Ukr.SColorCaptionText;
     SColorActiveBorder : Result := RxResConst_Ukr.SColorActiveBorder;
     SColorInactiveBorder : Result := RxResConst_Ukr.SColorInactiveBorder;
     SColorAppWorkSpace : Result := RxResConst_Ukr.SColorAppWorkSpace;
     SColorHighlight : Result := RxResConst_Ukr.SColorHighlight;
     SColorHighlightText : Result := RxResConst_Ukr.SColorHighlightText;
     SColorBtnFace : Result := RxResConst_Ukr.SColorBtnFace;
     SColorBtnShadow : Result := RxResConst_Ukr.SColorBtnShadow;
     SColorGrayText : Result := RxResConst_Ukr.SColorGrayText;
     SColorBtnText : Result := RxResConst_Ukr.SColorBtnText;
     SColorInactiveCaptionText : Result := RxResConst_Ukr.SColorInactiveCaptionText;
     SColorBtnHighlight : Result := RxResConst_Ukr.SColorBtnHighlight;
     SColor3DDkShadow : Result := RxResConst_Ukr.SColor3DDkShadow;
     SColor3DLight : Result := RxResConst_Ukr.SColor3DLight;
     SColorInfoText : Result := RxResConst_Ukr.SColorInfoText;
     SColorInfoBk : Result := RxResConst_Ukr.SColorInfoBk;
     SColorCream : Result := RxResConst_Ukr.SColorCream;
     SColorMoneyGreen : Result := RxResConst_Ukr.SColorMoneyGreen;
     SColorSkyBlue : Result := RxResConst_Ukr.SColorSkyBlue;
     SColorNone : Result := RxResConst_Ukr.SColorNone;
     SColorDefault : Result := RxResConst_Ukr.SColorDefault;
     SColorCustom : Result := RxResConst_Ukr.SColorCustom;
     SMemNoRecords : Result := RxResConst_Ukr.SMemNoRecords;
     SInvalidFields : Result := RxResConst_Ukr.SInvalidFields;
     SAutoSpeedbarMode : Result := RxResConst_Ukr.SAutoSpeedbarMode;
    end;
  end;
  {$ENDIF}
begin              

  case RxDefaultLang of
    {$IFDEF RXLANG_Bul}rlBul: Result := RxResConst_Bul_RxLoadStr(IdMsg);{$ENDIF}
    {$IFDEF RXLANG_Chi}rlChi: Result := RxResConst_Chi_RxLoadStr(IdMsg);{$ENDIF}
    {$IFDEF RXLANG_Cht}rlCht: Result := RxResConst_Cht_RxLoadStr(IdMsg);{$ENDIF}
    {$IFDEF RXLANG_Cze}rlCze: Result := RxResConst_Cze_RxLoadStr(IdMsg);{$ENDIF}
    {$IFDEF RXLANG_Dan}rlDan: Result := RxResConst_Dan_RxLoadStr(IdMsg);{$ENDIF}
    {$IFDEF RXLANG_Dut}rlDut: Result := RxResConst_Dut_RxLoadStr(IdMsg);{$ENDIF}
    {$IFDEF RXLANG_Eng}rlEng: Result := RxResConst_Eng_RxLoadStr(IdMsg);{$ENDIF}
    {$IFDEF RXLANG_Fra}rlFra: Result := RxResConst_Fra_RxLoadStr(IdMsg);{$ENDIF}
    {$IFDEF RXLANG_Gal}rlGal: Result := RxResConst_Gal_RxLoadStr(IdMsg);{$ENDIF}
    {$IFDEF RXLANG_Ger}rlGer: Result := RxResConst_Ger_RxLoadStr(IdMsg);{$ENDIF}
    {$IFDEF RXLANG_Gre}rlGre: Result := RxResConst_Gre_RxLoadStr(IdMsg);{$ENDIF}
    {$IFDEF RXLANG_Hun}rlHun: Result := RxResConst_Hun_RxLoadStr(IdMsg);{$ENDIF}
    {$IFDEF RXLANG_Ita}rlIta: Result := RxResConst_Ita_RxLoadStr(IdMsg);{$ENDIF}
    {$IFDEF RXLANG_Kor}rlKor: Result := RxResConst_Kor_RxLoadStr(IdMsg);{$ENDIF}
    {$IFDEF RXLANG_Mex}rlMex: Result := RxResConst_Mex_RxLoadStr(IdMsg);{$ENDIF}
    {$IFDEF RXLANG_Pol}rlPol: Result := RxResConst_Pol_RxLoadStr(IdMsg);{$ENDIF}
    {$IFDEF RXLANG_Por}rlPor: Result := RxResConst_Por_RxLoadStr(IdMsg);{$ENDIF}
    {$IFDEF RXLANG_Rom}rlRom: Result := RxResConst_Rom_RxLoadStr(IdMsg);{$ENDIF}
    {$IFDEF RXLANG_Rus}rlRus: Result := RxResConst_Rus_RxLoadStr(IdMsg);{$ENDIF}
    {$IFDEF RXLANG_Slo}rlSlo: Result := RxResConst_Slo_RxLoadStr(IdMsg);{$ENDIF}
    {$IFDEF RXLANG_Spa}rlSpa: Result := RxResConst_Spa_RxLoadStr(IdMsg);{$ENDIF}
    {$IFDEF RXLANG_Srb}rlSrb: Result := RxResConst_Srb_RxLoadStr(IdMsg);{$ENDIF}
    {$IFDEF RXLANG_Svk}rlSvk: Result := RxResConst_Svk_RxLoadStr(IdMsg);{$ENDIF}
    {$IFDEF RXLANG_Swe}rlSwe: Result := RxResConst_Swe_RxLoadStr(IdMsg);{$ENDIF}
    {$IFDEF RXLANG_Tur}rlTur: Result := RxResConst_Tur_RxLoadStr(IdMsg);{$ENDIF}
    {$IFDEF RXLANG_Ukr}rlUkr: Result := RxResConst_Ukr_RxLoadStr(IdMsg);{$ENDIF}
  {$IFDEF UNICODE}
  else
    {for unsupported lang}
    Result := RxResConst_Eng_RxLoadStr(IdMsg);
  {$ENDIF}
  end;
  {$IFDEF UNICODE}
  {when is null then try English as default}
  if Result = '' then
    Result := RxResConst_Eng_RxLoadStr(IdMsg);
  {$ENDIF}
end;


{-------------------------------------------------------------------------------
  Procedure: GetDefaultLangByLocale
  Author:    jaro.benes
  DateTime:  2011.03.02
  Arguments: None
  Result:    TRxLang
  Purpose:   Local function for evaluate language by system locale
             (by default is English)
-------------------------------------------------------------------------------}
function GetDefaultLangByLocale:TRxLang;
var
  L: Integer;
begin
  L := Sysutils.SysLocale.PriLangID;
  case L of
    LANG_NEUTRAL: Result := rlEng;
    LANG_BULGARIAN: Result := rlBul;
    LANG_CHINESE:
    begin
      case Sysutils.SysLocale.SubLangID of
        SUBLANG_CHINESE_TRADITIONAL: Result := rlCht;
        SUBLANG_CHINESE_SIMPLIFIED: Result := rlChi;
      else
        Result := rlChi;
      end;
    end;
    LANG_CZECH: Result := rlCze;
    LANG_DANISH: Result := rlDan;
    LANG_DUTCH: Result := rlDut;
    LANG_ENGLISH: Result := rlEng;
    LANG_FRENCH: Result := rlFra;
    //not default lang
    //cannot be evaluate by locale, must be set up manually
    //LANG_GALICIAN: Result := rlGal;
    LANG_GERMAN: Result := rlGer;
    LANG_GREEK: Result := rlGre;
    LANG_HUNGARIAN: Result := rlHun;
    LANG_ITALIAN: Result := rlIta;
    LANG_KOREAN: Result := rlKor;
    LANG_POLISH: Result := rlPol;
    LANG_PORTUGUESE: Result := rlPor;
    LANG_ROMANIAN: Result := rlRom;
    LANG_RUSSIAN: Result := rlRus;
    LANG_SLOVENIAN: Result := rlSlo;
    LANG_SPANISH:
    begin
      case Sysutils.SysLocale.SubLangID of
        SUBLANG_SPANISH_MEXICAN: Result := rlMex;
      else
        Result := rlSpa;
      end;
    end;
    LANG_SERBIAN: Result := rlSrb;
    LANG_SLOVAK: Result := rlSvk;
    LANG_SWEDISH: Result := rlSwe;
    LANG_TURKISH: Result := rlTur;
    LANG_UKRAINIAN: Result := rlUkr;
  end;
end;

initialization
  RxDefaultLang := GetDefaultLangByLocale;
end.